package org.processmining.decomposelogs;

import java.util.HashMap;
import java.util.HashSet;
import java.util.Iterator;
import java.util.Map;
import java.util.Set;

import org.deckfour.xes.extension.std.XConceptExtension;
import org.deckfour.xes.factory.XFactory;
import org.deckfour.xes.factory.XFactoryRegistry;
import org.deckfour.xes.model.XAttributeMap;
import org.deckfour.xes.model.XEvent;
import org.deckfour.xes.model.XLog;
import org.deckfour.xes.model.XTrace;
import org.processmining.acceptingpetrinet.models.AcceptingPetriNet;
import org.processmining.acceptingpetrinet.models.impl.AcceptingPetriNetImpl;
import org.processmining.framework.plugin.ProMCanceller;
import org.processmining.models.graphbased.directed.petrinet.Petrinet;
import org.processmining.models.graphbased.directed.petrinet.elements.Transition;
import org.processmining.projectedrecallandprecision.framework.CompareParameters;
import org.processmining.projectedrecallandprecision.helperclasses.AutomatonFailedException;
import org.processmining.projectedrecallandprecision.plugins.CompareLog2PetriNetPlugin;
import org.processmining.projectedrecallandprecision.result.ProjectedRecallPrecisionResult;
import org.processmining.projectedrecallandprecision.result.ProjectedRecallPrecisionResult.ProjectedMeasuresFailedException;

public class DefaultConformance {

	public static float[] pccConformance(XLog log, Petrinet pn) {
		float[] fitPre = new float[3];
		int totEvts = 0, evtsCovered = 0;
		Set<String> actsInNet = new HashSet<String>();
		for (Transition t : pn.getTransitions())
			if (!t.isInvisible())
				actsInNet.add(t.getLabel());

		XFactory factory = XFactoryRegistry.instance().currentDefault();
		XLog subLog = factory.createLog((XAttributeMap) log.getAttributes().clone());
		subLog.getClassifiers().addAll(log.getClassifiers());
		subLog.getExtensions().addAll(log.getExtensions());
		subLog.getGlobalTraceAttributes().addAll(log.getGlobalTraceAttributes());
		subLog.getGlobalEventAttributes().addAll(log.getGlobalEventAttributes());

		//Walk thought each trace of the original log
		Map<Integer, XTrace> tracesMap = new HashMap<Integer, XTrace>();
		Iterator<XTrace> itTrace = log.iterator();
		int num = 0;
		for (XTrace trace : log) {
			XTrace trace2 = factory.createTrace((XAttributeMap) trace.getAttributes().clone());
			for (XEvent event : trace) {

				if (actsInNet.contains(XConceptExtension.instance().extractName(event).toString())) {
					XEvent event2 = (XEvent) event.clone();
					trace2.add(event2);
					evtsCovered++;
				}
				totEvts++;
			}
			subLog.add(trace2);
		}

		AcceptingPetriNet model = new AcceptingPetriNetImpl(pn);
		CompareLog2PetriNetPlugin compare = new CompareLog2PetriNetPlugin();
		CompareParameters param = new CompareParameters();

		//		System.out.println("\n LOG: ");
		//		for (XTrace t : log)
		//			for (XEvent e : t)
		//				System.out.println(XConceptExtension.instance().extractName(e));
		if (actsInNet.size() < 3)
			return fitPre;
		param.setK(actsInNet.size() > 1 ? 2 : 1);
		ProjectedRecallPrecisionResult result = null;
		try {
			result = compare.measure(subLog, model, param, new ProMCanceller() {
				public boolean isCancelled() {
					return false;
				}
			});
		} catch (AutomatonFailedException | InterruptedException | ProjectedMeasuresFailedException e) {
			e.printStackTrace();
		}
		try {
			fitPre[0] = (float) result.getRecall();
			fitPre[1] = (float) result.getPrecision();
		} catch (ProjectedMeasuresFailedException e) {
			// TODO Auto-generated catch block
			e.printStackTrace();
		}
		fitPre[2] = (float) evtsCovered / (float) totEvts;

		return fitPre;
	}
}
