package org.processmining.decomposelogs;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Set;
import java.util.concurrent.ExecutorService;
import java.util.concurrent.Executors;
import java.util.concurrent.TimeUnit;

import org.deckfour.xes.classification.XEventNameClassifier;
import org.deckfour.xes.extension.std.XConceptExtension;
import org.deckfour.xes.model.XEvent;
import org.deckfour.xes.model.XLog;
import org.deckfour.xes.model.XTrace;
import org.processmining.prabhakardixit.alignments.CombinationUtilsL;

public class Log2DeclareL {

	public static AggregatedHeuristicsL getAllAggPairHeuristics(final XLog log) {

		final Set<String> acts = new HashSet<String>();
		for (XTrace t : log)
			for (XEvent e : t)
				acts.add(XConceptExtension.instance().extractName(e));

		long msecs = System.currentTimeMillis();
		int NTHREADS = Math.max(1, Runtime.getRuntime().availableProcessors());
		final AggregatedHeuristicsL[] ahls = new AggregatedHeuristicsL[log.size() > NTHREADS ? NTHREADS : 1];
		if (log.size() > NTHREADS) {
			ExecutorService exec = Executors.newFixedThreadPool(NTHREADS - 1);
			final int segmentLen = (log.size() / (NTHREADS - 1)) < 1 ? 1 : log.size() / NTHREADS - 1;
			int offset = 0;
			for (int i = 0; i < NTHREADS - 1; i++) {
				final int j = i;
				final int from = offset;
				final int to = offset + segmentLen;
				exec.execute(new Runnable() {

					@Override
					public void run() {
						getAggPairHeuristics(from, to, log, ahls, j, acts);
						//					work(array, from, to);
					}
				});
				offset += segmentLen;
			}

			// do last segment on main thread
			getAggPairHeuristics(offset, log.size(), log, ahls, ahls.length - 1, acts);

			//process all the dependencies between clusters and other acts
			//			processMappingClusters(context, pn, mapper, actCombClusters, logAgg, fitness, precision);

			//				 wait for completion
			exec.shutdown();
			try {
				exec.awaitTermination(Long.MAX_VALUE, TimeUnit.NANOSECONDS);
			} catch (InterruptedException e) {
				// TODO Auto-generated catch block
				e.printStackTrace();
			}
		} else {
			getAggPairHeuristics(0, log.size(), log, ahls, ahls.length - 1, acts);
		}

		//merge ahls
		AggregatedHeuristicsL ahl = new AggregatedHeuristicsL();
		System.out.println("Start Merging");
		int count = 0;
		for (AggregatedHeuristicsL ahli : ahls) {
			System.out.println("Start Merging" + count++);
			for (String a1 : ahli.actAbsOccCaseOcc.keySet()) {
				if (!ahl.actAbsOccCaseOcc.containsKey(a1))
					ahl.actAbsOccCaseOcc.put(a1, new HashMap<Short, Short>());
				for (Short key : ahli.actAbsOccCaseOcc.get(a1).keySet()) {
					short val = (short) (ahl.actAbsOccCaseOcc.get(a1).containsKey(key) ? ahl.actAbsOccCaseOcc.get(a1)
							.get(key) + ahli.actAbsOccCaseOcc.get(a1).get(key) : ahli.actAbsOccCaseOcc.get(a1).get(key));
					ahl.actAbsOccCaseOcc.get(a1).put(key, val);
				}
			}

			if (ahl.getActComb().isEmpty()) {
				ahl = ahli;
			} else {
				for (Set<String> ac : ahli.actComb)
					if (!ahl.getActComb().contains(ac))
						ahl.getActComb().add(ac);

				for (int i = 0; i < ahli.actComb.size(); i++) {
					int i2 = -1;
					for (Set<String> ac : ahl.getActComb())
						if (ac.equals(ahli.getActComb().get(i)))
							i2 = ahl.getActComb().indexOf(ac);

					for (DeclarePairsL dp : ahli.getDeclarePairs().get(i)) {
						boolean added = false;
						for (DeclarePairsL dp2 : ahl.getDeclarePairs().get(i2))
							if (!added && dp.getAct1().equals(dp2.getAct1())
									&& dp.getTemplate().equals(dp2.getTemplate())
									&& dp.getLastAct().equals(dp2.getLastAct())) {
								dp2.setFreq((short) (dp2.getFreq() + dp.getFreq()));
								added = true;
							}

						if (!added)
							ahl.getDeclarePairs().get(i2).add(dp);

					}

				}

			}
		}
		return ahl;
	}

	public static void getAggPairHeuristics(int start, int end, XLog log, AggregatedHeuristicsL[] ahls, int index,
			Set<String> acts) {
		//		XFactory f = XFactoryRegistry.instance().currentDefault();
		//		XLog log2 = f.createLog();
		//				log2.getClassifiers().addAll();
		//				log2.getExtensions().addAll(log.getExtensions());
		//				log2.getGlobalTraceAttributes().addAll(log.getGlobalTraceAttributes());
		//				log2.getGlobalEventAttributes().addAll(log.getGlobalEventAttributes());
		//				log2.getAttributes().putAll(log.getAttributes());
		final XEventNameClassifier lifecycleClassifier = new XEventNameClassifier();
		//		for (int i = start; i < end; i++) {
		//			XTrace t = (XTrace) log.get(i).clone();
		//			//			t.getExtensions().addAll(log.get(i).getExtensions());
		//			//			t.getAttributes().putAll(log.get(i).getAttributes());
		//			log2.add(t);
		//
		//		}

		System.out.println(index);
		//		for (XEventClass e : log.getInfo(lifecycleClassifier).getEventClasses().getClasses())
		//			acts.add(e.toString());
		ahls[index] = getAggPairHeuristics(log, acts, start, end);
	}

	public static AggregatedHeuristicsL getAggPairHeuristics(XLog log, Set<String> acts, int start, int end) {
		AggregatedHeuristicsL ah = new AggregatedHeuristicsL();
		//		Set<String> acts = new HashSet<String>();
		//		Set<String> acts = new HashSet<String>();
		//		final XEventNameClassifier lifecycleClassifier = new XEventNameClassifier();
		//		XLogInfo info = XLogInfoFactory.createLogInfo(log, lifecycleClassifier);
		//								for (XEventClass e : log.getInfo(lifecycleClassifier).getEventClasses().getClasses())
		//		for (XEventClass e : info.getEventClasses().getClasses())
		//			acts.add(e.toString());
		Map<Integer, String> actMap = new HashMap<Integer, String>();
		int j = 0;
		for (String t : acts)
			actMap.put(j++, t);
		Set<Set<Integer>> combs = CombinationUtilsL.getAllCombs(actMap.keySet(), 2);
		final List<Set<String>> actComb = new ArrayList<Set<String>>(combs.size()); //list of act combs
		List<Set<DeclarePairsL>> aggLogH = new ArrayList<Set<DeclarePairsL>>();

		for (Set<Integer> c : combs) {
			Set<String> ac = new HashSet<String>();
			for (int l : c)
				ac.add(actMap.get(l));
			actComb.add(ac);
			aggLogH.add(new HashSet<DeclarePairsL>());
		}
		ah.setActComb(actComb);
		//		Set<DeclarePairs> globaldps = new HashSet<DeclarePairs>();

		//declare constraint generation code starts
		for (int i = start; i < end; i++) {
			XTrace trace = log.get(i);
			Set<DeclarePairsL> dps = new HashSet<DeclarePairsL>();
			Set<String> addedActs = new HashSet<String>();
			Set<String> actsNotAdded = new HashSet<String>(acts);
			Map<String, Short> actOccurrences = new HashMap<String, Short>();
			for (XEvent event : trace) {
				String act = XConceptExtension.instance().extractName(event);
				addedActs.add(act);

				//remove declare pairs
				removeInvaliadtedPairs(act, addedActs, dps);

				//add new declare pairs
				if (actsNotAdded.remove(act))
					addDeclarePairs(act, actsNotAdded, dps);

				//update last act
				updateLastAct(dps, act);
				actOccurrences.put(act, (short) (actOccurrences.containsKey(act) ? actOccurrences.get(act) + 1 : 1));

			}

			for (String a1 : actsNotAdded) {
				for (String a2 : actsNotAdded)
					if (!a1.equals(a2)) {
						DeclarePairsL dp = new DeclarePairsL(a1, a2, DeclareTemplatesL.None);
						dp.setLastAct(a1);
						dps.add(dp);

					}
			}

			//update single act relations
			for (String act : actOccurrences.keySet()) {
				if (actOccurrences.get(act) > 1)
					System.out.println();

				if (ah.actAbsOccCaseOcc.containsKey(act)) {
					boolean added = false;
					for (Short key : ah.actAbsOccCaseOcc.get(act).keySet()) {
						if (key.equals(actOccurrences.get(act))) {
							added = true;
							ah.actAbsOccCaseOcc.get(act).put(key, (short) (ah.actAbsOccCaseOcc.get(act).get(key) + 1));
						}
					}
					if (!added)
						ah.actAbsOccCaseOcc.get(act).put(actOccurrences.get(act), (short) 1);
				} else {
					ah.actAbsOccCaseOcc.put(act, new HashMap<Short, Short>());
					ah.actAbsOccCaseOcc.get(act).put(actOccurrences.get(act), (short) 1);
				}
			}

			//keep only valid constraints with last acts
			removeInvalidLastActs(dps);

			//tidy up by removing very broad constraints
			keepMostStrictConstraint(dps, acts);

			//update global dps obj
			//			updateGlobalDps(globaldps, dps);

			//update list candidates
			updateListCandidates(aggLogH, dps, actComb);
		}

		//		for (DeclarePairs dp : globaldps) {
		//			System.out.println(dp.getAct1() + "  " + dp.getAct2() + "  " + dp.getTemplate() + " " + dp.getFreq());
		//		}

		ah.dp = aggLogH;
		return ah;

		//declare constraint generation code ends	
	}

	private static void removeInvalidLastActs(Set<DeclarePairsL> dps) {
		Set<DeclarePairsL> pairsToRemove = new HashSet<DeclarePairsL>();
		for (DeclarePairsL dp : dps) {
			switch (dp.getTemplate()) {
				case Alternate :
				case RepeatAAlternate :
				case RepeatBResponse :
				case RepeatBothAlternate :
				case RepeatBothResponse :
				case Response :
				case RepeatAResponse :
				case RepeatBAlternate :
					if (dp.getLastAct().equals(dp.getAct1()))
						pairsToRemove.add(dp);
					break;
				case Alternate_A :
				case Choice :
				case RepeatAAlternate_A :
				case RepeatBAlternate_A :
				case RepeatBothAlternate_A :
					if (!dp.getLastAct().equals(dp.getAct1()))
						pairsToRemove.add(dp);
					break;
				default :
					break;

			}
		}
		dps.removeAll(pairsToRemove);
	}

	//	private static void updateGlobalDps(Set<DeclarePairs> globaldps, Set<DeclarePairs> dps) {
	//		Set<DeclarePairs> newdps = new HashSet<DeclarePairs>();
	//		for (DeclarePairs dp : dps) {
	//			boolean ind = false;
	//			for (DeclarePairs d : globaldps) {
	//				if (dp.getAct1().equals(d.getAct1()) && dp.getAct2().equals(d.getAct2())
	//						&& dp.getTemplate().equals(d.getTemplate())) {
	//					ind = true;
	//					d.setFreq((short) (d.getFreq() + 1));
	//
	//				}
	//
	//			}
	//			if (!ind) {
	//				newdps.add(new DeclarePairs(dp));
	//				globaldps.addAll(newdps);
	//			}
	//
	//		}
	//
	//	}

	private static void updateListCandidates(List<Set<DeclarePairsL>> aggLogH, Set<DeclarePairsL> dps,
			List<Set<String>> actCombs) {
		Set<DeclarePairsL> newdps = new HashSet<DeclarePairsL>();

		for (DeclarePairsL dp : dps) {
			int index = -1, i = 0;
			boolean ind = false, end = false;
			while (!end) {
				if (actCombs.get(i).contains(dp.getAct1()) && actCombs.get(i).contains(dp.getAct2()))
					index = i;
				i++;

				if (i >= actCombs.size() || index != -1)
					end = true;

			}
			for (DeclarePairsL d : aggLogH.get(index)) {
				if (dp.getAct1().equals(d.getAct1()) && dp.getAct2().equals(d.getAct2())
						&& dp.getTemplate().equals(d.getTemplate())) {
					ind = true;
					d.setFreq((short) (d.getFreq() + 1));
				}

			}
			if (!ind) {
				aggLogH.get(index).add(new DeclarePairsL(dp));
			}
		}
	}

	private static void updateLastAct(Set<DeclarePairsL> dps, String act) {
		for (DeclarePairsL dp : dps)
			if (dp.getAct1().equals(act) || dp.getAct2().equals(act))
				dp.setLastAct(act);

	}

	private static void keepMostStrictConstraint(Set<DeclarePairsL> dps, Set<String> events) {
		Set<DeclarePairsL> declarePairs = new HashSet<DeclarePairsL>();
		Map<Integer, String> actMap = new HashMap<Integer, String>();
		int i = 0;
		for (String e : events)
			actMap.put(i++, e);

		Set<Set<Integer>> combs = CombinationUtilsL.getAllCombs(actMap.keySet(), 2);

		for (Set<Integer> comb : combs) {
			List<String> acts = new ArrayList<String>();
			Set<DeclarePairsL> pairsWithAct = new HashSet<DeclarePairsL>();
			for (int j : comb)
				acts.add(actMap.get(j));
			for (DeclarePairsL dp : dps) {
				if (acts.contains(dp.getAct1()) && acts.contains(dp.getAct2()))
					pairsWithAct.add(dp);
			}

			if (acts.contains("c") && acts.contains("d")) {
				System.out.println("");
			}

			//keep restricting pairs
			DeclarePairsL d = null;
			int id = 99;
			for (DeclarePairsL dp : pairsWithAct) {
				if (scoreDeclareTemplates(dp.getTemplate()) <= id) {
					if (id == scoreDeclareTemplates(dp.getTemplate())) {
						d = (d.getAct1().equals(d.getLastAct()) && d.getTemplate().toString().contains("_A")) ? d : dp;
					} else {
						id = scoreDeclareTemplates(dp.getTemplate());
						d = dp;
					}
				}
			}

			//remove all other
			if (d != null) {
				dps.removeAll(pairsWithAct);
				dps.add(d);
			}
		}

	}

	//add declare pairs corresponding to the newly added activity
	private static void addDeclarePairs(String act, Set<String> actsNotAdded, Set<DeclarePairsL> dps) {
		for (String actNotAdded : actsNotAdded) {
			for (DeclareTemplatesL temp : DeclareTemplatesL.values()) {
				if (!temp.equals(DeclareTemplatesL.None)) {
					DeclarePairsL dp1 = new DeclarePairsL(act, actNotAdded);
					dp1.setTemplate(temp);
					dps.add(dp1);
				}
			}
		}

	}

	//remove pairs which are no longer valid
	private static void removeInvaliadtedPairs(String act, Set<String> addedActs, Set<DeclarePairsL> dps) {
		Set<DeclarePairsL> dpsToRemove = new HashSet<DeclarePairsL>();
		if (act.equals("b"))
			System.out.println();

		for (DeclarePairsL dp : dps) {
			if (dp.getAct1().equals(act) || dp.getAct2().equals(act)) {
				switch (dp.getTemplate()) {
				//					case None :
				//						dpsToRemove.add(dp);
				//						break;
					case Choice :
						if (dp.getAct2().equals(act))
							dpsToRemove.add(dp);
						break;

					case RepeatBothResponse :
						if (dp.getLastAct().equals(dp.getAct2()) && dp.getAct1().equals(act)) //BA not allowed
							dpsToRemove.add(dp);
						break;

					case Response :
						if (dp.getLastAct().equals(dp.getAct2()) && dp.getAct1().equals(act)
								|| (dp.getLastAct().equals(act))) //BA or BB or AA not allowed
							dpsToRemove.add(dp);
						break;

					case RepeatAResponse :
						if ((dp.getLastAct().equals(dp.getAct2()) && dp.getAct1().equals(act))
								|| (dp.getLastAct().equals(act) && dp.getAct2().equals(act)))//BA or BB  not allowed
							dpsToRemove.add(dp);
						break;
					case RepeatBResponse :
						if ((dp.getLastAct().equals(dp.getAct2()) && dp.getAct1().equals(act))
								|| (dp.getLastAct().equals(act) && dp.getAct1().equals(act)))//BA or AA  not allowed
							dpsToRemove.add(dp);
						break;

					case Alternate :
					case Alternate_A :
						if (dp.getLastAct().equals(act)) //repeats not allowed
							dpsToRemove.add(dp);
						break;

					case RepeatAAlternate :
					case RepeatAAlternate_A :
						if (dp.getLastAct().equals(act) && dp.getAct2().equals(act)) //BB allowed
							dpsToRemove.add(dp);
						break;

					case RepeatBAlternate :
					case RepeatBAlternate_A :
						if (dp.getLastAct().equals(act) && dp.getAct1().equals(act)) //AA not allowed
							dpsToRemove.add(dp);
						break;

					case RepeatBothAlternate : // last one standing
						break;
					case RepeatBothAlternate_A :
						break;
					default :
						break;

				}
			}
		}

		dps.removeAll(dpsToRemove);

	}

	//create empty declare constraint - each pair has ALL the pairs - remove the pairs one by one 
	private static Set<DeclarePairsL> getInitialDP(Set<String> events) {
		Set<DeclarePairsL> declarePairs = new HashSet<DeclarePairsL>();
		Map<Integer, String> actMap = new HashMap<Integer, String>();
		int i = 0;
		for (String e : events)
			actMap.put(i++, e);

		Set<Set<Integer>> combs = CombinationUtilsL.getAllCombs(actMap.keySet(), 2);

		for (Set<Integer> comb : combs) {
			List<String> acts = new ArrayList<String>();
			for (int j : comb)
				acts.add(actMap.get(j));

			DeclarePairsL dp1 = new DeclarePairsL(acts.get(0), acts.get(1));
			DeclarePairsL dp2 = new DeclarePairsL(acts.get(1), acts.get(0));
			for (DeclareTemplatesL temp : DeclareTemplatesL.values()) {
				dp1.setTemplate(temp);
				dp2.setTemplate(temp);
			}
			declarePairs.add(dp1);
			declarePairs.add(dp2);
		}

		return declarePairs;

	}

	private static Integer scoreDeclareTemplates(DeclareTemplatesL dt) {
		switch (dt) {
			case None :
				return 10;
			case Choice :
				return 1;
			case Response :
				return 2;
			case RepeatAResponse :
				return 3;
			case RepeatBResponse :
				return 4;
			case RepeatBothResponse :
				return 5;

			case Alternate :
			case Alternate_A :
				return 6;

			case RepeatAAlternate :
			case RepeatAAlternate_A :
				return 7;

			case RepeatBAlternate :
			case RepeatBAlternate_A :
				return 8;

			case RepeatBothAlternate :
			case RepeatBothAlternate_A :
				return 9;
			default :
				break;

		}
		return -1;
	}
}
