package org.processmining.decomposelogs;

import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Set;
import java.util.concurrent.ExecutorService;
import java.util.concurrent.Executors;
import java.util.concurrent.TimeUnit;

import org.deckfour.xes.extension.std.XConceptExtension;
import org.deckfour.xes.model.XEvent;
import org.deckfour.xes.model.XLog;
import org.deckfour.xes.model.XTrace;
import org.processmining.acceptingpetrinet.models.AcceptingPetriNet;
import org.processmining.acceptingpetrinet.models.impl.AcceptingPetriNetImpl;
import org.processmining.framework.packages.PackageManager.Canceller;
import org.processmining.framework.plugin.PluginContext;
import org.processmining.framework.util.Pair;
import org.processmining.models.graphbased.directed.petrinet.Petrinet;
import org.processmining.models.graphbased.directed.petrinet.PetrinetNode;
import org.processmining.models.graphbased.directed.petrinet.elements.Place;
import org.processmining.models.graphbased.directed.petrinet.elements.Transition;
import org.processmining.plugins.InductiveMiner.reduceacceptingpetrinet.ReduceAcceptingPetriNetKeepLanguage;
import org.processmining.prabhakardixit.fcnets.models.PetriNetMapL;
import org.processmining.reduction.ReductionL;

public class LogModelComparisonL {

	static long msecs1 = 0;
	static long msecs2 = 0;
	static long msecs3 = 0;
	static long msecs4 = 0;

	public static void compareModelAndLog(final PluginContext context, final Petrinet pn, XLog log) {
		final Set<String> acts = new HashSet<String>();
		for (XTrace t : log)
			for (XEvent e : t)
				acts.add(XConceptExtension.instance().extractName(e));
		AggregatedHeuristicsL ahl = Log2DeclareL.getAggPairHeuristics(log, acts, 0, log.size());
		compareModelAndLog(context, pn, ahl);

	}

	public static void compareModelAndLog(final PluginContext context, final Petrinet pn,
			final AggregatedHeuristicsL logAgg) {
		msecs1 = 0;
		msecs2 = 0;
		msecs3 = 0;
		msecs4 = 0;
		long msecs = System.currentTimeMillis();
		final List<Set<String>> actComb = logAgg.getActComb();
		final float[] fitness = new float[actComb.size()]; //fitness value corresponding to index from list
		final float[] precision = new float[actComb.size()];

		//get all clustered set of acts
		//		Set<String> clusterdActs = new HashSet<String>();
		//		Map<String, Set<String>> actCombClusters = findSimilarClusters(pn, clusterdActs);
		//
		//		final List<Integer> indices = new ArrayList<Integer>();
		//		Map<Set<String>, Integer> mapper = new HashMap<Set<String>, Integer>();
		//		for (int i = 0; i < actComb.size(); i++) {
		//			if (Collections.disjoint(clusterdActs, actComb.get(i))
		//					|| anyClusterContainsAllActs(actCombClusters, actComb.get(i)))
		//				indices.add(i);
		//			else
		//				mapper.put(actComb.get(i), i);
		//		}

		int NTHREADS = Math.max(1, Runtime.getRuntime().availableProcessors());
		if (actComb.size() > NTHREADS) {
			ExecutorService exec = Executors.newFixedThreadPool(NTHREADS - 1);
			final int segmentLen = (actComb.size() / (NTHREADS - 1)) < 1 ? 1 : actComb.size() / NTHREADS - 1;
			int offset = 0;
			for (int i = 0; i < NTHREADS - 1; i++) {
				final int from = offset;
				final int to = offset + segmentLen;
				exec.execute(new Runnable() {

					@Override
					public void run() {
						compareLogAndModels(context, pn, actComb, logAgg, fitness, precision, from, to);
						//					work(array, from, to);
					}
				});
				offset += segmentLen;
			}

			// do last segment on main thread
			compareLogAndModels(context, pn, actComb, logAgg, fitness, precision, segmentLen * (NTHREADS - 1),
					actComb.size());

			//process all the dependencies between clusters and other acts
			//			processMappingClusters(context, pn, mapper, actCombClusters, logAgg, fitness, precision);

			//				 wait for completion
			exec.shutdown();
			try {
				exec.awaitTermination(Long.MAX_VALUE, TimeUnit.NANOSECONDS);
			} catch (InterruptedException e) {
				// TODO Auto-generated catch block
				e.printStackTrace();
			}
		} else {
			compareLogAndModels(context, pn, actComb, logAgg, fitness, precision, 0, actComb.size() - 1);
		}
		float fit = 0, pre = 0;
		for (int i = 0; i < fitness.length; i++) {
			fit += fitness[i];
			pre += precision[i];
		}
		System.out.println(" Fitness : " + fit / fitness.length + " Precision : " + pre / precision.length);
		System.out.println("Time spent : " + (System.currentTimeMillis() - msecs));

		System.out.println("msecs1 :" + msecs1 + " msecs2 :" + msecs2 + " msecs3 :" + msecs3 + " msecs4 :" + msecs4);
		//		for (int i = 0; i < fitness.length; i++) {
		//			if (precision[i] < 1f) {
		//				System.out.println(actComb.get(i) + " : " + precision[i]);
		//			}
		//		}

		//		System.out.println(fitness + "" + precision);
		//		try {
		//			exec.awaitTermination(10, TimeUnit.SECONDS);
		//		} catch (InterruptedException ignore) {
		//		}
	}

	private static void processMappingClusters(PluginContext context, Petrinet pn, Map<Set<String>, Integer> mapper,
			Map<String, Set<String>> actCombClusters, AggregatedHeuristicsL ahl, float[] fitness, float[] precision) {

		for (String s : actCombClusters.keySet()) {
			for (Set<String> ac : mapper.keySet()) {
				if (actCombClusters.get(s).equals(ac)) {
					PetriNetMapL pn2 = new PetriNetMapL(pn);
					AcceptingPetriNet apn = new AcceptingPetriNetImpl(pn2.getPetriNet());
					reduceNet(context, apn, ac);
					Set<Place> initMarking = new HashSet<Place>();
					Set<Place> finalMarking = new HashSet<Place>();
					pupulateInitialAndFinalMarkings(initMarking, finalMarking, pn2.getPetriNet());
					Set<DeclarePairsL> dps = Net2DeclareL.net2Declare(pn2.getPetriNet(), initMarking, finalMarking);
					Pair<Float, Float> result = getFitnessAndPrecision(ahl, dps, mapper.get(ac));
					fitness[mapper.get(ac)] = result.getFirst();
					precision[mapper.get(ac)] = result.getSecond();

					for (String s2 : actCombClusters.get(s)) {
						Set<String> ac2 = new HashSet<String>(ac);
						ac2.remove(s);
						ac2.add(s);
						int id = getId(ac2, mapper);
						Set<DeclarePairsL> dps2 = new HashSet<DeclarePairsL>();
						for (DeclarePairsL dp : dps)
							dps2.add(new DeclarePairsL(dp.getAct1().equals(s) ? s2 : dp.getAct1(), dp.getAct2()
									.equals(s) ? s2 : dp.getAct2(), dp.getTemplate()));
						Pair<Float, Float> result2 = getFitnessAndPrecision(ahl, dps2, id);
						fitness[mapper.get(s)] = result2.getFirst();
						precision[mapper.get(s)] = result2.getSecond();
					}

				}
			}

		}

	}

	private static int getId(Set<String> s, Map<Set<String>, Integer> mapper) {
		for (Set<String> s2 : mapper.keySet())
			if (s2.equals(s))
				return mapper.get(s2);
		return -1;
	}

	private static boolean anyClusterContainsAllActs(Map<String, Set<String>> actCombClusters, Set<String> set) {
		for (String s : actCombClusters.keySet())
			if (actCombClusters.get(s).containsAll(set) || actCombClusters.get(s).equals(set))
				return true;
		return false;
	}

	private static Map<String, Set<String>> findSimilarClusters(Petrinet pn, Set<String> allStrs) {
		ReductionL rd = new ReductionL();
		Set<Set<String>> coms = rd.reduceNetGetClusters(new PetriNetMapL(pn).getPetriNet());
		System.out.println(coms + "");

		Map<String, Set<String>> actSetMapper = new HashMap<String, Set<String>>();
		for (Set<String> com : coms)
			if (com.size() > 1) {
				actSetMapper.put(com.iterator().next(), com);
				allStrs.addAll(com);
			}
		return actSetMapper;
	}

	//multi threading
	private static void compareLogAndModels(PluginContext context, Petrinet pn, List<Set<String>> actComb,
			AggregatedHeuristicsL logAgg, float[] fit, float[] pre, int from, int to) {
		for (int j = from; j < to; j++) {
			Set<String> ac = actComb.get(j);
			//compare the agg heu
			Pair<Float, Float> fitPre = compareLogAndModel(context, logAgg, ac, j, pn);
			fit[j] = fitPre.getFirst();
			pre[j] = fitPre.getSecond();
		}
	}

	private static Pair<Float, Float> compareLogAndModel(PluginContext context, AggregatedHeuristicsL ahl,
			Set<String> ac, int index, Petrinet pn) {
		//reduce the net to the given pair

		PetriNetMapL pn2 = new PetriNetMapL(pn);

		AcceptingPetriNet apn = new AcceptingPetriNetImpl(pn2.getPetriNet());
		long msecs = System.currentTimeMillis();
		reduceNet(context, apn, ac);
		msecs1 += System.currentTimeMillis() - msecs;
		//		reduceNet(context, pn2.getPetriNet(), ac);
		if (ac.contains("IV Liquid") && ac.contains("IV Antibiotics"))
			context.getProvidedObjectManager().createProvidedObject("Petrinet" + ac, pn2.getPetriNet(), Petrinet.class,
					context);

		//calculate agg heu for the net
		msecs = System.currentTimeMillis();
		Set<Place> initMarking = new HashSet<Place>();
		Set<Place> finalMarking = new HashSet<Place>();
		pupulateInitialAndFinalMarkings(initMarking, finalMarking, pn2.getPetriNet());
		msecs2 += System.currentTimeMillis() - msecs;
		msecs = System.currentTimeMillis();
		Set<DeclarePairsL> dps = Net2DeclareL.net2Declare(pn2.getPetriNet(), initMarking, finalMarking);
		msecs3 += System.currentTimeMillis() - msecs;
		msecs = System.currentTimeMillis();
		Pair<Float, Float> result = getFitnessAndPrecision(ahl, dps, index);
		msecs4 += System.currentTimeMillis() - msecs;
		return result;

	}

	private static void pupulateInitialAndFinalMarkings(Set<Place> initMarking, Set<Place> finalMarking, Petrinet pn) {
		for (Place p : pn.getPlaces()) {
			if (pn.getInEdges(p).isEmpty())
				initMarking.add(p);
			if (pn.getOutEdges(p).isEmpty())
				finalMarking.add(p);
		}
	}

	private static Pair<Float, Float> getFitnessAndPrecision(AggregatedHeuristicsL ahlog, Set<DeclarePairsL> dps,
			int index) {

		float fitCases = 0;
		float tot = 0;
		float preCases = 0;
		String act1 = "";

		Set<DeclareTemplatesL> allTemplatesInModel1 = new HashSet<DeclareTemplatesL>();
		Set<DeclareTemplatesL> allTemplatesInModel2 = new HashSet<DeclareTemplatesL>();

		for (DeclarePairsL dp : dps) {
			if (act1 == "")
				act1 = dp.getAct1();
			if (act1.equals(dp.getAct1()))
				allTemplatesInModel1.add(dp.getTemplate());
			else
				allTemplatesInModel2.add(dp.getTemplate());
		}

		for (DeclarePairsL dp : ahlog.getDeclarePairs().get(index)) {
			if ((dp.getAct1().equals(act1) && allTemplatesInModel1.contains(dp.getTemplate()))
					|| (dp.getAct2().equals(act1) && allTemplatesInModel2.contains(dp.getTemplate()))) {
				fitCases += dp.getFreq();
				preCases++;
			}
			//				else
			//					System.out.println(dp.getTemplate() + " Act 1 " + act1);
			tot += dp.getFreq();
		}

		Pair<Float, Float> fitPre = new Pair<Float, Float>(Float.isNaN(fitCases / tot) ? 1 : fitCases / tot,
				Float.isNaN(preCases / (allTemplatesInModel1.size() + allTemplatesInModel2.size())) ? 0 : preCases
						/ (allTemplatesInModel1.size() + allTemplatesInModel2.size()));

		//		System.out.println("Fitness Pre : " + fitPre);
		//		if (fitPre.getFirst() < 0.1)
		//			System.out.println(acts);
		return fitPre;
	}

	private static void reduceNet(PluginContext context, AcceptingPetriNet apn, Set<String> ac) {
		//		//		Murata m = new Murata();
		//		//		MurataParameters parameters = new MurataParameters();
		//		//		parameters.setAllowFPTSacredNode(false);
		//		//		try {
		//		//			m.runPreserveBehavior(context, pn, PetrinetUtils.getInitialMarking(pn));
		//		//		} catch (ConnectionCannotBeObtained e) {
		//		//			// TODO Auto-generated catch block
		//		//			e.printStackTrace();
		//		//		}
		//		//		AcceptingPetriNet apn = new AcceptingPetriNetImpl(pn);

		for (Transition t : apn.getNet().getTransitions()) {
			if (!ac.contains(t.getLabel()))
				t.setInvisible(true);
		}
		ReduceAcceptingPetriNetKeepLanguage.reduce(apn, new Canceller() {
			public boolean isCancelled() {
				return false;
			}
		});
	}

	private static void reduceNet(PluginContext context, Petrinet pn, Set<String> ac) {
		Set<PetrinetNode> transTokeep = new HashSet<PetrinetNode>();
		for (Transition t : pn.getTransitions())
			if (!t.isInvisible() && ac.contains(t.getLabel()))
				transTokeep.add(t);
		ReductionL rd = new ReductionL(pn, transTokeep);
		for (Transition t : pn.getTransitions())
			if (!ac.contains(t.getLabel()))
				t.setInvisible(true);

	}
}
