package org.processmining.decomposelogs;

import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Set;
import java.util.concurrent.ExecutorService;
import java.util.concurrent.Executors;
import java.util.concurrent.TimeUnit;

import org.deckfour.xes.extension.std.XConceptExtension;
import org.deckfour.xes.model.XEvent;
import org.deckfour.xes.model.XLog;
import org.deckfour.xes.model.XTrace;
import org.processmining.acceptingpetrinet.models.AcceptingPetriNet;
import org.processmining.acceptingpetrinet.models.impl.AcceptingPetriNetImpl;
import org.processmining.framework.packages.PackageManager.Canceller;
import org.processmining.framework.plugin.PluginContext;
import org.processmining.framework.util.Pair;
import org.processmining.models.graphbased.directed.petrinet.Petrinet;
import org.processmining.models.graphbased.directed.petrinet.elements.Place;
import org.processmining.models.graphbased.directed.petrinet.elements.Transition;
import org.processmining.models.semantics.petrinet.Marking;
import org.processmining.models.semantics.petrinet.impl.EfficientPetrinetSemanticsImpl;
import org.processmining.plugins.InductiveMiner.reduceacceptingpetrinet.ReduceAcceptingPetriNetKeepLanguage;
import org.processmining.prabhakardixit.fcnets.models.PetriNetMapL;
import org.processmining.reduction.ReductionL;
import org.processmining.reduction.ReductionNewL;

public class LogModelComparisonNewL {

	static long msecs1 = 0;
	static long msecs2 = 0;
	static long msecs3 = 0;
	static long msecs4 = 0;

	public static ConstraintsBasedFitnessPrecision compareModelAndLog(final PluginContext context, final Petrinet pn,
			XLog log, boolean multiThreaded) {
		final Set<String> acts = new HashSet<String>();
		for (XTrace t : log)
			for (XEvent e : t)
				acts.add(XConceptExtension.instance().extractName(e));
		AggregatedHeuristicsL ahl = Log2DeclareL.getAggPairHeuristics(log, acts, 0, log.size());
		//		PairwiseLogHeuristic plh = PairwiseLog2Declare.getAggPairHeuristics(log);

		float[] res = compareModelAndLog(context, pn, ahl, null, multiThreaded);
		ConstraintsBasedFitnessPrecision cbfp = new ConstraintsBasedFitnessPrecision();
		cbfp.fit1 = res[0];
		cbfp.fit2 = res[1];
		cbfp.fit3 = res[2];
		cbfp.pre1 = res[3];
		cbfp.pre2 = res[4];
		cbfp.pre3 = res[5];
		return cbfp;
	}

	public static float[] compareModelAndLog1(final PluginContext context, final Petrinet pn,
			final AggregatedHeuristicsL logAgg, XLog log) {
		float[] fitPre = new float[6];
		float res[] = new float[6];

		System.out.println("Size 123: " + Thread.getAllStackTraces().size());

		long msecs = System.currentTimeMillis();
		final List<Set<String>> actComb = logAgg.getActComb();

		Set<String> actsInModel = new HashSet<String>();
		for (Transition t : pn.getTransitions()) {
			if (!t.isInvisible())
				actsInModel.add(t.getLabel());
		}

		//		Set<Set<String>> removeCombs = new HashSet<Set<String>>();

		final Set<Integer> skipIndexes = new HashSet<Integer>();

		for (Set<String> ac : actComb)
			if (!actsInModel.containsAll(ac))
				skipIndexes.add(actComb.indexOf(ac));

		//		skipIndexes.removeAll(removeCombs);

		final float[] fitness = new float[actComb.size()]; //fitness value corresponding to index from list
		final float[] precision = new float[actComb.size()];

		System.out.println("Size 2: " + Thread.getAllStackTraces().size());
		compareLogAndModels(context, pn, actComb, logAgg, fitness, precision, 0, actComb.size(), skipIndexes);
		System.out.println("Size 3: " + Thread.getAllStackTraces().size());
		float fit = 0, pre = 0, tot2 = 0;
		for (int i = 0; i < fitness.length; i++) {
			if (!skipIndexes.contains(i)) {
				fit += fitness[i];
				pre += precision[i];
				tot2++;
			}

		}

		float fit1 = 0, tot1 = 0;
		float pre1 = 0;
		Map<String, Pair<Short, Short>> minMax = getUnaryConstraints(context, pn);


		for (String s : minMax.keySet()) {
			short fitting = 0, total = 0;
			short mmax = minMax.get(s).getSecond(), lmax = 0, lmin = 0;

			if (logAgg.actAbsOccCaseOcc.containsKey(s)) {
				for (short nbr : logAgg.actAbsOccCaseOcc.get(s).keySet()) {
					if (nbr > lmax)
						lmax = nbr;
					if (nbr >= minMax.get(s).getFirst() && nbr <= minMax.get(s).getSecond()) {
						fitting += logAgg.actAbsOccCaseOcc.get(s).get(nbr);
					}
					total += logAgg.actAbsOccCaseOcc.get(s).get(nbr);
				}
				if (total == 0)
					total++;
				fit1 += (float) fitting / (float) total;
				if (mmax > 999) //max of model
					mmax = (short) (lmax + 1);

				if (lmax > mmax)
					pre1 += 1;
				else
					pre1 += (float) lmax / (float) mmax;
				tot1++;
			}
		}

		int totInModel = 0, totInLog = 0;
		for (String s : logAgg.actAbsOccCaseOcc.keySet()) {
			for (short i : logAgg.actAbsOccCaseOcc.get(s).keySet()) {
				if (minMax.containsKey(s))
					totInModel += logAgg.actAbsOccCaseOcc.get(s).get(i);

				totInLog += logAgg.actAbsOccCaseOcc.get(s).get(i);
			}
		}

		float aggFit = fit1 / tot1;
		float aggPre = pre1 / tot1;
		float aggFit2 = fit / fitness.length;
		float aggPre2 = pre / precision.length;

		res[0] = (fit + fit1) / (tot2 + tot1);
		res[1] = aggFit;
		res[2] = aggFit2;
		res[3] = (pre + pre1) / (tot2 + tot1);
		res[4] = aggPre;
		res[5] = aggPre2;

		//calc log coverage
		fitPre[0] = (fit + fit1) / (tot2 + tot1);
		fitPre[1] = (pre + pre1) / (tot2 + tot1);
//		fitPre[2] = (float) totInModel / (float) totInLog;
		float totEvts = 0f, evtsCovered = 0f;
		if (log != null) {
			for (XTrace trace : log) {
				for (XEvent event : trace) {

					if (actsInModel.contains(XConceptExtension.instance().extractName(event).toString())) {
						evtsCovered++;
					}
					totEvts++;
				}
			}

		}
		fitPre[2] = evtsCovered / totEvts;

		return res;
	}

	public static float[] compareModelAndLog(final PluginContext context, final Petrinet pn,
			final AggregatedHeuristicsL logAgg, XLog log, boolean multiThreaded) {
		float[] fitPre = new float[6];
		float res[] = new float[6];
		msecs1 = 0;
		msecs2 = 0;
		msecs3 = 0;
		msecs4 = 0;
		//		
		//		Set<Thread> threads = Thread.getAllStackTraces().keySet();
		//		System.out.println("BEFORE comparing!");
		//		for (Thread t : threads) {
		//		    String name = t.getName();
		//		    Thread.State state = t.getState();
		//		    int priority = t.getPriority();
		//		    String type = t.isDaemon() ? "Daemon" : "Normal";
		//		    System.out.printf("%-20s \t %s \t %d \t %s\n", name, state, priority, type);
		//		}

//		System.out.println("Size : " + Thread.getAllStackTraces().size());
		if (logAgg != null) {
			long msecs = System.currentTimeMillis();
			final List<Set<String>> actComb = logAgg.getActComb();

			Set<String> actsInModel = new HashSet<String>();
			for (Transition t : pn.getTransitions()) {
				if (!t.isInvisible())
					actsInModel.add(t.getLabel());
			}

			//		Set<Set<String>> removeCombs = new HashSet<Set<String>>();

			final Set<Integer> skipIndexes = new HashSet<Integer>();

			for (Set<String> ac : actComb)
				if (!actsInModel.containsAll(ac))
					skipIndexes.add(actComb.indexOf(ac));

			//		skipIndexes.removeAll(removeCombs);

			final float[] fitness = new float[actComb.size()]; //fitness value corresponding to index from list
			final float[] precision = new float[actComb.size()];

			int NTHREADS = Math.max(1, Runtime.getRuntime().availableProcessors());
			if (!multiThreaded)
				NTHREADS = 999999; //Set very high to avoid multi threading
			if (actComb.size() > NTHREADS) {
				ExecutorService exec = Executors.newFixedThreadPool(NTHREADS - 1);
				final int segmentLen = (actComb.size() / (NTHREADS - 1)) < 1 ? 1 : actComb.size() / NTHREADS - 1;
				int offset = 0;
				for (int i = 0; i < NTHREADS - 1; i++) {
					final int from = offset;
					final int to = offset + segmentLen;
					exec.execute(new Runnable() {

						@Override
						public void run() {
							compareLogAndModels(context, pn, actComb, logAgg, fitness, precision, from, to, skipIndexes);
							//					work(array, from, to);
						}
					});
					offset += segmentLen;
				}

				// do last segment on main thread
				compareLogAndModels(context, pn, actComb, logAgg, fitness, precision, segmentLen * (NTHREADS - 1),
						actComb.size(), skipIndexes);

				//process all the dependencies between clusters and other acts
				//			processMappingClusters(context, pn, mapper, actCombClusters, logAgg, fitness, precision);

				//				 wait for completion
				exec.shutdown();
				try {
					exec.awaitTermination(Long.MAX_VALUE, TimeUnit.NANOSECONDS);
				} catch (InterruptedException e) {
					// TODO Auto-generated catch block
					e.printStackTrace();
				}
			} else {

//				System.out.println("Size 2: " + Thread.getAllStackTraces().size());
				compareLogAndModels(context, pn, actComb, logAgg, fitness, precision, 0, actComb.size(), skipIndexes);
//				System.out.println("Size 3: " + Thread.getAllStackTraces().size());
			}
			float fit = 0, pre = 0, tot2 = 0;
			for (int i = 0; i < fitness.length; i++) {
				if (!skipIndexes.contains(i)) {
					fit += fitness[i];
					pre += precision[i];
					tot2++;
				}

			}

			Map<String, Pair<Short, Short>> minMax = getUnaryConstraints(context, pn);
			float fit1 = 0, tot1 = 0;
			float pre1 = 0;

			for (String s : minMax.keySet()) {
				short fitting = 0, total = 0;
				short mmax = minMax.get(s).getSecond(), lmax = 0, lmin = 0;

				if (logAgg.actAbsOccCaseOcc.containsKey(s)) {
					for (short nbr : logAgg.actAbsOccCaseOcc.get(s).keySet()) {
						if (nbr > lmax)
							lmax = nbr;
						if (nbr >= minMax.get(s).getFirst() && nbr <= minMax.get(s).getSecond()) {
							fitting += logAgg.actAbsOccCaseOcc.get(s).get(nbr);
						}
						total += logAgg.actAbsOccCaseOcc.get(s).get(nbr);
					}
					if (total == 0)
						total++;
					fit1 += (float) fitting / (float) total;
					if (mmax > 999) //max of model
						mmax = (short) (lmax + 1);

					if (lmax > mmax)
						pre1 += 1;
					else
						pre1 += (float) lmax / (float) mmax;
					tot1++;
				}
			}

			int totInModel = 0, totInLog = 0;
			for (String s : logAgg.actAbsOccCaseOcc.keySet()) {
				for (short i : logAgg.actAbsOccCaseOcc.get(s).keySet()) {
					if (minMax.containsKey(s))
						totInModel += logAgg.actAbsOccCaseOcc.get(s).get(i);

					totInLog += logAgg.actAbsOccCaseOcc.get(s).get(i);
				}
			}

			float aggFit = fit1 / tot1;
			float aggPre = pre1 / tot1;
			float aggFit2 = fit / fitness.length;
			float aggPre2 = pre / precision.length;
			//				System.out.println(" Fitness : " + fit / fitness.length + " Precision : " + pre / precision.length
			//						+ " Aggr fit : " + aggFit + " Aggr Pre : " + aggPre);

			float fitLen = (fitness.length) / (fitness.length + tot1);
			float preLen = (precision.length) / (precision.length + tot1);

			System.out.println(" Fitness : " + (fit + fit1) / (tot2 + tot1) + " Precision : " + (pre + pre1)
					/ (tot2 + tot1) + " Aggr fit : " + aggFit + " Aggr Pre : " + aggPre + " Aggr fit2 : "
					+ ((aggFit * (1 - fitLen) + aggFit2 * fitLen) / 2) + " Aggr Pre2 : "
					+ ((aggPre * (1 - preLen) + aggPre2 * preLen) / 2));
			//		System.out.println("Time spent : " + (System.currentTimeMillis() - msecs));

			System.out
					.println("msecs1 :" + msecs1 + " msecs2 :" + msecs2 + " msecs3 :" + msecs3 + " msecs4 :" + msecs4);

			res[0] = (fit + fit1) / (tot2 + tot1);
			res[1] = aggFit;
			res[2] = aggFit2;
			res[3] = (pre + pre1) / (tot2 + tot1);
			res[4] = aggPre;
			res[5] = aggPre2;

			//		for (int i = 0; i < fitness.length; i++) {
			//			if (precision[i] < 1f) {
			//				System.out.println(actComb.get(i) + " : " + precision[i]);
			//			}
			//		}

			//		System.out.println(fitness + "" + precision);
			//		try {
			//			exec.awaitTermination(10, TimeUnit.SECONDS);
			//		} catch (InterruptedException ignore) {
			//		}

			//calc log coverage
			fitPre[0] = (fit + fit1) / (tot2 + tot1);
			fitPre[1] = (pre + pre1) / (tot2 + tot1);
			fitPre[2] = (float) totInModel / (float) totInLog;
			float totEvts = 0f, evtsCovered = 0f;
			if (log != null) {
				for (XTrace trace : log) {
					for (XEvent event : trace) {

						if (actsInModel.contains(XConceptExtension.instance().extractName(event).toString())) {
							evtsCovered++;
						}
						totEvts++;
					}
				}

			}
			fitPre[2] = evtsCovered / totEvts;
			//		if (returnFitness)
			//			return (fit + fit1) / (fitness.length + tot1);
			//		else
			//			return (pre + pre1) / (precision.length + tot1);
		} else {
			fitPre = DefaultConformance.pccConformance(log, pn);
		}
		//		return fitPre;

		//		 threads = Thread.getAllStackTraces().keySet();
		//		System.out.println("AFTER comparing!");
		//		for (Thread t : threads) {
		//		    String name = t.getName();
		//		    Thread.State state = t.getState();
		//		    int priority = t.getPriority();
		//		    String type = t.isDaemon() ? "Daemon" : "Normal";
		//		    System.out.printf("%-20s \t %s \t %d \t %s\n", name, state, priority, type);
		//		}
		return res;
	}

	private static Map<String, Pair<Short, Short>> getUnaryConstraints(PluginContext context, Petrinet pn) {
		Map<String, Pair<Short, Short>> unary = new HashMap<String, Pair<Short, Short>>();

		for (Transition t : pn.getTransitions()) {
			PetriNetMapL pnm = new PetriNetMapL(pn);
			if (!t.isInvisible() && !unary.containsKey(t.getLabel())) {
				for (Transition t2 : pnm.getPetriNet().getTransitions()) {
					if (!t2.getLabel().equals(t.getLabel()))
						t2.setInvisible(true);
				}

				short min = 0, max = 0;
				Set<String> acts = new HashSet<String>();
				acts.add(t.getLabel());

				AcceptingPetriNet apn = new AcceptingPetriNetImpl(pnm.getPetriNet());

				reduceNet(apn.getNet(), acts);
				//				if (t.getLabel().contains("IC")){
				//					System.out.println("");
//				context.getProvidedObjectManager().createProvidedObject("Petrinet", apn, AcceptingPetriNet.class,
//						context);
				//					
				//				}
				EfficientPetrinetSemanticsImpl semantics = new EfficientPetrinetSemanticsImpl(apn.getNet(),
						apn.getInitialMarking());
				//				new Marking(PetrinetUtils.getInitialMarking(apn.getNet())
				/* A -> Choice */
				Marking m1 = Net2DeclareNewL.replay(apn, semantics, apn.getInitialMarking(), new HashSet<Marking>(),
						t.getLabel());

				if (m1 != null) {
					max++;
					Marking m2 = Net2DeclareNewL.replay(apn, semantics, m1, new HashSet<Marking>(), t.getLabel());
					if (m2 != null) {
						max++;
						m1 = Net2DeclareNewL.replay(apn, semantics, m2, new HashSet<Marking>(), t.getLabel());
						if (m1 != null) {
							max++;
							m2 = Net2DeclareNewL.replay(apn, semantics, m1, new HashSet<Marking>(), t.getLabel());
							if (m2 != null) {
								max++;
								m1 = Net2DeclareNewL.replay(apn, semantics, m2, new HashSet<Marking>(), t.getLabel());
								if (m1 != null) {
									max = 1000; //TODO 
									m2 = Net2DeclareNewL.replay(apn, semantics, m1, new HashSet<Marking>(),
											t.getLabel());
								}
							}
						}
					}
				}

				//TODO check for more min
				if (!Net2DeclareNewL.isEmptyTraceAllowed(apn))
					min++;

				unary.put(t.getLabel(), new Pair<Short, Short>(min, max));
			}
		}
		return unary;
	}

	private static void processMappingClusters(PluginContext context, Petrinet pn, Map<Set<String>, Integer> mapper,
			Map<String, Set<String>> actCombClusters, AggregatedHeuristicsL ahl, float[] fitness, float[] precision) {

		for (String s : actCombClusters.keySet()) {
			for (Set<String> ac : mapper.keySet()) {
				if (actCombClusters.get(s).equals(ac)) {
					PetriNetMapL pn2 = new PetriNetMapL(pn);
					AcceptingPetriNet apn = new AcceptingPetriNetImpl(pn2.getPetriNet());
					reduceNet(context, apn, ac);
					Set<Place> initMarking = new HashSet<Place>();
					Set<Place> finalMarking = new HashSet<Place>();
					pupulateInitialAndFinalMarkings(initMarking, finalMarking, pn2.getPetriNet());
					Set<DeclarePairsL> dps = Net2DeclareL.net2Declare(pn2.getPetriNet(), initMarking, finalMarking);
					Pair<Float, Float> result = getFitnessAndPrecision(ahl, dps, mapper.get(ac));
					fitness[mapper.get(ac)] = result.getFirst();
					precision[mapper.get(ac)] = result.getSecond();

					for (String s2 : actCombClusters.get(s)) {
						Set<String> ac2 = new HashSet<String>(ac);
						ac2.remove(s);
						ac2.add(s);
						int id = getId(ac2, mapper);
						Set<DeclarePairsL> dps2 = new HashSet<DeclarePairsL>();
						for (DeclarePairsL dp : dps)
							dps2.add(new DeclarePairsL(dp.getAct1().equals(s) ? s2 : dp.getAct1(), dp.getAct2().equals(
									s) ? s2 : dp.getAct2(), dp.getTemplate()));
						Pair<Float, Float> result2 = getFitnessAndPrecision(ahl, dps2, id);
						fitness[mapper.get(s)] = result2.getFirst();
						precision[mapper.get(s)] = result2.getSecond();
					}

				}
			}

		}

	}

	private static int getId(Set<String> s, Map<Set<String>, Integer> mapper) {
		for (Set<String> s2 : mapper.keySet())
			if (s2.equals(s))
				return mapper.get(s2);
		return -1;
	}

	private static boolean anyClusterContainsAllActs(Map<String, Set<String>> actCombClusters, Set<String> set) {
		for (String s : actCombClusters.keySet())
			if (actCombClusters.get(s).containsAll(set) || actCombClusters.get(s).equals(set))
				return true;
		return false;
	}

	private static Map<String, Set<String>> findSimilarClusters(Petrinet pn, Set<String> allStrs) {
		ReductionL rd = new ReductionL();
		Set<Set<String>> coms = rd.reduceNetGetClusters(new PetriNetMapL(pn).getPetriNet());
		System.out.println(coms + "");

		Map<String, Set<String>> actSetMapper = new HashMap<String, Set<String>>();
		for (Set<String> com : coms)
			if (com.size() > 1) {
				actSetMapper.put(com.iterator().next(), com);
				allStrs.addAll(com);
			}
		return actSetMapper;
	}

	//multi threading
	private static void compareLogAndModels(PluginContext context, Petrinet pn, List<Set<String>> actComb,
			AggregatedHeuristicsL logAgg, float[] fit, float[] pre, int from, int to, Set<Integer> skipIndexes) {
		for (int j = from; j < to; j++) {
			if (skipIndexes.contains(j))
				continue;
			Set<String> ac = actComb.get(j);
			//			if (ac.contains("Vital signs") && ac.contains("Lab tests")) {
			//				System.out.println();
			//			}
			//compare the agg heu
			Pair<Float, Float> fitPre = compareLogAndModel(context, logAgg, ac, j, pn);
			fit[j] = fitPre.getFirst();
			pre[j] = fitPre.getSecond();
			//			if (fitPre.getFirst() < 1)
			//				System.out.println("2");
		}
	}

	private static Pair<Float, Float> compareLogAndModel(PluginContext context, AggregatedHeuristicsL ahl,
			Set<String> ac, int index, Petrinet pn) {
		//reduce the net to the given pair
		Thread currThread = Thread.currentThread();

//		System.out.println("current thread = " + currThread);
		PetriNetMapL pn2 = new PetriNetMapL(pn);

		AcceptingPetriNet apn = new AcceptingPetriNetImpl(pn2.getPetriNet());
		long msecs = System.currentTimeMillis();

		//				reduceNet(context, apn, ac);
		reduceNet(pn2.getPetriNet(), ac);
		msecs1 += System.currentTimeMillis() - msecs;
		//		reduceNet(context, pn2.getPetriNet(), ac);

		//calculate agg heu for the net
		msecs = System.currentTimeMillis();
		Set<Place> initMarking = new HashSet<Place>();
		Set<Place> finalMarking = new HashSet<Place>();
		//		pupulateInitialAndFinalMarkings(initMarking, finalMarking, pn2.getPetriNet());
		msecs2 += System.currentTimeMillis() - msecs;
		msecs = System.currentTimeMillis();
		//		Set<DeclarePairs> dps = Net2DeclareAuto.net2Declare(context, apn);
		Set<DeclarePairsL> dps = Net2DeclareNewL.net2Declare(context, apn);
		msecs3 += System.currentTimeMillis() - msecs;
		msecs = System.currentTimeMillis();
		Pair<Float, Float> result = getFitnessAndPrecision(ahl, dps, index);
		msecs4 += System.currentTimeMillis() - msecs;
		//		if (result.getFirst() <= 1f) {
		//			context.getProvidedObjectManager().createProvidedObject("Petrinet" + ac + " " + dps.size(),
		//					pn2.getPetriNet(), Petrinet.class, context);
		//			context.getProvidedObjectManager().createProvidedObject("Petrinet" + ac, apn, AcceptingPetriNet.class,
		//					context);
		//		}
		return result;

	}

	private static void pupulateInitialAndFinalMarkings(Set<Place> initMarking, Set<Place> finalMarking, Petrinet pn) {
		for (Place p : pn.getPlaces()) {
			if (pn.getInEdges(p).isEmpty())
				initMarking.add(p);
			if (pn.getOutEdges(p).isEmpty())
				finalMarking.add(p);
		}
	}

	private static Pair<Float, Float> getFitnessAndPrecision(AggregatedHeuristicsL ahlog, Set<DeclarePairsL> dps,
			int index) {

		float fitCases = 0;
		float tot = 0;
		float preCases = 0;
		String act1 = "";
		String act2 = "";
		Set<DeclareTemplatesL> allTemplatesInModel1 = new HashSet<DeclareTemplatesL>();
		Set<DeclareTemplatesL> allTemplatesInModel2 = new HashSet<DeclareTemplatesL>();

		for (DeclarePairsL dp : dps) {
			if (act1 == "") {
				act1 = dp.getAct1();
				act2 = dp.getAct2();
			}
			if (act1.equals(dp.getAct1()))
				allTemplatesInModel1.add(dp.getTemplate());
			else
				allTemplatesInModel2.add(dp.getTemplate());
		}

		for (DeclarePairsL dp : ahlog.getDeclarePairs().get(index)) {
			if ((dp.getAct1().equals(act1) && allTemplatesInModel1.contains(dp.getTemplate()))
					|| (dp.getAct2().equals(act1) && allTemplatesInModel2.contains(dp.getTemplate()))) {
				fitCases += dp.getFreq();
				preCases++;
			}
			//			else
			//				System.out.println(dp.getTemplate() + " Act 1 " + act1 + " Act 2 : " + act2);
			tot += dp.getFreq();
		}

		Pair<Float, Float> fitPre = new Pair<Float, Float>(Float.isNaN(fitCases / tot) ? 1 : fitCases / tot,
				Float.isNaN(preCases / (allTemplatesInModel1.size() + allTemplatesInModel2.size())) ? 0 : preCases
						/ (allTemplatesInModel1.size() + allTemplatesInModel2.size()));

		//		System.out.println("Fitness Pre : " + fitPre);
		if (fitPre.getFirst() < 1.)
			System.out.println("");
		return fitPre;
	}

	private static void reduceNet(PluginContext context, AcceptingPetriNet apn, Set<String> ac) {
		//		//		Murata m = new Murata();
		//		//		MurataParameters parameters = new MurataParameters();
		//		//		parameters.setAllowFPTSacredNode(false);
		//		//		try {
		//		//			m.runPreserveBehavior(context, pn, PetrinetUtils.getInitialMarking(pn));
		//		//		} catch (ConnectionCannotBeObtained e) {
		//		//			// TODO Auto-generated catch block
		//		//			e.printStackTrace();
		//		//		}
		//		//		AcceptingPetriNet apn = new AcceptingPetriNetImpl(pn);

		for (Transition t : apn.getNet().getTransitions()) {
			if (!ac.contains(t.getLabel()))
				t.setInvisible(true);
		}
		ReduceAcceptingPetriNetKeepLanguage.reduce(apn, new Canceller() {
			public boolean isCancelled() {
				return false;
			}
		});
	}

	private static void reduceNet(Petrinet pn, Set<String> ac) {
		Set<Transition> transTokeep = new HashSet<Transition>();
		for (Transition t : pn.getTransitions())
			if (!t.isInvisible() && ac.contains(t.getLabel()))
				transTokeep.add(t);
			else
				t.setInvisible(true);
		ReductionNewL rd = new ReductionNewL(pn, transTokeep);
	}
}
