package org.processmining.decomposelogs;

import java.util.Collection;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Map;
import java.util.Set;

import org.deckfour.xes.factory.XFactory;
import org.deckfour.xes.factory.XFactoryRegistry;
import org.deckfour.xes.model.XAttribute;
import org.deckfour.xes.model.XEvent;
import org.deckfour.xes.model.XLog;
import org.deckfour.xes.model.XTrace;
import org.deckfour.xes.model.impl.XEventImpl;
import org.ejml.simple.SimpleMatrix;
import org.processmining.acceptingpetrinet.models.AcceptingPetriNet;
import org.processmining.framework.plugin.PluginContext;
import org.processmining.framework.plugin.ProMCanceller;
import org.processmining.models.graphbased.AttributeMap;
import org.processmining.models.graphbased.directed.petrinet.Petrinet;
import org.processmining.models.graphbased.directed.petrinet.PetrinetEdge;
import org.processmining.models.graphbased.directed.petrinet.PetrinetNode;
import org.processmining.models.graphbased.directed.petrinet.elements.Place;
import org.processmining.models.graphbased.directed.petrinet.elements.Transition;
import org.processmining.models.semantics.petrinet.Marking;
import org.processmining.models.semantics.petrinet.impl.EfficientPetrinetSemanticsImpl;
import org.processmining.plugins.petrinet.replayresult.PNRepResult;
import org.processmining.plugins.replayer.replayresult.SyncReplayResult;
import org.processmining.prabhakardixit.fcnets.matrix.MatrixUtilsL;
import org.processmining.prabhakardixit.utils.PetrinetUtilsL;
import org.processmining.projectedrecallandprecision.helperclasses.AcceptingPetriNet2automaton;
import org.processmining.projectedrecallandprecision.helperclasses.AutomatonFailedException;
import org.processmining.projectedrecallandprecision.helperclasses.automata.DFA;
import org.processmining.projectedrecallandprecision.helperclasses.automata.RunAutomaton2;

public class Net2DeclareAuto {

	public static Set<DeclarePairsL> net2Declare(PluginContext context, AcceptingPetriNet pn) {
		Set<DeclarePairsL> dps = new HashSet<DeclarePairsL>();
		String a1 = "", a2 = "";

		for (Transition t : pn.getNet().getTransitions())
			if (!t.isInvisible()) {
				if (a1 == "")
					a1 = t.getLabel();
				else if (a2 == "")
					a2 = t.getLabel();
			}
		dps.addAll(getDeclareTemplatesUsingReplay(a1, a2, pn));
		//		dps.addAll(getDeclareTemplates(a2, a1, pn));
		//		getStats(context, pn.getNet(), a1, a2);
		return dps;
	}

	private static Set<DeclarePairsL> validateUsingMatrix(Petrinet pn, String A, String B) {
		boolean flag = true;

		Set<Place> initMarking = new HashSet<Place>();
		Set<Place> finalMarking = new HashSet<Place>();
		Map<Transition, Integer> tranInciMap = new HashMap<Transition, Integer>();
		Map<Place, Integer> placeInciMap = new HashMap<Place, Integer>();
		SimpleMatrix im = MatrixUtilsL.createInciMap(tranInciMap, placeInciMap, pn);
		initMarking.clear();
		finalMarking.clear();
		for (Place p : pn.getPlaces()) {
			if (pn.getInEdges(p).isEmpty())
				initMarking.add(p);
		}

		/* A -> Choice */
		//set the output as final marking - if act is last then set the final state as final marking
		for (Transition t : pn.getTransitions())
			if (!t.isInvisible() && t.getLabel().equals(A))
				for (PetrinetEdge<? extends PetrinetNode, ? extends PetrinetNode> eOut : pn.getOutEdges(t))
					finalMarking.add((Place) eOut.getTarget());
		//M' = M + im . seq
		SimpleMatrix Mi = MatrixUtilsL.createInitialAnOrFinalMarking(placeInciMap, pn, initMarking);
		SimpleMatrix Mf = MatrixUtilsL.createInitialAnOrFinalMarking(placeInciMap, pn, finalMarking);
		SimpleMatrix t_MfMinusMi = Mf.minus(Mi).transpose();
		SimpleMatrix t_Inci = im.transpose();
		Map<Integer, Integer> coeffs = getCoeff(A, pn, tranInciMap);

		if (flag) {
			//is end marking reachable?

			// AB -> Response
			// ABB -> RepeatBReponse
			// ABBA -> RepeatBAlternate_A
			// ABBAB -> RepeatBAlternate
			// ABA -> Alternate_A
			//ABAB -> Alternate			
		}

		// AAB -> RepeatAResponse
		// AABA
		//AABAB -> RepeatAAlternate
		//AABABA -> RepeatAAlternate_A 
		// AABB -> RepeatBothResponse
		//AABBAB -> RepeatBothAlternate
		//AABBABA -> RepeatBothAlternate_A

		for (int i = 0; i < 50; i++) {
			coeffs.clear();

			System.out.println(MatrixUtilsL.canTraceBeReplayed(t_Inci, t_MfMinusMi, coeffs));
		}

		return null;
	}

	//sets the coeff corresponding to non visibile trans not labeled as act to zero
	private static Map<Integer, Integer> getCoeff(String act, Petrinet pn, Map<Transition, Integer> tranInciMap) {
		Map<Integer, Integer> coeffs = new HashMap<Integer, Integer>();
		for (Transition t : pn.getTransitions()) {
			if (!t.isInvisible())
				if (t.getLabel().equals(act))
					coeffs.put(tranInciMap.get(t), 1);
				else
					coeffs.put(tranInciMap.get(t), 0);
		}
		return coeffs;
	}

	//use PCC to get fitting and non fitting traces
	private static Set<DeclarePairsL> getDeclareTemplates(String A, String B, AcceptingPetriNet pn) {
		Set<DeclarePairsL> dps = new HashSet<DeclarePairsL>();
		for (Transition t : pn.getNet().getTransitions())
			if (!t.isInvisible())
				if (t.getLabel().equals(A)) {
					t.getAttributeMap().put(AttributeMap.LABEL, (char) 0 + "");
				} else {
					t.getAttributeMap().put(AttributeMap.LABEL, (char) 1 + "");
				}

		ProMCanceller c = new ProMCanceller() {
			public boolean isCancelled() {
				return false;
			}
		};
		DFA result = null;
		try {
			result = AcceptingPetriNet2automaton.convert(pn, 250000l, c);
		} catch (AutomatonFailedException e) {
			// TODO Auto-generated catch block
			e.printStackTrace();
		}

		RunAutomaton2 runAutomatonModel = new RunAutomaton2(result, c);

		/* A -> Choice */
		short t[] = new short[] { 0 };
		if (runAutomatonModel.accepts(new short[] { 0 }))
			dps.add(new DeclarePairsL(A, B, DeclareTemplatesL.Choice));
		// AB -> Response
		if (runAutomatonModel.accepts(new short[] { 0, 1 }))
			dps.add(new DeclarePairsL(A, B, DeclareTemplatesL.Response));
		// ABB -> RepeatBReponse
		if (runAutomatonModel.accepts(new short[] { 0, 1, 1 }))
			dps.add(new DeclarePairsL(A, B, DeclareTemplatesL.RepeatBResponse));
		// ABBA -> RepeatBAlternate_A
		if (runAutomatonModel.accepts(new short[] { 0, 1, 1, 0 }))
			dps.add(new DeclarePairsL(A, B, DeclareTemplatesL.RepeatBAlternate_A));
		// ABBAB -> RepeatBAlternate
		if (runAutomatonModel.accepts(new short[] { 0, 1, 1, 0, 1 }))
			dps.add(new DeclarePairsL(A, B, DeclareTemplatesL.RepeatBAlternate));

		// ABA -> Alternate_A
		if (runAutomatonModel.accepts(new short[] { 0, 1, 0 }))
			dps.add(new DeclarePairsL(A, B, DeclareTemplatesL.Alternate_A));
		//ABAB -> Alternate		
		if (runAutomatonModel.accepts(new short[] { 0, 1, 0, 1 }))
			dps.add(new DeclarePairsL(A, B, DeclareTemplatesL.Alternate));
		// AAB -> RepeatAResponse
		if (runAutomatonModel.accepts(new short[] { 0, 0, 1 }))
			dps.add(new DeclarePairsL(A, B, DeclareTemplatesL.RepeatAResponse));
		// AABA
		if (runAutomatonModel.accepts(new short[] { 0, 0, 1, 0 }))
			dps.add(new DeclarePairsL(A, B, DeclareTemplatesL.RepeatAAlternate));
		//AABABA -> RepeatAAlternate_A 
		if (runAutomatonModel.accepts(new short[] { 0, 0, 1, 0, 1, 0 }))
			dps.add(new DeclarePairsL(A, B, DeclareTemplatesL.RepeatAAlternate_A));
		// AABB -> RepeatBothResponse
		if (runAutomatonModel.accepts(new short[] { 0, 0, 1, 1 }))
			dps.add(new DeclarePairsL(A, B, DeclareTemplatesL.RepeatBothResponse));
		//AABBAB -> RepeatBothAlternate
		if (runAutomatonModel.accepts(new short[] { 0, 0, 1, 1, 0, 1 }))
			dps.add(new DeclarePairsL(A, B, DeclareTemplatesL.RepeatBothAlternate));

		//AABBABA -> RepeatBothAlternate_A
		if (runAutomatonModel.accepts(new short[] { 0, 0, 1, 1, 0, 1, 0 }))
			dps.add(new DeclarePairsL(A, B, DeclareTemplatesL.RepeatBothAlternate_A));

		//other way
		/* B -> Choice */
		if (runAutomatonModel.accepts(new short[] { 1 }))
			dps.add(new DeclarePairsL(B, A, DeclareTemplatesL.Choice));
		// BA -> Response
		if (runAutomatonModel.accepts(new short[] { 1, 0 }))
			dps.add(new DeclarePairsL(B, A, DeclareTemplatesL.Response));
		// BAA -> RepeatBReponse
		if (runAutomatonModel.accepts(new short[] { 1, 0, 0 }))
			dps.add(new DeclarePairsL(B, A, DeclareTemplatesL.RepeatBResponse));
		// BAAB -> RepeatBAlternate_A
		if (runAutomatonModel.accepts(new short[] { 1, 0, 0, 1 }))
			dps.add(new DeclarePairsL(B, A, DeclareTemplatesL.RepeatBAlternate_A));
		// BAABA -> RepeatBAlternate
		if (runAutomatonModel.accepts(new short[] { 1, 0, 0, 1, 0 }))
			dps.add(new DeclarePairsL(B, A, DeclareTemplatesL.RepeatBAlternate));

		// BAB -> Alternate_A
		if (runAutomatonModel.accepts(new short[] { 1, 0, 1 }))
			dps.add(new DeclarePairsL(B, A, DeclareTemplatesL.Alternate_A));
		//BABA -> Alternate		
		if (runAutomatonModel.accepts(new short[] { 1, 0, 1, 0 }))
			dps.add(new DeclarePairsL(B, A, DeclareTemplatesL.Alternate));
		// BBA -> RepeatAResponse
		if (runAutomatonModel.accepts(new short[] { 1, 1, 0 }))
			dps.add(new DeclarePairsL(B, A, DeclareTemplatesL.RepeatAResponse));
		// BBAB
		if (runAutomatonModel.accepts(new short[] { 1, 1, 0, 1 }))
			dps.add(new DeclarePairsL(B, A, DeclareTemplatesL.RepeatAAlternate));
		//BBABAB -> RepeatAAlternate_A 
		if (runAutomatonModel.accepts(new short[] { 1, 1, 0, 1, 0, 1 }))
			dps.add(new DeclarePairsL(B, A, DeclareTemplatesL.RepeatAAlternate_A));
		// BBAA -> RepeatBothResponse
		if (runAutomatonModel.accepts(new short[] { 1, 1, 0, 0 }))
			dps.add(new DeclarePairsL(B, A, DeclareTemplatesL.RepeatBothResponse));
		//BBAABA -> RepeatBothAlternate
		if (runAutomatonModel.accepts(new short[] { 1, 1, 0, 0, 1, 0 }))
			dps.add(new DeclarePairsL(B, A, DeclareTemplatesL.RepeatBothAlternate));

		//BBAABAB -> RepeatBothAlternate_A
		if (runAutomatonModel.accepts(new short[] { 1, 1, 0, 0, 1, 0, 1 }))
			dps.add(new DeclarePairsL(B, A, DeclareTemplatesL.RepeatBothAlternate_A));

		return dps;

	}

	private static Set<DeclarePairsL> getDeclareTemplatesUsingReplay(String A, String B, AcceptingPetriNet pn) {
		System.out.println(A + B);

		Set<DeclarePairsL> dps = new HashSet<DeclarePairsL>();
		for (Transition t : pn.getNet().getTransitions())
			if (!t.isInvisible())
				if (t.getLabel().equals(A)) {
					t.getAttributeMap().put(AttributeMap.LABEL, 0 + "");
				} else {
					t.getAttributeMap().put(AttributeMap.LABEL, 1 + "");
				}
		Marking initialMarking = new Marking();
		initialMarking.add(PetrinetUtilsL.getInitialPlace(pn.getNet()));
		Marking finalMarking = new Marking();
		finalMarking.add(PetrinetUtilsL.getFinalPlace(pn.getNet()));
		EfficientPetrinetSemanticsImpl semantics = new EfficientPetrinetSemanticsImpl(pn.getNet(), initialMarking);

		/* A -> Choice */
		short t[] = new short[] { 0 };
		if (replay(finalMarking, semantics, initialMarking, new HashSet<byte[]>(), (short) 0, new short[] { 0 }))
			dps.add(new DeclarePairsL(A, B, DeclareTemplatesL.Choice));
		// AB -> Response
		if (replay(finalMarking, semantics, initialMarking, new HashSet<byte[]>(), (short) 0, new short[] { 0, 1 }))
			dps.add(new DeclarePairsL(A, B, DeclareTemplatesL.Response));
		// ABB -> RepeatBReponse
		if (replay(finalMarking, semantics, initialMarking, new HashSet<byte[]>(), (short) 0, new short[] { 0, 1, 1 }))
			dps.add(new DeclarePairsL(A, B, DeclareTemplatesL.RepeatBResponse));
		// ABBA -> RepeatBAlternate_A
		if (replay(finalMarking, semantics, initialMarking, new HashSet<byte[]>(), (short) 0,
				new short[] { 0, 1, 1, 0 }))
			dps.add(new DeclarePairsL(A, B, DeclareTemplatesL.RepeatBAlternate_A));
		// ABBAB -> RepeatBAlternate
		if (replay(finalMarking, semantics, initialMarking, new HashSet<byte[]>(), (short) 0, new short[] { 0, 1, 1, 0,
				1 }))
			dps.add(new DeclarePairsL(A, B, DeclareTemplatesL.RepeatBAlternate));

		// ABA -> Alternate_A
		if (replay(finalMarking, semantics, initialMarking, new HashSet<byte[]>(), (short) 0, new short[] { 0, 1, 0 }))
			dps.add(new DeclarePairsL(A, B, DeclareTemplatesL.Alternate_A));
		//ABAB -> Alternate		
		if (replay(finalMarking, semantics, initialMarking, new HashSet<byte[]>(), (short) 0,
				new short[] { 0, 1, 0, 1 }))
			dps.add(new DeclarePairsL(A, B, DeclareTemplatesL.Alternate));
		// AAB -> RepeatAResponse
		if (replay(finalMarking, semantics, initialMarking, new HashSet<byte[]>(), (short) 0, new short[] { 0, 0, 1 }))
			dps.add(new DeclarePairsL(A, B, DeclareTemplatesL.RepeatAResponse));
		// AABA
		if (replay(finalMarking, semantics, initialMarking, new HashSet<byte[]>(), (short) 0,
				new short[] { 0, 0, 1, 0 }))
			dps.add(new DeclarePairsL(A, B, DeclareTemplatesL.RepeatAAlternate));
		//AABABA -> RepeatAAlternate_A 
		if (replay(finalMarking, semantics, initialMarking, new HashSet<byte[]>(), (short) 0, new short[] { 0, 0, 1, 0,
				1, 0 }))
			dps.add(new DeclarePairsL(A, B, DeclareTemplatesL.RepeatAAlternate_A));
		// AABB -> RepeatBothResponse
		if (replay(finalMarking, semantics, initialMarking, new HashSet<byte[]>(), (short) 0,
				new short[] { 0, 0, 1, 1 }))
			dps.add(new DeclarePairsL(A, B, DeclareTemplatesL.RepeatBothResponse));
		//AABBAB -> RepeatBothAlternate
		if (replay(finalMarking, semantics, initialMarking, new HashSet<byte[]>(), (short) 0, new short[] { 0, 0, 1, 1,
				0, 1 }))
			dps.add(new DeclarePairsL(A, B, DeclareTemplatesL.RepeatBothAlternate));

		//AABBABA -> RepeatBothAlternate_A
		if (replay(finalMarking, semantics, initialMarking, new HashSet<byte[]>(), (short) 0, new short[] { 0, 0, 1, 1,
				0, 1, 0 }))
			dps.add(new DeclarePairsL(A, B, DeclareTemplatesL.RepeatBothAlternate_A));

		//other way
		/* B -> Choice */
		if (replay(finalMarking, semantics, initialMarking, new HashSet<byte[]>(), (short) 0, new short[] { 1 }))
			dps.add(new DeclarePairsL(B, A, DeclareTemplatesL.Choice));
		// BA -> Response
		if (replay(finalMarking, semantics, initialMarking, new HashSet<byte[]>(), (short) 0, new short[] { 1, 0 }))
			dps.add(new DeclarePairsL(B, A, DeclareTemplatesL.Response));
		// BAA -> RepeatBReponse
		if (replay(finalMarking, semantics, initialMarking, new HashSet<byte[]>(), (short) 0, new short[] { 1, 0, 0 }))
			dps.add(new DeclarePairsL(B, A, DeclareTemplatesL.RepeatBResponse));
		// BAAB -> RepeatBAlternate_A
		if (replay(finalMarking, semantics, initialMarking, new HashSet<byte[]>(), (short) 0,
				new short[] { 1, 0, 0, 1 }))
			dps.add(new DeclarePairsL(B, A, DeclareTemplatesL.RepeatBAlternate_A));
		// BAABA -> RepeatBAlternate
		if (replay(finalMarking, semantics, initialMarking, new HashSet<byte[]>(), (short) 0, new short[] { 1, 0, 0, 1,
				0 }))
			dps.add(new DeclarePairsL(B, A, DeclareTemplatesL.RepeatBAlternate));

		// BAB -> Alternate_A
		if (replay(finalMarking, semantics, initialMarking, new HashSet<byte[]>(), (short) 0, new short[] { 1, 0, 1 }))
			dps.add(new DeclarePairsL(B, A, DeclareTemplatesL.Alternate_A));
		//BABA -> Alternate		
		if (replay(finalMarking, semantics, initialMarking, new HashSet<byte[]>(), (short) 0,
				new short[] { 1, 0, 1, 0 }))
			dps.add(new DeclarePairsL(B, A, DeclareTemplatesL.Alternate));
		// BBA -> RepeatAResponse
		if (replay(finalMarking, semantics, initialMarking, new HashSet<byte[]>(), (short) 0, new short[] { 1, 1, 0 }))
			dps.add(new DeclarePairsL(B, A, DeclareTemplatesL.RepeatAResponse));
		// BBAB
		if (replay(finalMarking, semantics, initialMarking, new HashSet<byte[]>(), (short) 0,
				new short[] { 1, 1, 0, 1 }))
			dps.add(new DeclarePairsL(B, A, DeclareTemplatesL.RepeatAAlternate));
		//BBABAB -> RepeatAAlternate_A 
		if (replay(finalMarking, semantics, initialMarking, new HashSet<byte[]>(), (short) 0, new short[] { 1, 1, 0, 1,
				0, 1 }))
			dps.add(new DeclarePairsL(B, A, DeclareTemplatesL.RepeatAAlternate_A));
		// BBAA -> RepeatBothResponse
		if (replay(finalMarking, semantics, initialMarking, new HashSet<byte[]>(), (short) 0,
				new short[] { 1, 1, 0, 0 }))
			dps.add(new DeclarePairsL(B, A, DeclareTemplatesL.RepeatBothResponse));
		//BBAABA -> RepeatBothAlternate
		if (replay(finalMarking, semantics, initialMarking, new HashSet<byte[]>(), (short) 0, new short[] { 1, 1, 0, 0,
				1, 0 }))
			dps.add(new DeclarePairsL(B, A, DeclareTemplatesL.RepeatBothAlternate));

		//BBAABAB -> RepeatBothAlternate_A
		if (replay(finalMarking, semantics, initialMarking, new HashSet<byte[]>(), (short) 0, new short[] { 1, 1, 0, 0,
				1, 0, 1 }))
			dps.add(new DeclarePairsL(B, A, DeclareTemplatesL.RepeatBothAlternate_A));

		return dps;

	}

	//assuming non-duplicates - else TODO do full recursion -i.e. without checks of first loop
	private static boolean replay(Marking finalMarking, EfficientPetrinetSemanticsImpl semantics, Marking marking,
			Set<byte[]> execMarkings, short ind, short[] trace) {
		System.out.println();
		for (short s : trace)
			System.out.print(s);
		Marking markingReached = null;
		boolean flag = false;
		//initiate only once
		if (marking != null)
			semantics.setStateAsMarking(marking);

		if (!execMarkings.add(semantics.getState()))
			return false;
		if (finalMarking.equals(semantics.getCurrentState()))
			if (ind == trace.length)
				return true;
			else
				return false;
		Collection<Transition> enabledTrans = semantics.getExecutableTransitions();
		Marking currMarking = semantics.getCurrentState();
		if (ind < trace.length) {
			short ind2 = -1;
			for (Transition t : enabledTrans)
				if (!t.isInvisible() && ind < trace.length && t.getLabel().equals(String.valueOf(trace[ind]))) {
					ind2 = ++ind;
					if (flag)
						return flag;
					semantics.setCurrentState(currMarking);
					semantics.directExecuteExecutableTransition(t);
					markingReached = semantics.getStateAsMarking();
					execMarkings.clear();
					ind = ind2;
					flag = replay(finalMarking, semantics, null, execMarkings, ind2, trace);

				}
			if (ind2 == -1)
				//only tau's remaining?
				for (Transition t2 : enabledTrans) {
					if (t2.isInvisible()) {
						if (flag)
							return flag;
						semantics.setCurrentState(currMarking);
						semantics.directExecuteExecutableTransition(t2);
						flag = replay(finalMarking, semantics, null, execMarkings, ind, trace);
					}
				}
		}
		return flag;
	}

	public static Set<DeclarePairsL> getStats(PluginContext context, Petrinet pn, String A, String B) {
		Set<DeclarePairsL> dps = new HashSet<DeclarePairsL>();
		for (Transition t : pn.getTransitions())
			if (!t.isInvisible())
				if (t.getLabel().equals(A)) {
					t.getAttributeMap().put(AttributeMap.LABEL, 0 + "");
				} else {
					t.getAttributeMap().put(AttributeMap.LABEL, 1 + "");
				}

		XLog log = getLog();
		PNRepResult result = AlignmentCalculationsL.calculateNAlignment1(context, log, pn, 8, false, null);

		Set<DeclarePairsL> dps2 = new HashSet<DeclarePairsL>();

		for (SyncReplayResult sync : result) {
			if (sync.getInfo().get(PNRepResult.TRACEFITNESS) == 1.) {
				for (int i : sync.getTraceIndex()) {
					String a1, a2;
					if (i > 12) {
						a1 = B;
						a2 = A;
					} else {
						a1 = A;
						a2 = B;
					}
					switch (i) {
						case 0 :
						case 13 :
							dps2.add(new DeclarePairsL(a1, a2, DeclareTemplatesL.Choice));
							break;
						case 1 :
						case 14 :
							dps2.add(new DeclarePairsL(a1, a2, DeclareTemplatesL.Response));
							break;
						case 2 :
						case 15 :
							dps2.add(new DeclarePairsL(a1, a2, DeclareTemplatesL.Alternate));
							break;
						case 3 :
						case 16 :
							dps2.add(new DeclarePairsL(a1, a2, DeclareTemplatesL.Alternate_A));
							break;
						case 4 :
						case 17 :
							dps2.add(new DeclarePairsL(a1, a2, DeclareTemplatesL.RepeatAResponse));
							break;
						case 5 :
						case 18 :
							dps2.add(new DeclarePairsL(a1, a2, DeclareTemplatesL.RepeatBResponse));
							break;
						case 6 :
						case 19 :
							dps2.add(new DeclarePairsL(a1, a2, DeclareTemplatesL.RepeatAAlternate));
							break;
						case 7 :
						case 20 :
							dps2.add(new DeclarePairsL(a1, a2, DeclareTemplatesL.RepeatAAlternate_A));
							break;
						case 8 :
						case 21 :
							dps2.add(new DeclarePairsL(a1, a2, DeclareTemplatesL.RepeatBAlternate));
							break;
						case 9 :
						case 22 :
							dps2.add(new DeclarePairsL(a1, a2, DeclareTemplatesL.RepeatBAlternate_A));
							break;
						case 10 :
						case 23 :
							dps2.add(new DeclarePairsL(a1, a2, DeclareTemplatesL.RepeatBothResponse));
							break;
						case 11 :
						case 24 :
							dps2.add(new DeclarePairsL(a1, a2, DeclareTemplatesL.RepeatBothAlternate));
							break;
						case 12 :
						case 25 :
							dps2.add(new DeclarePairsL(a1, a2, DeclareTemplatesL.RepeatBothAlternate_A));
							break;

					}
				}
			}
		}

		return dps2;
	}

	public static XLog getLog() {
		XFactory f = XFactoryRegistry.instance().currentDefault();
		XLog log = f.createLog();
		XTrace t = f.createTrace();
		XEvent e = new XEventImpl();
		log.add(addTrace(new short[] { 0 }));
		log.add(addTrace(new short[] { 0, 1 }));
		log.add(addTrace(new short[] { 0, 1, 0, 1 }));
		log.add(addTrace(new short[] { 0, 1, 0 }));
		log.add(addTrace(new short[] { 0, 0, 1 }));
		log.add(addTrace(new short[] { 0, 1, 1 }));
		log.add(addTrace(new short[] { 0, 0, 1, 0, 1 }));
		log.add(addTrace(new short[] { 0, 0, 1, 0 }));
		log.add(addTrace(new short[] { 0, 1, 1, 0, 1 }));
		log.add(addTrace(new short[] { 0, 1, 1, 0 }));
		log.add(addTrace(new short[] { 0, 0, 1, 1 }));
		log.add(addTrace(new short[] { 0, 0, 1, 1, 0, 1 }));
		log.add(addTrace(new short[] { 0, 0, 1, 1, 0, 1, 0 }));

		log.add(addTrace(new short[] { 1 }));
		log.add(addTrace(new short[] { 1, 0 }));
		log.add(addTrace(new short[] { 1, 0, 1, 0 }));
		log.add(addTrace(new short[] { 1, 0, 1 }));
		log.add(addTrace(new short[] { 1, 1, 0 }));
		log.add(addTrace(new short[] { 1, 0, 0 }));
		log.add(addTrace(new short[] { 1, 1, 0, 1, 0 }));
		log.add(addTrace(new short[] { 1, 1, 0, 1 }));
		log.add(addTrace(new short[] { 1, 0, 0, 1, 0 }));
		log.add(addTrace(new short[] { 1, 0, 0, 1 }));
		log.add(addTrace(new short[] { 1, 1, 0, 0 }));
		log.add(addTrace(new short[] { 1, 1, 0, 0, 1, 0 }));
		log.add(addTrace(new short[] { 1, 1, 0, 0, 1, 0, 1 }));
		return log;
	}

	private static XTrace addTrace(short[] s) {
		XFactory f = XFactoryRegistry.instance().currentDefault();
		XTrace t = f.createTrace();
		for (int i = 0; i < s.length; i++) {
			XEvent e = new XEventImpl();
			XAttribute att = f.createAttributeLiteral("concept:name", s[i] + "", null);
			e.getAttributes().put("concept:name", att);
			t.add(e);
		}

		return t;
	}
}
