package org.processmining.humbertogarciacaballero.models.controller;

import java.util.ArrayList;
import java.util.Iterator;
import java.util.List;
import java.util.Observable;

import javax.swing.JComboBox;
import javax.swing.JComponent;
import javax.swing.JFrame;
import javax.swing.JLabel;
import javax.swing.JPanel;

import org.processmining.humbertogarciacaballero.models.gui.GraphVisualizationL;
import org.processmining.humbertogarciacaballero.models.gui.MiniGraphViewL;
import org.processmining.humbertogarciacaballero.models.gui.PetriNetViewL;
import org.processmining.humbertogarciacaballero.models.model.PNConstantsL;
import org.processmining.prabhakardixit.fcnets.models.InteractivePetriNetL;

import prefuse.Visualization;
import prefuse.action.layout.Layout;
import prefuse.action.layout.SpecifiedLayout;
import prefuse.data.Graph;
import prefuse.data.Node;
import prefuse.data.Tuple;

public class PNetVisualizerControllerL extends Observable {

	// Singleton instance
	//	private static PNetVisualizerController instance = new PNetVisualizerController();

	// Component
	private JComponent component;
	private JPanel statisticsPanel; // TODO: do sth with the panel

	// Data
	private Graph graph;
	//	private HierarchicalLayouter layouter;

	// Visualization data
	private GraphVisualizationL visualization;
	private PetriNetViewL mainView;
	private MiniGraphViewL miniView;

	// Information JPanel
	private JPanel infoPanel;

	// Undo/redo
	//	private CommandManager commandManager;

	private List<Node> initialMarking;

	private boolean alignmentMode;
	//PMD Starts
	//	private String activity;
	private InteractivePetriNetL ipn;
	private boolean conformanceMode;

	JComboBox<String> activityCombo = null;
	Graph tempGraph;

	//also add array of IPNs later to include history
	//PMD Ends
	JLabel fitPre = new JLabel("Fitness: -" + " Precision: -");

	public PNetVisualizerControllerL() {
		graph = new Graph(true);
		initialMarking = new ArrayList<>();

	}

	public void initController(Graph graph, JFrame component) {
		visualization = new GraphVisualizationL(graph, this);
		mainView = new PetriNetViewL(visualization, this);
		miniView = new MiniGraphViewL(visualization);
		//		soundnessInteractionGUI = new SoundnessInteractionGUI(visualization.getVisualization(), mainView.getDisplay());
		this.graph.clear();
		this.graph = graph;
		initialMarking.clear();
		Iterator<Tuple> iter = graph.getNodes().tuples(PNConstantsL.STARTING_NODE_PREDICATE);
		while (iter.hasNext()) {
			initialMarking.add(graph.getNode(iter.next().getRow()));
		}
	}

	//PMD process analytics
	public Graph getData() {
		return graph;
	}

	public Layout getLayouter() {
		return new SpecifiedLayout("graph.nodes", "x", "y"); //PMD 2607
		//		return layouter;
	}

	public PetriNetViewL getMainView() {
		return mainView;
	}

	public MiniGraphViewL getMiniMap() {
		return miniView;
	}

	public void fitViews() {
		visualization.fitLayoutBounds();
	}

	public void executeLayout(boolean animate) {
		visualization.executeLayout(animate);
	}

	public void setHighlightNeighborsEnabled(boolean enabled) {
		mainView.setHighlightControlEnabled(enabled);
	}

	public void setSpanEnabled(boolean enabled) {
		mainView.setSpanEnabled(enabled);
	}

	public void searchFor(String query) {
		visualization.searchFor(query);
	}

	public Visualization getVisualization() {
		return visualization.getVisualization();
	}

	public void setBarDrawingEnabled(boolean enabled) {
		visualization.setBarDecoratorEnabled(enabled);
	}

	public boolean isBarDrawingEnabled() {
		return visualization.isBarDecoratorEnabled();
	}

	public void setLayerSimplifierControlEnabled(boolean enabled) {
		mainView.setLayerSimplifierControlEnabled(enabled);
		visualization.setSimplifierControlEnabled(enabled);
		if (enabled)
			visualization.runSimplifierDecorator();
	}

	public void setEdgeSimplifierControlEnabled(boolean enabled) {
		mainView.setEdgeSimplifierControlEnabled(enabled);
		if (enabled)
			visualization.runSimplifierDecorator();
	}

	public void setLoopHighlightEnabled(boolean enabled) {
		visualization.setLoopModeEnabled(enabled);
	}

	public JPanel getInfoPanel() {
		return infoPanel;
	}

	public void setInfoPanel(JPanel infoPanel) {
		this.infoPanel = infoPanel;
	}

	public int getTransitionsWidth() {
		return visualization.getTransitionsWidth();
	}

	public void setTransitionsWidth(int width) {
		visualization.setTransitionsWidth(width);
	}

	public int getTransitionsHeight() {
		return visualization.getTransitionsHeight();
	}

	public void setTransitionsHeight(int height) {
		visualization.setTransitionsHeight(height);
	}

	public int getPlacesSize() {
		return visualization.getPlacesSize();
	}

	public void setPlacesSize(int size) {
		visualization.setPlacesSize(size);
	}

	public int getFontSize() {
		return visualization.getFontSize();
	}

	public void setFontSize(int size) {
		visualization.setFontSize(size);
	}

	public int getDegreeFontSize() {
		return visualization.getDegreeFontSize();
	}

	public void setDegreeFontSize(int size) {
		visualization.setDegreeFontSize(size);
	}

	public long getAnimationLength() {
		return visualization.getAnimationLength();
	}

	public void setAnimationLength(long animationLength) {
		visualization.setAnimationLength(animationLength);
	}

	public void setRemoveModeEnabled(boolean enabled) {
		mainView.setRemoveModeEnabled(enabled);
	}

	public void setAddModeEnabled(boolean enabled) {
		mainView.setAddModeEnabled(enabled);
	}

	public boolean isTransitionsInOutDegreeEnabled() {
		return visualization.isTransitionInOutDegreeEnabled();
	}

	public int getEdgeColor() {
		return visualization.getEdgeDefaultColor();
	}

	public void setEdgeColor(int color) {
		visualization.setEdgeDefaultColor(color);
	}

	public int getStrokeColor() {
		return visualization.getShapeBorderColor();
	}

	public void setStrokeColor(int color) {
		visualization.setShapeBorderColor(color);
	}

	public List<Node> getInitialMarking() {
		return initialMarking;
	}

	public void setStartingNodeDecoratorEnabled(boolean enabled) {
		visualization.setStartingNodeDecoratorEnabled(enabled);
	}

	public void setStatisticsPanel(JPanel panel) {
		statisticsPanel = panel;
		statisticsPanel.repaint();
		statisticsPanel.revalidate();
	}

	// TODO: Use this method
	public JPanel getStatisticsPanel() {
		return statisticsPanel;
	}

	public void setTransitionSelectionEnabled(boolean selected) {
		mainView.setTransitionSelectionEnabled(selected);
	}

	public void setSelectionNodeEnabled(boolean selected) {
		mainView.setSelectNodeControlEnabled(selected);
	}

	public void setSelectedActivity(String activity) {
		visualization.getVisualization().run(PNConstantsL.VIS_FILL_NODES_ACTIONS_NAME);
		visualization.getVisualization().run(PNConstantsL.VIS_HIGHLIGHT_ACTIONS_NAME);
	}

	//PMD Starts
	public void setActivity(String activity) {
		ipn.setSelectedActivity(activity);
		visualization.getVisualization().repaint();
	}

	public String getActivity() {
		return activityCombo.getSelectedItem().toString();
	}

	public InteractivePetriNetL getIPN() {
		return ipn;
	}

	public void setIPN(InteractivePetriNetL ipn) {
		this.ipn = ipn;
	}

	public void setHighlightNeighborsEnabled_Loop(boolean enabled) {
		//				mainView.setHighlightControl_loopEnabled(enabled);
	}

	public boolean isConformanceMode() {
		return conformanceMode;
	}

	public void setConformanceMode(boolean conformanceMode) {
		this.conformanceMode = conformanceMode;
	}

	/**
	 * @return true if the alignment button is enabled
	 */
	public boolean isAlignmentMode() {
		return alignmentMode;
	}

	public void setAlignmentMode(boolean alignmentMode) {
		this.alignmentMode = alignmentMode;
	}

	public JComboBox<String> getActivityCombo() {
		return activityCombo;
	}

	public void setActivityCombo(JComboBox<String> activityCombo) {
		this.activityCombo = activityCombo;
	}

	public Graph getTempGraph() {
		return tempGraph;
	}

	public void setTempGraph(Graph tempGraph) {
		this.tempGraph = tempGraph;
	}

	public void setFitPre(String fitPre) {
		this.fitPre.setText(fitPre);
	}
	public JLabel getFitPre() {
		return this.fitPre;
	}
	//PMD Ends

}
