package org.processmining.humbertogarciacaballero.models.gui.controllers;

import java.awt.event.MouseEvent;
import java.util.Iterator;

import org.processmining.humbertogarciacaballero.models.model.PNConstantsL;

import prefuse.controls.ControlAdapter;
import prefuse.data.Edge;
import prefuse.util.ColorLib;
import prefuse.visual.AggregateItem;
import prefuse.visual.DecoratorItem;
import prefuse.visual.NodeItem;
import prefuse.visual.VisualItem;

public class LayerSimplifierControlL extends ControlAdapter {

    private int currentLayer;

    /**
     * @see prefuse.controls.Control#itemEntered(prefuse.visual.VisualItem,
     *      java.awt.event.MouseEvent)
     */
    public void itemEntered(VisualItem item, MouseEvent e) {
        if (item instanceof DecoratorItem) {
            // Look for the nodes
            int layer = item.getInt(PNConstantsL.NODE_LAYER_KEY);
            int maxLayer = item.getInt(PNConstantsL.NODE_MAX_LAYER_KEY);

            item.set(VisualItem.STROKECOLOR, ColorLib.rgba(181, 16, 16, 150));
            item.getVisualization().run(PNConstantsL.VIS_REPAINT_ACTIONS_NAME);
            currentLayer = layer == maxLayer ? layer - 1 : layer;
        }
    }

    /**
     * @see prefuse.controls.Control#itemExited(prefuse.visual.VisualItem,
     *      java.awt.event.MouseEvent)
     */
    public void itemExited(VisualItem item, MouseEvent e) {
        if (item instanceof DecoratorItem) {
            item.set(VisualItem.STROKECOLOR, ColorLib.rgba(255, 255, 255, 0));
            item.getVisualization().run(PNConstantsL.VIS_REPAINT_ACTIONS_NAME);
        }
    }

    /**
     * @see prefuse.controls.Control#itemClicked(prefuse.visual.VisualItem,
     *      java.awt.event.MouseEvent)
     */
    @SuppressWarnings("unchecked")
    public void itemClicked(VisualItem item, MouseEvent e) {
        if (item instanceof DecoratorItem) {

            boolean relayout = false;
            Iterator<VisualItem> iter = ((AggregateItem) ((DecoratorItem) item).getDecoratedItem()).items();

            while (iter.hasNext()) {
                VisualItem vi = iter.next();

                if (vi instanceof NodeItem) {
                    NodeItem node = (NodeItem) vi;

                    if (node.getInt(PNConstantsL.NODE_LAYER_KEY) == currentLayer) {
                        Iterator<Edge> iter2 = node.edges();

                        while (iter2.hasNext()) {
                            Edge edge = iter2.next();

                            if (edge.getAdjacentNode(node).getInt(PNConstantsL.NODE_LAYER_KEY) == currentLayer + 1) {
                                edge.setBoolean(PNConstantsL.EDGE_SIMPLIFICABLE,
                                        !edge.getBoolean(PNConstantsL.EDGE_SIMPLIFICABLE));
                                relayout = true;
                            }
                        }
                    }
                }
            }

            if (relayout) {
                item.getVisualization().cancel(PNConstantsL.VIS_ANIMATE_LAYOUT_ACTIONS_NAME);
                item.getVisualization().run(PNConstantsL.VIS_LAYOUT_NO_PAINT_ACTIONS_NAME);
                item.getVisualization().run(PNConstantsL.VIS_ANIMATE_LAYOUT_ACTIONS_NAME);
            }

        }
    }
}
