package org.processmining.humbertogarciacaballero.models.gui.decorator;

import java.awt.geom.Line2D;
import java.awt.geom.Point2D;
import java.awt.geom.Rectangle2D;
import java.util.Iterator;
import java.util.List;

import org.processmining.humbertogarciacaballero.models.model.PNConstantsL;
import org.processmining.humbertogarciacaballero.models.util.UtilsL;

import prefuse.action.layout.Layout;
import prefuse.visual.AggregateItem;
import prefuse.visual.DecoratorItem;
import prefuse.visual.EdgeItem;
import prefuse.visual.NodeItem;

public class LayerSimplifierDecoratorLayoutL extends Layout {

    private static final int OFFSET = 2;

    public LayerSimplifierDecoratorLayoutL(String group) {
        super(group);
    }

    @SuppressWarnings("unchecked")
    @Override
    public void run(double frac) {
        Iterator<DecoratorItem> iter = getVisualization().items(getGroup());
        while (iter.hasNext()) {
            DecoratorItem decorator = iter.next();
            AggregateItem decoratedItem = (AggregateItem) decorator.getDecoratedItem();
            Rectangle2D bounds = UtilsL.getBounds(decoratedItem);
            refineBounds(decoratedItem, bounds);

            decorator.setDouble(PNConstantsL.SIMPLIFIER_DECORATOR_WIDTH, bounds.getWidth());
            decorator.setDouble(PNConstantsL.SIMPLIFIER_DECORATOR_HEIGHT, bounds.getHeight());

            setX(decorator, null, bounds.getX());
            setY(decorator, null, bounds.getY());
        }
    }
    
    private void refineBounds(AggregateItem item, Rectangle2D bounds) {
        int layer = item.getInt(PNConstantsL.NODE_LAYER_KEY);
        Iterator<NodeItem> iter = item.items();
        while (iter.hasNext()) {
            NodeItem node = iter.next();
            if (node.getInt(PNConstantsL.NODE_LAYER_KEY) == layer) {
                Iterator<EdgeItem> edges = node.edges();
                
                while (edges.hasNext()) {
                    EdgeItem edge = edges.next();
                    NodeItem adjacent = edge.getAdjacentItem(node);
                    
                    if (adjacent.getInt(PNConstantsL.NODE_LAYER_KEY) > layer + 1) {
                        List<Point2D> bends = (List<Point2D>) edge.get(PNConstantsL.EDGE_BENDS_KEY);
                        Point2D point;
                        if (edge.getSourceItem().equals(node)) {
                            point = bends.get(0);
                        } else {
                            point = bends.get(bends.size() - 1);
                        }
                        
                        bounds.add(point);
                    }
                }
            }
        }
    }

    @SuppressWarnings("unchecked")
    private Rectangle2D getBounds(AggregateItem item) {
        Iterator<NodeItem> iter = item.items();
        if (iter.hasNext()) {
            NodeItem vi = iter.next();
            int layer = item.getInt(PNConstantsL.NODE_LAYER_KEY);
            Rectangle2D bounds = getBounds(vi, layer);

            while (iter.hasNext()) {
                vi = iter.next();
                if (bounds == null) {
                    bounds = getBounds(vi, layer);
                } else {
                    Rectangle2D aux = getBounds(vi, layer);
                    if (aux != null) {
                        bounds = bounds.createUnion(aux);
                    }
                }
            }

            return bounds;
        }

        // Should never happens
        return new Rectangle2D.Double(0, 0, 0, 0);
    }

    @SuppressWarnings("unchecked")
    private Rectangle2D getBounds(NodeItem item, int layer) {
        Iterator<EdgeItem> iter = item.edges();

        if (iter.hasNext()) {
            EdgeItem edge;
            int adjacentLayer;
            do {
                edge = iter.next();
                adjacentLayer = edge.getAdjacentNode(item).getInt(PNConstantsL.NODE_LAYER_KEY);
            } while (iter.hasNext() && (adjacentLayer < layer));

            if (adjacentLayer >= layer) {
                Rectangle2D bounds;

                if (layer == adjacentLayer || layer + 1 == adjacentLayer) {
                    bounds = edge.getBounds();
                } else {
                    List<Point2D> bends = (List<Point2D>) edge.get(PNConstantsL.EDGE_BENDS_KEY);
                    Point2D point;
                    if (edge.getSourceItem().equals(item)) {
                        point = bends.get(0);
                    } else {
                        point = bends.get(bends.size() - 1);
                    }

                    bounds = new Line2D.Double(edge.getX(), edge.getY(), point.getX(), point.getY()).getBounds2D();
                }

                while (iter.hasNext()) {
                    edge = iter.next();
                    adjacentLayer = edge.getAdjacentNode(item).getInt(PNConstantsL.NODE_LAYER_KEY);
                    if (adjacentLayer >= layer) {
                        if (layer == adjacentLayer || layer + 1 == adjacentLayer) {
                            bounds = bounds.createUnion(edge.getBounds());
                        } else {
                            List<Point2D> bends = (List<Point2D>) edge.get(PNConstantsL.EDGE_BENDS_KEY);
                            Point2D point;
                            if (edge.getSourceItem().equals(item)) {
                                point = bends.get(0);
                            } else {
                                point = bends.get(bends.size() - 1);
                            }

                            bounds.createUnion(new Line2D.Double(edge.getX(), edge.getY(), point.getX(), point.getY())
                                    .getBounds2D());
                        }
                    }
                }

                return bounds;
            }
        }

        return null;
    }
}
