package org.processmining.humbertogarciacaballero.models.gui.render;

import java.awt.Color;
import java.awt.Font;
import java.awt.FontMetrics;
import java.awt.Graphics2D;
import java.awt.Image;
import java.awt.geom.AffineTransform;
import java.awt.geom.Point2D;
import java.awt.geom.RectangularShape;
import java.util.LinkedList;
import java.util.List;

import org.processmining.humbertogarciacaballero.models.model.PNConstantsL;

import prefuse.Constants;
import prefuse.render.LabelRenderer;
import prefuse.util.ColorLib;
import prefuse.util.FontLib;
import prefuse.util.GraphicsLib;
import prefuse.visual.VisualItem;

public class CustomLabelRendereForEdgesBPMNL extends LabelRenderer {
	private AffineTransform m_transform = new AffineTransform();

	public CustomLabelRendereForEdgesBPMNL() {
		this(VisualItem.LABEL);
	}

	public CustomLabelRendereForEdgesBPMNL(String field) {
		super(field);
		setHorizontalPadding(1);
	}

	/**
	 * @see prefuse.render.Renderer#render(java.awt.Graphics2D,
	 *      prefuse.visual.VisualItem)
	 */
	public void render(Graphics2D g, VisualItem item) {
		RectangularShape shape = (RectangularShape) getShape(item);
		if (shape == null)
			return;
		m_font = FontLib.getFont("Tahoma", Font.BOLD, 20);
		//		EdgeItem e = (EdgeItem) item;
		//		Node s = e.getSourceNode();
		//		s.getDouble("y");
		// fill the shape, if requested
		int type = getRenderType(item);
		if (type == RENDER_TYPE_FILL || type == RENDER_TYPE_DRAW_AND_FILL)
			GraphicsLib.paint(g, item, shape, getStroke(item), RENDER_TYPE_FILL);

		// now render the image and text
		String text = m_text;
		Image img = getImage(item);

		if (text == null && img == null)
			return;

		double size = item.getSize();
		boolean useInt = 1.5 > Math.max(g.getTransform().getScaleX(), g.getTransform().getScaleY());
		double x = shape.getMinX() + size * m_horizBorder;
		double y = shape.getMinY() + size * m_vertBorder;

		// render image
		if (img != null) {
			double w = size * img.getWidth(null);
			double h = size * img.getHeight(null);
			double ix = x, iy = y;

			// determine one co-ordinate based on the image position
			switch (m_imagePos) {
				case Constants.LEFT :
					x += w + size * m_imageMargin;
					break;
				case Constants.RIGHT :
					ix = shape.getMaxX() - size * m_horizBorder - w;
					break;
				case Constants.TOP :
					y += h + size * m_imageMargin;
					break;
				case Constants.BOTTOM :
					iy = shape.getMaxY() - size * m_vertBorder - h;
					break;
				default :
					throw new IllegalStateException("Unrecognized image alignment setting.");
			}

			// determine the other coordinate based on image alignment
			switch (m_imagePos) {
				case Constants.LEFT :
				case Constants.RIGHT :
					// need to set image y-coordinate
					switch (m_vImageAlign) {
						case Constants.TOP :
							break;
						case Constants.BOTTOM :
							iy = shape.getMaxY() - size * m_vertBorder - h;
							break;
						case Constants.CENTER :
							iy = shape.getCenterY() - h / 2;
							break;
					}
					break;
				case Constants.TOP :
				case Constants.BOTTOM :
					// need to set image x-coordinate
					switch (m_hImageAlign) {
						case Constants.LEFT :
							break;
						case Constants.RIGHT :
							ix = shape.getMaxX() - size * m_horizBorder - w;
							break;
						case Constants.CENTER :
							ix = shape.getCenterX() - w / 2;
							break;
					}
					break;
			}

			if (useInt && size == 1.0) {
				// if possible, use integer precision
				// results in faster, flicker-free image rendering
				g.drawImage(img, (int) ix, (int) iy, null);
			} else {
				m_transform.setTransform(size, 0, 0, size, ix, iy);
				g.drawImage(img, m_transform, null);
			}
		}

		// render text
		int textColor = item.getTextColor();
		if (text != null && ColorLib.alpha(textColor) > 0) {
			item.setTextColor(ColorLib.color(Color.RED));
			g.setPaint(ColorLib.getColor(textColor));
			g.setFont(m_font);
			FontMetrics fm = DEFAULT_GRAPHICS.getFontMetrics(m_font);

			// compute available width
			double tw;
			switch (m_imagePos) {
				case Constants.TOP :
				case Constants.BOTTOM :
					tw = shape.getWidth() - 2 * size * m_horizBorder;
					break;
				default :
					tw = m_textDim.width;
			}

			// compute available height
			double th;
			switch (m_imagePos) {
				case Constants.LEFT :
				case Constants.RIGHT :
					th = shape.getHeight() - 2 * size * m_vertBorder;
					break;
				default :
					th = m_textDim.height;
			}

			// compute starting y-coordinate
			y += fm.getAscent();
			switch (m_vTextAlign) {
				case Constants.TOP :
					break;
				case Constants.BOTTOM :
					y += th - m_textDim.height;
					break;
				case Constants.CENTER :
					y += (th - m_textDim.height) / 2;
			}

			LinkedList<Point2D> bends = (LinkedList<Point2D>) (item.get(PNConstantsL.EDGE_BENDS_KEY) != null ? new LinkedList<>(
					(List<Point2D>) item.get(PNConstantsL.EDGE_BENDS_KEY)) : new LinkedList<>());
			y = bends.getLast().getY() - 9;
			// render each line of text
			int lh = fm.getHeight(); // the line height
			int start = 0, end = text.indexOf(m_delim);
			for (; end >= 0; y += lh) {
				drawString(g, fm, text.substring(start, end), useInt, x, y, tw);
				start = end + 1;
				end = text.indexOf(m_delim, start);
			}
			drawString2(g, fm, text.substring(start), useInt, x, y, tw, bends.getLast());
		}

		// draw border
		if (type == RENDER_TYPE_DRAW || type == RENDER_TYPE_DRAW_AND_FILL) {
			GraphicsLib.paint(g, item, shape, getStroke(item), RENDER_TYPE_DRAW);
		}
	}

	private final void drawString(Graphics2D g, FontMetrics fm, String text, boolean useInt, double x, double y,
			double w) {
		// compute the x-coordinate
		double tx;
		switch (m_hTextAlign) {
			case Constants.LEFT :
				tx = x;
				break;
			case Constants.RIGHT :
				tx = x + w - fm.stringWidth(text);
				break;
			case Constants.CENTER :
				tx = x + (w - fm.stringWidth(text)) / 2;
				break;
			default :
				throw new IllegalStateException("Unrecognized text alignment setting.");
		}
		// use integer precision unless zoomed-in
		// results in more stable drawing
		if (useInt) {
			g.drawString(text, (int) tx, (int) y);
		} else {
			g.drawString(text, (float) tx, (float) y);
		}
	}

	private final void drawString2(Graphics2D g, FontMetrics fm, String text, boolean useInt, double x, double y,
			double w, Point2D lastBends) {

		text = text.replace(")", "");
		text = text.replace("(", "");
		text = text.replace("AND", "&&");
		String[] str2 = text.split("\\|\\|");

		int count = 0;

		text = "";
		for (String s : str2) {
			if (text == "")
				text = "(";
			else
				text += "||";
			text += s.replaceAll("&&", ",");
			;
		}
		text += ")";

		//PREV NEXT remove
		int index = text.indexOf("PREV");
		if (index != -1) {
			String text2 = text.toString();
			text = "";
			for (String s : text2.split("")) {
				if (text2.indexOf(s) > index + 4 || text2.indexOf(s) < index)
					if (s.equals("(") || s.equals(")"))
						text += s;
					else
						text += s + "1";
			}
		}

		index = text.indexOf("NEXT");
		if (index != -1) {
			String text2 = text.toString();
			text = "";
			for (String s : text2.split("")) {
				if (text2.indexOf(s) > index + 3 || text2.indexOf(s) < index - 1)
					if (s.equals("(") || s.equals(")"))
						text += s;
					else
						text += s + "2";
			}
		}
		String text1 = "";
		String text2 = "";

		int len = text.split("\\|\\|").length;
		len = (len % 2 == 0 || len == 1) ? len : len + 1;

		for (String str3 : text.split("\\|\\|")) {
			count++;
			if (count > len / 2 && len != 1)
				if (count != text.split("\\|\\|").length) {
					text2 += " (" + str3 + "),";
				} else {
					text2 += " (" + str3;
				}
			else {
				if (count == 1)
					text1 += (len != 1) ? str3 + ")," : str3;
				else
					text1 += " (" + str3 + "),";
			}
		}

		for (int i = 0; i < 2; i++) {

			if (i == 0) {
				text = text1;
			} else {
				if (text2 != "")
					text = text2;
				y += 25;
				if (len < 2)
					continue;
			}
			String[] str = text.split("");
			g.setColor(Color.RED);
			double tx = 0;
			double strWidth = 0.;
			boolean setRed = false;
			g.setPaint(ColorLib.getColor(ColorLib.red(1)));

			for (String str1 : str) {
				//			tx += strWidth;
				if (str1.equals("!")) {
					setRed = true;
					continue;
				}
				if (setRed) {
					g.setColor((Color.RED));
				} else {
					g.setPaint(ColorLib.getColor(ColorLib.gray(4)));
					g.setColor(Color.BLACK);
				}
				setRed = false;
				switch (m_hTextAlign) {
					case Constants.LEFT :
						tx = x;
						break;
					case Constants.RIGHT :
						tx = x + w - fm.stringWidth(str1);
						break;
					case Constants.CENTER :
						tx = x + (w - fm.stringWidth(text)) / 2 + strWidth;
						break;
					default :
						throw new IllegalStateException("Unrecognized text alignment setting.");
				}
				// use integer precision unless zoomed-in
				// results in more stable drawing
				if (useInt) {
					g.drawString(str1, (int) tx, (int) y);
				} else {
					g.drawString(str1, (float) tx, (float) y);
				}
				strWidth += fm.stringWidth(str1);
			}
		}
	}
}
