package org.processmining.humbertogarciacaballero.models.gui.render;

import java.awt.BasicStroke;
import java.awt.Color;
import java.awt.Graphics2D;
import java.awt.Shape;
import java.awt.Stroke;
import java.awt.geom.AffineTransform;
import java.awt.geom.Ellipse2D;
import java.awt.geom.GeneralPath;
import java.awt.geom.Point2D;
import java.awt.geom.Rectangle2D;
import java.awt.geom.RectangularShape;

import org.processmining.humbertogarciacaballero.models.controller.PNetVisualizerControllerL;
import org.processmining.humbertogarciacaballero.models.model.PNConstantsL;
import org.processmining.humbertogarciacaballero.models.util.UtilsL;

import prefuse.Constants;
import prefuse.render.AbstractShapeRenderer;
import prefuse.render.ShapeRenderer;
import prefuse.util.ColorLib;
import prefuse.util.GraphicsLib;
import prefuse.visual.VisualItem;

public class CustomShapeRendererL extends ShapeRenderer {

	private GeneralPath m_path;
	private PNetVisualizerControllerL controller;
	private boolean transitionsInOutDegree;

	public CustomShapeRendererL(PNetVisualizerControllerL con) {
		super();
		setBaseSize(1);
//		controller = PNetVisualizerController.getInstance();
		controller = con;
		m_path = new GeneralPath();
	}

	/**
	 * Draws the specified shape into the provided Graphics context, using
	 * stroke and fill color values from the specified VisualItem. This method
	 * can be called by subclasses in custom rendering routines.
	 */
	protected void drawShape(Graphics2D g, VisualItem item, Shape shape) {
		if (item.getBoolean("temp")) {
			item.setStrokeColor(PNConstantsL.TEMP_NODES_EDGES_COLOR);
//			item.setFillColor(ColorLib.rgb(144, 195, 212));
			g.setColor(Color.CYAN);
		}
		if (item.getInt(PNConstantsL.FILLING_EXTRA_COLOR_KEY) != 0) {
			int extraColor = item.getInt(PNConstantsL.FILLING_EXTRA_COLOR_KEY);
			paint(g, item, shape, getStroke(item), getRenderType(item), ColorLib.getColor(extraColor));
		} else {
			GraphicsLib.paint(g, item, shape, getStroke(item), getRenderType(item));
		}
	}

	/**
	 * @see prefuse.render.AbstractShapeRenderer#getRawShape(prefuse.visual.VisualItem)
	 */
	protected Shape getRawShape(VisualItem item) {
		synchronized (item.getVisualization()) {
			
		
		int stype = item.getShape();
		double x = item.getX();
		if (Double.isNaN(x) || Double.isInfinite(x))
			x = 0;
		double y = item.getY();
		if (Double.isNaN(y) || Double.isInfinite(y))
			y = 0;

		double width = getBaseSize() * UtilsL.getWidth(item);
		double height = getBaseSize() * UtilsL.getHeight(item);
		if (item.getBoolean(PNConstantsL.NODE_TAU_TRANSITION_KEY)) {
					width = PNConstantsL.TAU_TRANSITIONS_SIZE;

		}

		// Center the shape around the specified x and y
		if (width > 1) {
			x = x - width / 2;
			y = y - height / 2;
		}

		switch (stype) {
			case PNConstantsL.SHAPE_TRANSITION :
				if (item.getBoolean(PNConstantsL.NODE_TAU_TRANSITION_KEY)) {
					return rectangle(x, y, width, height);
				} else {
					width = width;
					return transition(x, y, width*2, height);
				}
			case Constants.SHAPE_ELLIPSE :
			case 0 : //places size
				return place(x, y, width);
			default :
				throw new IllegalStateException("Unknown shape type: " + stype);
		}
		}
	}

	private Shape place(double x, double y, double size) {
		return ellipse(x, y, size, size);
	}

	/**
	 * Returns a rectangle of the given dimensions. The resulting rectangle
	 * includes two sub-rectangles inside. This shape represents a Petri Net
	 * Transition.
	 */
	private Shape transition(double x, double y, double width, double height) {
		Shape rectangle = rectangle(x, y, width, height);
		if (!transitionsInOutDegree) {
			return rectangle;
		}

		double subrectWidth = (width / 2) - 1;
		double subrectHeight = ((height * 3) / 10);
		m_path.reset();
		// Add the main rectangle
		m_path.append(rectangle, false);

		Point2D p0, p1, p2, p3, p4, p5;

		// Create subrectangles
//		switch (controller.getLayoutOrientation()) {
//			case TOP_TO_BOTTOM :
//				p0 = new Point2D.Double(x + subrectWidth, y);
//				p1 = new Point2D.Double(x + subrectWidth, y + subrectHeight);
//				p2 = new Point2D.Double(x, y + subrectHeight);
//				p3 = new Point2D.Double(x, y + height - subrectHeight);
//				p4 = new Point2D.Double(x + subrectWidth, y + height - subrectHeight);
//				p5 = new Point2D.Double(x + subrectWidth, y + height);
//				break;
//			case BOTTOM_TO_TOP :
//				p0 = new Point2D.Double(x + width - subrectWidth, y);
//				p1 = new Point2D.Double(x + width - subrectWidth, y + subrectHeight);
//				p2 = new Point2D.Double(x + width, y + subrectHeight);
//				p3 = new Point2D.Double(x + width, y + height - subrectHeight);
//				p4 = new Point2D.Double(x + width - subrectWidth, y + height - subrectHeight);
//				p5 = new Point2D.Double(x + width - subrectWidth, y + height);
//				break;
//			case LEFT_TO_RIGHT :
//				p0 = new Point2D.Double(x, y + height - subrectHeight);
//				p1 = new Point2D.Double(x + subrectWidth, y + height - subrectHeight);
//				p2 = new Point2D.Double(x + subrectWidth, y + height);
//				p3 = new Point2D.Double(x + width - subrectWidth, y + height);
//				p4 = new Point2D.Double(x + width - subrectWidth, y + height - subrectHeight);
//				p5 = new Point2D.Double(x + width, y + height - subrectHeight);
//				break;
//			default :
				p0 = new Point2D.Double(x, y + subrectHeight);
				p1 = new Point2D.Double(x + subrectWidth, y + subrectHeight);
				p2 = new Point2D.Double(x + subrectWidth, y);
				p3 = new Point2D.Double(x + width - subrectWidth, y);
				p4 = new Point2D.Double(x + width - subrectWidth, y + subrectHeight);
				p5 = new Point2D.Double(x + width, y + subrectHeight);
//				break;
//		}

		m_path.moveTo(p0.getX(), p0.getY());
		m_path.lineTo(p1.getX(), p1.getY());
		m_path.lineTo(p2.getX(), p2.getY());

		m_path.moveTo(p3.getX(), p3.getY());
		m_path.lineTo(p4.getX(), p4.getY());
		m_path.lineTo(p5.getX(), p5.getY());

		return m_path;
	}

	public void setTransitionsInOutDegreeEnabled(boolean enabled) {
		transitionsInOutDegree = enabled;
	}

	public boolean isTransitionsInOutDegreeEnabled() {
		return transitionsInOutDegree;
	}

	public static void paint(Graphics2D g, VisualItem item, Shape shape, BasicStroke stroke, int type, Color extraColor) {
		// if render type is NONE, then there is nothing to do
		if (type == AbstractShapeRenderer.RENDER_TYPE_NONE)
			return;

		// set up colors
		Color strokeColor = ColorLib.getColor(item.getStrokeColor());
		Color fillColor = ColorLib.getColor(item.getFillColor());
		boolean sdraw = (type == AbstractShapeRenderer.RENDER_TYPE_DRAW || type == AbstractShapeRenderer.RENDER_TYPE_DRAW_AND_FILL)
				&& strokeColor.getAlpha() != 0;
		boolean fdraw = (type == AbstractShapeRenderer.RENDER_TYPE_FILL || type == AbstractShapeRenderer.RENDER_TYPE_DRAW_AND_FILL)
				&& fillColor.getAlpha() != 0;
		if (!(sdraw || fdraw))
			return;

		Stroke origStroke = null;
		if (sdraw) {
			origStroke = g.getStroke();
			g.setStroke(stroke);
		}

		int x, y, w, h, aw, ah;
		double xx, yy, ww, hh;

		// see if an optimized (non-shape) rendering call is available for us
		// these can speed things up significantly on the windows JRE
		// it is stupid we have to do this, but we do what we must
		// if we are zoomed in, we have no choice but to use
		// full precision rendering methods.
		AffineTransform at = g.getTransform();
		double scale = Math.max(at.getScaleX(), at.getScaleY());
		if (scale > 1.5) {
			if (fdraw) {
				g.setPaint(extraColor);
				g.fill(shape);

				Rectangle2D r = shape.getBounds2D();
				xx = r.getX();
				ww = r.getWidth();
				yy = r.getY();
				hh = r.getHeight();

				x = (int) xx;
				y = (int) yy;
				w = (int) (ww + xx - x);
				h = (int) (hh + yy - y);

				int factor = (w / 2);
				int width = w;

				g.setPaint(fillColor);
				Shape clip = g.getClip();
				g.setClip(x, y, width -= factor, h);
				g.fill(shape);
				g.setClip(clip);
			}
			if (sdraw) {
				g.setPaint(strokeColor);
				g.draw(shape);
			}
		} else if (shape instanceof RectangularShape) {
			RectangularShape r = (RectangularShape) shape;
			xx = r.getX();
			ww = r.getWidth();
			yy = r.getY();
			hh = r.getHeight();

			x = (int) xx;
			y = (int) yy;
			w = (int) (ww + xx - x);
			h = (int) (hh + yy - y);

			if (shape instanceof Rectangle2D) {
				if (fdraw) {
					int factor = (w / 2);
					int width = w;
					g.setPaint(extraColor);
					g.fillRect(x, y, w, h);

					g.setPaint(fillColor);
					g.setClip(x, y, width -= factor, h);
					g.fillRect(x, y, w, h);

				}
				if (sdraw) {
					g.setPaint(strokeColor);
					g.drawRect(x, y, w, h);
				}
			} else if (shape instanceof Ellipse2D) {
				if (fdraw) {
					int factor = (w / 2);
					int width = w;
					g.setPaint(extraColor);
					g.fillOval(x, y, w, h);

					g.setPaint(fillColor);
					g.setClip(x, y, width -= factor, h);
					g.fillOval(x, y, w, h);
				}
				if (sdraw) {
					g.setPaint(strokeColor);
					g.drawOval(x, y, w, h);
				}
			} else {
				if (fdraw) {
					int factor = (w / 2);
					int width = w;
					g.setPaint(extraColor);
					g.fill(shape);

					g.setPaint(fillColor);
					g.setClip(x, y, width -= factor, h);
					g.fill(shape);
				}
				if (sdraw) {
					g.setPaint(strokeColor);
					g.draw(shape);
				}
			}
		}
		if (sdraw) {
			g.setStroke(origStroke);
		}
	}
}
