package org.processmining.humbertogarciacaballero.models.gui.render;

import java.awt.Color;
import java.awt.Graphics2D;
import java.awt.Shape;
import java.awt.geom.Ellipse2D;
import java.awt.geom.Path2D;
import java.awt.geom.Rectangle2D;
import java.awt.geom.RectangularShape;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Map;
import java.util.Set;

import org.processmining.humbertogarciacaballero.models.controller.PNetVisualizerControllerL;
import org.processmining.humbertogarciacaballero.models.gui.GraphVisualizationL;
import org.processmining.humbertogarciacaballero.models.model.PNConstantsL;
import org.processmining.models.graphbased.directed.petrinet.elements.Place;
import org.processmining.models.graphbased.directed.petrinet.elements.Transition;
import org.processmining.prabhakardixit.fcnets.models.ControllerViewModeL;
import org.processmining.prabhakardixit.fcnets.models.InteractivePetriNetL;

import prefuse.Constants;
import prefuse.util.ColorLib;
import prefuse.visual.NodeItem;
import prefuse.visual.VisualItem;

public class VizToAlignmentMapperL extends CustomShapeRendererL {
	PNetVisualizerControllerL controller;
	GraphVisualizationL graphVisualization;

	public VizToAlignmentMapperL(PNetVisualizerControllerL controller, GraphVisualizationL graphVisualization) {
		super(controller);
		new CustomShapeRendererL(controller);
		this.controller = controller;
		this.graphVisualization = graphVisualization;
	}

	public void render(Graphics2D g, VisualItem vitem) {

		switch (controller.getIPN().getViewMode()) {
			case ALIGNMENT :
				renderAlignments(g, vitem);
				break;
			case DIRECTLYFOLLOWS :
				renderFollowsPrecedes(g, vitem, false);
				break;
			case EVENTUALLYFOLLOWS :
				renderFollowsPrecedes(g, vitem, false);
				break;
			case PERFORMANCE :
				break;
			default :
				renderFollowsPrecedes(g, vitem, true);
				break;

		}

	}

	private void renderFollowsPrecedes(Graphics2D g, VisualItem vitem, boolean defaultMode) {
		if (vitem.get(PNConstantsL.NODE_TYPE_KEY).equals("TRANSITION")) {

			NodeItem item = (NodeItem) vitem;
			float precedesVal = 0f;
			float followsVal = 0f;
			float maxVal = 0f;
			String selectedAct = controller.getActivity();
			String label = (String) item.get("label");
			// get precedes and follows relations
			if (vitem.getBoolean(PNConstantsL.NODE_TAU_TRANSITION_KEY) == false && !defaultMode
					&& !vitem.isHighlighted() && !selectedAct.equals(label)) {
				Map<String, Integer> follows;
				Map<String, Integer> precedes;
				if (controller.getIPN().getViewMode().equals(ControllerViewModeL.DIRECTLYFOLLOWS)) {
					follows = controller.getIPN().getLogInfo().getDirectlyFollows();
					precedes = controller.getIPN().getLogInfo().getDirectlyPrecedes();
				} else {
					follows = controller.getIPN().getLogInfo().getEventuallyFollows();
					precedes = controller.getIPN().getLogInfo().getEventuallyPrecedes();
				}

				//filter to contain only the acts present in net
				Set<String> actsInNet = new HashSet<String>();
				Map<String, Integer> followsFiltered = new HashMap<String, Integer>(follows);
				Map<String, Integer> precedesFiltered = new HashMap<String, Integer>(precedes);
				for (Transition t : controller.getIPN().getselectedPetriNetMap().getPetriNet().getTransitions()) {
					actsInNet.add(t.getLabel());
				}
				for (String s : follows.keySet())
					if (!actsInNet.contains(s))
						followsFiltered.remove(s);

				for (String s : precedes.keySet())
					if (!actsInNet.contains(s))
						precedesFiltered.remove(s);

				Map<String, Integer> combinedVal = new HashMap<String, Integer>();
				for (String t : followsFiltered.keySet()) {
					if (t.equals(label))
						followsVal = followsFiltered.get(t);
					maxVal = maxVal > followsVal ? maxVal : followsVal;
					combinedVal.put(t, followsFiltered.get(t));
				}
				for (String t : precedesFiltered.keySet()) {
					if (t != "") {
						int val = combinedVal.containsKey(t) ? combinedVal.get(t) + precedesFiltered.get(t)
								: precedesFiltered.get(t);
						if (t.equals(label))
							precedesVal = precedesFiltered.get(t);
						maxVal = maxVal > val ? maxVal : val;
						combinedVal.put(t, val);
					}
				}
				Double ratioWidth = (double) (precedesVal / (followsVal + precedesVal));
				Float ratioCoOccurs = ((followsVal + precedesVal) / maxVal);

				// render the edge line
				Shape shape = getRawShape(item);
				if (shape != null) {
					Shape tmpnod = shape.getBounds();
					Double w = ((RectangularShape) tmpnod).getWidth();

					Double w1 = w * ratioWidth;
					Double w2 = w - w1;

					Shape r = new Rectangle2D.Double(((RectangularShape) tmpnod).getX(),
							((RectangularShape) tmpnod).getY(), w1, ((RectangularShape) tmpnod).getHeight());
					Shape r2 = new Rectangle2D.Double(((RectangularShape) tmpnod).getX() + w1,
							((RectangularShape) tmpnod).getY(), w2, ((RectangularShape) tmpnod).getHeight());

					// ColorLib.gray(100);
					// int i =(int) Math.round(100*coOccurs);
					Float f = ratioCoOccurs;
					//									g.draw(shape);
					drawShape(g, vitem, shape);
					Color c1 = new Color(1f, 0f, 1f, f);
					Color c2 = new Color(1f, 1f, 0f, f);

					if (w1 != 0) {
						g.setColor(c1);
						g.fill(r);
						g.draw(r);
					}
					if (w2 != 0) {
						g.setColor(c2);
						g.draw(r2);
						g.fill(r2);
					}
					g.setColor(ColorLib.getColor(PNConstantsL.EDGE_DEFAULT_COLOR));
					//									g.draw(shape);

				}
			} else {

				Shape shape = getRawShape(item);
				Shape tmpnod = shape.getBounds();
				Shape r = new Rectangle2D.Double(((RectangularShape) tmpnod).getX(),
						((RectangularShape) tmpnod).getY(), ((RectangularShape) tmpnod).getWidth(),
						((RectangularShape) tmpnod).getHeight());

				// int i =(int) Math.round(100*coOccurs);

				// Color c = (Color) ColorLib.gray(100);

				g.setColor(ColorLib.getColor(PNConstantsL.EDGE_DEFAULT_COLOR));

				if (label == "") {
					g.setColor(Color.BLACK);
					g.fill(r);
				}
				//								g.draw(r);
				drawShape(g, vitem, shape);
			}
			// render the edge arrow head, if appropriate
			g.setPaint(ColorLib.getColor(item.getFillColor()));
			//}
		} else {
			NodeItem item = (NodeItem) vitem;
			Shape shape = getRawShape(item);
			g.setColor(ColorLib.getColor(PNConstantsL.EDGE_DEFAULT_COLOR));
			g.draw(shape);
			if (vitem.get(PNConstantsL.NODE_TYPE_KEY).equals("PLACE") && !(shape instanceof Ellipse2D.Double)) {
				item.setShape(Constants.SHAPE_ELLIPSE);
				shape = getShape(item);
				Ellipse2D ellipse = (Ellipse2D) shape;
				Ellipse2D el = new Ellipse2D.Double(ellipse.getHeight(), ellipse.getWidth(), ellipse.getX(),
						ellipse.getY());
				g.setColor(ColorLib.getColor(PNConstantsL.EDGE_DEFAULT_COLOR));
				g.draw((Shape) el);
				g.setPaint(ColorLib.getColor(vitem.getFillColor()));
			} else {
				if (shape != null)
					drawShape(g, vitem, shape);
			}
		}

	}

	private void renderAlignments(Graphics2D g, VisualItem vitem) {

		//conformance visualization
		NodeItem item = (NodeItem) vitem;

		//if full view, then use placeinfo in the viz
		Map<Place, Integer> placeMarkingInfo = getPlaceMarking(controller.getIPN());

		//		System.out.println(vitem.get("label") + " " + vitem.get(PNConstants.NODE_TYPE_KEY));
		if (vitem.get(PNConstantsL.NODE_TYPE_KEY).equals("TRANSITION")) {
			//			System.out.println(" Transition is : " + vitem.get("label"));
			//set sync-error ratio
			Shape shape = getRawShape(item);
			drawShape(g, vitem, shape);
			if (vitem.getBoolean(PNConstantsL.NODE_TAU_TRANSITION_KEY) == false && controller.isAlignmentMode()) {
				int nbrOfOcc = 0;
				Transition tOrg = controller.getIPN().getTranBasedOnID(Integer.valueOf(item.get("ID").toString()));
				boolean nodeSelected = false;
				double nbrLogMoves = 0;
				double nbrSyncMoves = 0;
				double nbrModelMoves = 0;

				double maxSyncMoveCount = 0;

				if (graphVisualization.getTpa() != null) {
					nodeSelected = true;
					//					tOrg = graphVisualization.getTpa().getPetriNetMap().getMapTransitions().get(tOrg);
					for (Transition t : graphVisualization.getTpa().getPetriNetMap().getPetriNet().getTransitions())
						if (!t.isInvisible()) {
							//get max syncMoveScore
							maxSyncMoveCount = graphVisualization.getTpa().getSyncMoveMap().get(t) > maxSyncMoveCount ? graphVisualization
									.getTpa().getSyncMoveMap().get(t)
									: maxSyncMoveCount;
							if (t.getLabel().equals(vitem.get("label")))
								nbrOfOcc++;
						}
					if (graphVisualization.getTpa().getLogMoveMap().containsKey(tOrg.getLabel()))
						nbrLogMoves = graphVisualization.getTpa().getLogMoveMap().get(tOrg.getLabel()) / nbrOfOcc;

					if (graphVisualization.getTpa().getSyncMoveMap().containsKey(tOrg))
						nbrSyncMoves = graphVisualization.getTpa().getSyncMoveMap().get(tOrg);

					if (graphVisualization.getTpa().getModelMove().containsKey(tOrg))
						nbrModelMoves = graphVisualization.getTpa().getModelMove().get(tOrg);

				} else {

					for (Transition t : controller.getIPN().getselectedPetriNetMap().getPetriNet().getTransitions())
						//						if (!t.isInvisible() && t.getLabel().equals(vitem.get("label")))
						//							nbrOfOcc++;
						if (!t.isInvisible()) {
							//get max syncMoveScore
							maxSyncMoveCount = controller.getIPN().getAggrSyncMoves().get(t) > maxSyncMoveCount ? controller
									.getIPN().getAggrSyncMoves().get(t)
									: maxSyncMoveCount;
							if (t.getLabel().equals(vitem.get("label")))
								nbrOfOcc++;
						}
					if (controller.getIPN().getAggrLogMoves().containsKey(tOrg.getLabel()))
						nbrLogMoves = controller.getIPN().getAggrLogMoves().get(tOrg.getLabel()) / nbrOfOcc;

					if (controller.getIPN().getAggrSyncMoves().containsKey(tOrg))
						nbrSyncMoves = controller.getIPN().getAggrSyncMoves().get(tOrg);

					if (controller.getIPN().getAggrModelMoves().containsKey(tOrg))
						nbrModelMoves = controller.getIPN().getAggrModelMoves().get(tOrg);
				}
				System.out.println(" Transition is : " + tOrg.getLabel() + " Sync : " + nbrSyncMoves + " Model : "
						+ nbrModelMoves + " Log : " + nbrLogMoves);

				Shape tmpnod = shape.getBounds();
				Double w = ((RectangularShape) tmpnod).getWidth();
				Double w1 = (((double) nbrSyncMoves) / (nbrSyncMoves + nbrLogMoves + nbrModelMoves)) * w;
				Double w2 = w - w1;

				System.out.println("W1 : " + w1 + " w2 : " + w2);

				Shape r = new Rectangle2D.Double(((RectangularShape) tmpnod).getX(),
						((RectangularShape) tmpnod).getY() - 3, w1, ((RectangularShape) tmpnod).getHeight() / 8);
				Shape r2 = new Rectangle2D.Double(((RectangularShape) tmpnod).getX() + w1,
						((RectangularShape) tmpnod).getY() - 3, w2, ((RectangularShape) tmpnod).getHeight() / 8);

				Color c1 = new Color(0f, 1f, 0f);
				Color c2 = new Color(1f, 0f, 0f);

				g.setColor(c2);
				g.fill(r2);
				g.draw(r2);
				g.setColor(c1);
				g.draw(r);
				g.fill(r);
				if (maxSyncMoveCount > 0.) {
					Color c4 = new Color(0f, 0f, 1f, (float) (nbrSyncMoves / maxSyncMoveCount));
					g.setColor(c4);
					g.fill(tmpnod);
				}
				g.setColor(ColorLib.getColor(PNConstantsL.EDGE_DEFAULT_COLOR));

				//check if the current node is selected

				if (graphVisualization.getTpa() == null) { //draw diamond incase this is the selected tran
					Path2D path = new Path2D.Double();
					Double valoresX[] = { ((RectangularShape) tmpnod).getX() + w / 2,
							((RectangularShape) tmpnod).getX() + w / 2 - 5.,
							((RectangularShape) tmpnod).getX() + w / 2, ((RectangularShape) tmpnod).getX() + w / 2 + 5. };
					Double valoresY[] = { ((RectangularShape) tmpnod).getY(), ((RectangularShape) tmpnod).getY() - 3.,
							((RectangularShape) tmpnod).getY() - 6, ((RectangularShape) tmpnod).getY() - 3 };
					path.moveTo(valoresX[0], valoresY[0]);
					for (int i = 1; i < valoresX.length; ++i) {
						path.lineTo(valoresX[i], valoresY[i]);
					}
					path.closePath();
					g.fill(path);
				}
			}

		} else {
			Shape shape = getRawShape(item);
			g.setColor(ColorLib.getColor(PNConstantsL.EDGE_DEFAULT_COLOR));
			g.draw(shape);
			if (vitem.get(PNConstantsL.NODE_TYPE_KEY).equals("PLACE") && !(shape instanceof Ellipse2D.Double)) {
				item.setShape(Constants.SHAPE_ELLIPSE);
				shape = getShape(item);
				Ellipse2D ellipse = (Ellipse2D) shape;
				Ellipse2D el = new Ellipse2D.Double(ellipse.getHeight(), ellipse.getWidth(), ellipse.getX(),
						ellipse.getY());
				g.setColor(ColorLib.getColor(PNConstantsL.EDGE_DEFAULT_COLOR));
				g.draw((Shape) el);
				g.setPaint(ColorLib.getColor(vitem.getFillColor()));
			} else {
				if (shape != null)
					drawShape(g, vitem, shape);
			}

		}
	}

	private Map<Place, Integer> getPlaceMarking(InteractivePetriNetL ipn) {
		Map<Place, Integer> placeMarking = new HashMap<Place, Integer>();

		return null;
	}

}
