package org.processmining.incrementaldiscovery;

import java.util.HashSet;
import java.util.Set;

import org.deckfour.xes.model.XLog;
import org.processmining.acceptingpetrinet.models.AcceptingPetriNet;
import org.processmining.acceptingpetrinet.models.impl.AcceptingPetriNetImpl;
import org.processmining.contexts.uitopia.UIPluginContext;
import org.processmining.decomposelogs.AggregatedHeuristicsL;
import org.processmining.decomposelogs.LogModelComparisonNewL;
import org.processmining.framework.plugin.PluginContext;
import org.processmining.framework.plugin.ProMCanceller;
import org.processmining.incrementaldiscovery.models.IncrementalModelL;
import org.processmining.models.graphbased.directed.petrinet.Petrinet;
import org.processmining.models.graphbased.directed.petrinet.elements.Transition;
import org.processmining.plugins.InductiveMiner.efficienttree.EfficientTreeReduce.ReductionFailedException;
import org.processmining.projectedrecallandprecision.framework.CompareParameters;
import org.processmining.projectedrecallandprecision.helperclasses.AutomatonFailedException;
import org.processmining.projectedrecallandprecision.plugins.CompareLog2PetriNetPlugin;
import org.processmining.projectedrecallandprecision.plugins.ComparePetriNet2PetriNetPlugin;
import org.processmining.projectedrecallandprecision.result.ProjectedRecallPrecisionResult;
import org.processmining.projectedrecallandprecision.result.ProjectedRecallPrecisionResult.ProjectedMeasuresFailedException;

public class AlignmentsL {

	public static float[] getProjectedFitnessAndPrecision(XLog log, IncrementalModelL mnew)
			throws ProjectedMeasuresFailedException {
		Petrinet pn = mnew.getNet();
		AcceptingPetriNet model = new AcceptingPetriNetImpl(pn);
		float[] fitPre = new float[2];
		CompareLog2PetriNetPlugin compare = new CompareLog2PetriNetPlugin();
		CompareParameters param = new CompareParameters();

		Set<String> act = new HashSet<String>();
		for (Transition t : model.getNet().getTransitions())
			if (!t.isInvisible()) {
				//				System.out.println("\n Transition : " + t);
				act.add(t.getLabel());

			}
		//		System.out.println("\n LOG: ");
		//		for (XTrace t : log)
		//			for (XEvent e : t)
		//				System.out.println(XConceptExtension.instance().extractName(e));
		param.setK(act.size() > 1 ? 2 : 1);
		ProjectedRecallPrecisionResult result = null;
		try {
			result = compare.measure(log, model, param, new ProMCanceller() {
				public boolean isCancelled() {
					return false;
				}
			});
		} catch (AutomatonFailedException | InterruptedException | ProjectedMeasuresFailedException e) {
			e.printStackTrace();
		}
		fitPre[0] = (float) result.getRecall();
		fitPre[1] = (float) result.getPrecision();

		mnew.getImn().setProjectedResult(result);

		return fitPre;
	}

	//	public static float[] getTraditionalFitnessAndPrecision(PluginContext context, XLog log, IncrementalModel mnew) {
	//		float[] fitPre = new float[2];
	//		PNRepResult replay = AlignmentCalculations.calculateNAlignment1(context, log, mnew.getNet(), 50, false, null);
	////		if (replay != null) {
	//			fitPre[0] = ((Double) replay.getInfo().get(PNRepResult.TRACEFITNESS)).floatValue();
	//
	//			//calculate precision
	//			AlignmentPrecGen precGen = new AlignmentPrecGen();
	//			AlignmentPrecGenRes precRes = precGen.measureConformanceAssumingCorrectAlignment(context,
	//					AlignmentCalculations.transitionToEventClassMapperByLabel(log, mnew.getNet(), null, false), replay,
	//					mnew.getNet(), PetrinetUtils.getInitialMarking(mnew.getNet()), true);
	//
	//			fitPre[1] = ((Double) precRes.getPrecision()).floatValue();
	//
	//			mnew.getImn().setReplayReult(replay);
	//		//		}
	//		return fitPre;
	//	}

	public static float[] getProjectedFitnessAndPrecisionUsingPatterns(PluginContext context,
			AggregatedHeuristicsL ahl, IncrementalModelL mnew, XLog log) throws ProjectedMeasuresFailedException {
		Petrinet pn = mnew.getNet();
		Set<Thread> threads = Thread.getAllStackTraces().keySet();
//		System.out.println("BEFORE comparing!");
//		for (Thread t : threads) {
//			String name = t.getName();
//			Thread.State state = t.getState();
//			int priority = t.getPriority();
//			String type = t.isDaemon() ? "Daemon" : "Normal";
//			System.out.printf("%-20s \t %s \t %d \t %s\n", name, state, priority, type);
//		}
		return LogModelComparisonNewL.compareModelAndLog(context, pn, ahl, log,true);
		//
		//		getProjectedFitnessAndPrecision(log,mnew);
//		threads = Thread.getAllStackTraces().keySet();
//		System.out.println("AFTER comparing!");
//		for (Thread t : threads) {
//			String name = t.getName();
//			Thread.State state = t.getState();
//			int priority = t.getPriority();
//			String type = t.isDaemon() ? "Daemon" : "Normal";
//			System.out.printf("%-20s \t %s \t %d \t %s\n", name, state, priority, type);
//		}
//		return getProjectedFitnessAndPrecision(log, mnew);
	}

	public static boolean getProjectedModel2ModelComparison(UIPluginContext context, IncrementalModelL m2,
			Set<IncrementalModelL> models) throws ProjectedMeasuresFailedException, ReductionFailedException {
		Set<Petrinet> nets = new HashSet<Petrinet>();
		AcceptingPetriNet net1 = new AcceptingPetriNetImpl(m2.getNet());

		for (IncrementalModelL m : models) {
			float[] fitPre = new float[2];
			ComparePetriNet2PetriNetPlugin compare = new ComparePetriNet2PetriNetPlugin();
			CompareParameters param = new CompareParameters();
			Set<String> act = new HashSet<String>();
			for (Transition t : m2.getNet().getTransitions())
				if (!t.isInvisible())
					act.add(t.getLabel());

			param.setK(act.size() > 1 ? 2 : 1);
			ProjectedRecallPrecisionResult result = null;
			try {
				result = compare.measure(net1, new AcceptingPetriNetImpl(m.getNet()), param, new ProMCanceller() {

					public boolean isCancelled() {
						// TODO Auto-generated method stub
						return false;
					}
				});
			} catch (AutomatonFailedException | InterruptedException | ProjectedMeasuresFailedException e) {
				e.printStackTrace();
			}
			try {
				if (result.getRecall() == 1. && result.getPrecision() == 1.) {
					if (net1.getNet().getEdges().size() < m.getNet().getEdges().size()) { //replace m with m2 if num of edges in m2 is less
						models.remove(m);
						models.add(m2);
					}
					return true;
				}
			} catch (NullPointerException n) {
				context.getProvidedObjectManager().createProvidedObject("net1", net1, AcceptingPetriNet.class, context);
				context.getProvidedObjectManager().createProvidedObject("net2", new AcceptingPetriNetImpl(m.getNet()),
						AcceptingPetriNet.class, context);
				System.out.println("Exported");
				//				context.getGlobalContext().getResourceManager().getResourceForInstance(net1).setFavorite(true);
			}

		}
		models.add(m2);
		return false;
	}
}
