package org.processmining.incrementaldiscovery;

import java.awt.GridBagConstraints;
import java.awt.GridBagLayout;
import java.util.ArrayList;
import java.util.Collections;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Map.Entry;
import java.util.Set;
import java.util.TreeMap;

import javax.swing.DefaultListModel;
import javax.swing.JButton;
import javax.swing.JComboBox;
import javax.swing.JPanel;
import javax.swing.ListModel;
import javax.swing.table.DefaultTableModel;

import org.deckfour.xes.model.XAttribute;
import org.deckfour.xes.model.XTrace;
import org.processmining.acceptingpetrinet.models.AcceptingPetriNetArray;
import org.processmining.acceptingpetrinet.models.impl.AcceptingPetriNetArrayFactory;
import org.processmining.framework.util.ui.widgets.ProMList;
import org.processmining.humbertogarciacaballero.models.controller.PNetVisualizerControllerL;
import org.processmining.incrementaldiscovery.models.IncrementalModelL;
import org.processmining.incrementaldiscovery.processconfiguration.GlobalAttributeMapL;
import org.processmining.models.graphbased.directed.petrinet.elements.Transition;
import org.processmining.plugins.InductiveMiner.efficienttree.EfficientTreeReduce.ReductionFailedException;
import org.processmining.plugins.petrinet.replayresult.PNRepResult;
import org.processmining.plugins.petrinet.replayresult.StepTypes;
import org.processmining.plugins.replayer.replayresult.SyncReplayResult;
import org.processmining.prabhakardixit.alignments.CombinationUtilsL;
import org.processmining.prabhakardixit.fcnets.models.InteractiveMatrixNetL;
import org.processmining.prabhakardixit.fcnets.models.InteractivePetriNetL;
import org.processmining.projectedrecallandprecision.result.ProjectedRecallPrecisionResult;
import org.processmining.projectedrecallandprecision.result.ProjectedRecallPrecisionResult.ProjectedMeasuresFailedException;

import com.fluxicon.slickerbox.factory.SlickerFactory;

public class DiscoveryPredictorL {

	//view names
	public final String selectview = "Select ";
	public final String alphabetical = "Alphabetical";
	public final String heuLtoR = "Heuristics L-to-R";

	private int currSelection;
	private PNetVisualizerControllerL controller;

	private JComboBox policyTypeBox; //updated from PredictorActions
	ProMList<ListModel> list;//used in PredictorActions
	DefaultListModel actModel;//used in PredictorActions

	private JButton addBtn; //used in Discovery Panel

	private Set<String> skipActs; //used in Discovery Panel
	private DefaultTableModel tableModel;//used in Discovery Panel
	List<IncrementalModelL> incM;
	
	private DefaultTableModel tableModelConfig;
	private String attributeChosen="";
	
	private float minFitness=0.5f;
	private float minPrecision=0.5f;
	private int fitWeight=100;
	private int preWeight=100;


	
	
	//TODO extend this to include more discovery algos
	public void discover(PNetVisualizerControllerL controller) {
		this.setController(controller);
		actModel = new DefaultListModel();
		list = new ProMList<ListModel>("", actModel);
		final Class[] classes = new Class[] { Integer.class, Float.class, Float.class};
		tableModel = new DefaultTableModel() {
			@Override
			public Class<?> getColumnClass(int columnIndex) {
				if (columnIndex < classes.length)
					return classes[columnIndex];
				return super.getColumnClass(columnIndex);
			}
		};
		
		final Class[] classes2 = new Class[] { Object.class, Integer.class, Float.class, Float.class, Float.class };
		setTableModelConfig(new DefaultTableModel() {
			@Override
			public Class<?> getColumnClass(int columnIndex) {
				if (columnIndex < classes2.length)
					return classes2[columnIndex];
				return super.getColumnClass(columnIndex);
			}
		});
		
		
		
		currSelection = -1;
		incM = new ArrayList<IncrementalModelL>();
		addBtn = SlickerFactory.instance().createButton("Use model and select next activity");
//		addBtn = new JButton("Use model & next act");
		skipActs = new HashSet<String>();
		initActModel();

		GridBagConstraints gbc = new GridBagConstraints();
		JPanel panel = controller.getStatisticsPanel();
		panel.removeAll();
		//		panel.add(DiscoveryUsingIM.getArray(controller.getActivity(), controller.getIPN(), panel.getWidth(),
		//				panel.getHeight()));

		PredictorActivityListPanelL pr = new PredictorActivityListPanelL(this);

		panel.setLayout(new GridBagLayout());
		gbc.gridx = 0;
		gbc.gridy = 0;
		DiscoveryPanelL p = new DiscoveryPanelL(this);
		panel.add(p);
		gbc.gridx = 1;
		panel.add(pr);

//		gbc.gridx = 2;
//
//		panel.add(new ProcessConfigurabilityL(this), gbc);
//		panel.revalidate();
//		panel.repaint();

	}

	private void initActModel() {
		for (int i = 0; i < controller.getActivityCombo().getModel().getSize(); i++) {
			String s = controller.getActivityCombo().getItemAt(i);
			actModel.addElement(s);
		}
	}

	public PNetVisualizerControllerL getController() {
		return controller;
	}

	public void setController(PNetVisualizerControllerL controller) {
		this.controller = controller;
	}

	//sets the next controller activity : based on some policy
	public void setControllerActivity(boolean skipCurrAct) {
		String currAct = controller.getActivity();
		List<String> activities;
		if (addBtn != null)
			switch (policyTypeBox.getSelectedItem().toString()) {
				case selectview :
					//do nothing
					addBtn.setEnabled(false);
					break;
				case alphabetical :
					//alphabetical
					addBtn.setEnabled(true);
					activities = new ArrayList<String>(controller.getIPN().getLogInfo().getActivities());
					Collections.sort(activities);
					setActivity(activities, skipCurrAct);
					break;
				case heuLtoR :
					addBtn.setEnabled(true);
					activities = new ArrayList<String>(controller.getIPN().getLogInfo().getOrderedBasedOnSeq());
					setActivity(activities, skipCurrAct);
					break;
			}

	}

	public void setActivity(List<String> activities, boolean skipCurrAct) {
		removeAcitivitiesPresentInNet(activities);
		for (String act : activities) {
			if (!skipActs.contains(act)) {
				controller.getActivityCombo().setSelectedItem(act);
				//				controller.setActivity(act);
				list.setSelectedIndex(actModel.indexOf(act));
				return;
			}
		}

	}

	public JComboBox getPolicyTypeBox() {
		return policyTypeBox;
	}

	public void setPolicyTypeBox(JComboBox policyTypeBox) {
		this.policyTypeBox = policyTypeBox;
	}

	private void removeAcitivitiesPresentInNet(List<String> activities) {
		//remove trans which are already present
		for (Transition t : controller.getIPN().getselectedPetriNetMap().getPetriNet().getTransitions()) {
			if (!t.isInvisible()) {
				activities.remove(t.getLabel());
			}
		}
	}

	public Set<String> getSkipActs() {
		return skipActs;
	}

	public void setSkipActs(Set<String> skipActs) {
		this.skipActs = skipActs;
	}

	public List<IncrementalModelL> getIncM() {
		return incM;
	}

	public void setIncM(List<IncrementalModelL> incM) {
		this.incM = incM;
	}

	public int getCurrSelection() {
		return currSelection;
	}

	public void setCurrSelection(int currSelection) {
		this.currSelection = currSelection;
	}

	public ProMList<ListModel> getList() {
		return list;
	}

	public void setList(ProMList<ListModel> list) {
		this.list = list;
	}

	public DefaultListModel getActModel() {
		return actModel;
	}

	public void setActModel(DefaultListModel actModel) {
		this.actModel = actModel;
	}

	public DefaultTableModel getTableModel() {
		return tableModel;
	}

	public void setTableModel(DefaultTableModel tableModel) {
		this.tableModel = tableModel;
	}

	public void addRelations(InteractivePetriNetL ipn) throws ProjectedMeasuresFailedException, ReductionFailedException {
		if (tableModel.getRowCount() > 0)
			for (int i = tableModel.getRowCount() - 1; i > -1; i--)
				tableModel.removeRow(i);
		IncrementalDiscoveryGuidedDFAlgorithmL algo = new IncrementalDiscoveryGuidedDFAlgorithmL();

		TreeMap<Float, IncrementalModelL> sortedMap = algo.generatePetriNetForSelectedActivity(ipn.getLog(),
				ipn.getUIContext(), ipn, 1, minFitness,minPrecision,fitWeight,preWeight, false);
		AcceptingPetriNetArray apnArr = AcceptingPetriNetArrayFactory.createAcceptingPetriNetArray();
		int count = 0;
		incM.clear();
		for (Float f : sortedMap.descendingMap().keySet()) {
			incM.add(sortedMap.get(f));
		}

		for (int i = 0; i < incM.size(); i++) {
			tableModel.addRow(new Object[] { i + 1, incM.get(i).getFitness(),
					incM.get(i).getPrecision() });
		}
	}

	public JButton getAddBtn() {
		return addBtn;
	}

	public DefaultTableModel getTableModelConfig() {
		return tableModelConfig;
	}

	public void setTableModelConfig(DefaultTableModel tableModelConfig) {
		this.tableModelConfig = tableModelConfig;
	}
	
	public void updateStats(String attKey) {
		//		attKey = "AMOUNT_REQ";
		//		for (int i = 0; i < tableModelConfig.getRowCount()+1; i++)
		//			tableModelConfig.removeRow(i);
		attKey = attributeChosen;
		tableModelConfig.getDataVector().removeAllElements();
		tableModelConfig.fireTableDataChanged();

		int defaultBucketSize = 6;
		Map<Object, GlobalAttributeMapL> valMap = new HashMap<Object, GlobalAttributeMapL>();

		InteractiveMatrixNetL imn = getController().getIPN().getInteractiveMatNet();
		if (currSelection != -1)
			imn = getIncM().get(currSelection).getImn();

		if (imn.getReplayReult() != null) {
			for (SyncReplayResult sync : imn.getReplayReult()) {
				int syncMoves = 0;
				int nonSyncMoves = 0;
				int actsCovered = 0;
				for (int actIndex = 0; actIndex < sync.getStepTypes().size(); actIndex++) {
					StepTypes type = sync.getStepTypes().get(actIndex);
					Object stepObj = sync.getNodeInstance().get(actIndex);
					switch (type) {
						case L :
							nonSyncMoves++;
							break;
						case LMGOOD :
							syncMoves++;
							actsCovered++;
							break;
						case MINVI :
							actsCovered++;
							break;
						case MREAL :
							nonSyncMoves++;
							break;
					}
				}

				float min = 999;
				float max = -1;
				for (int traceId : sync.getTraceIndex()) {
					XTrace t = getController().getIPN().getLog().get(traceId);
					for (Entry<String, XAttribute> s : t.getAttributes().entrySet()) {
						if (s.getKey().equals(attKey)) {
							if (s.getValue().toString().equals("10"))
								System.out.println("ere");
							if (!valMap.containsKey(s.getValue())) {
								valMap.put(s.getValue(), new GlobalAttributeMapL());

							}
							GlobalAttributeMapL gap = valMap.get(s.getValue());
							gap.setAllFitness(gap.getAllFitness() + sync.getInfo().get(PNRepResult.TRACEFITNESS));
							gap.setNumCases(gap.getNumCases() + 1);
							//							gap.setTotActsCovered(gap.getTotActsCovered() + actsCovered);
							//							gap.setTotFitActs(gap.getTotFitActs() + syncMoves);
							//							gap.setTotNonFitActs(nonSyncMoves + gap.getTotNonFitActs());
							gap.setNumCasesWithPerfectFitness(sync.getInfo().get(PNRepResult.TRACEFITNESS) == 1 ? gap
									.getNumCasesWithPerfectFitness() + 1 : gap.getNumCasesWithPerfectFitness());
							gap.setValue(s.getValue());

						}
					}
				}

			}
		} else {
			Map<Integer, String> allActs = new HashMap<Integer, String>();
			int i = 0;
			for (Transition t : imn.getselectedPetriNetMap().getPetriNet().getTransitions()) {
				if (!t.isInvisible())
					allActs.put(i++, t.getLabel());
			}

			Set<Set<Integer>> allCombs = CombinationUtilsL.getAllCombs(allActs.keySet(), 2);
			ProjectedRecallPrecisionResult result = imn.getProjectedResult();
			for (XTrace t : getController().getIPN().getLog()) {
				for (Entry<String, XAttribute> s : t.getAttributes().entrySet()) {
					if (s.getKey().equals(attKey)) {
						if (!valMap.containsKey(s.getValue())) {
							valMap.put(s.getValue(), new GlobalAttributeMapL());
						}
						double fitness = 0.;
						for (Set<Integer> comb : allCombs) {
							String[] actPair = new String[2];
							i = 0;
							for (int id : comb)
								actPair[i++] = allActs.get(id);
							try {
								fitness += result.getRecall(actPair);
							} catch (ProjectedMeasuresFailedException e) {
								// TODO Auto-generated catch block
								e.printStackTrace();
							}
						}

						GlobalAttributeMapL gap = valMap.get(s.getValue());
						gap.setAllFitness(gap.getAllFitness() + fitness / (allCombs.size()));
						gap.setNumCases(gap.getNumCases() + 1);
						//						gap.setTotActsCovered(gap.getTotActsCovered() + actsCovered);
						//						gap.setTotFitActs(gap.getTotFitActs() + syncMoves);
						//						gap.setTotNonFitActs(nonSyncMoves + gap.getTotNonFitActs());
						gap.setNumCasesWithPerfectFitness(fitness / (allCombs.size()) == 1 ? gap.getNumCasesWithPerfectFitness() + 1 : gap
								.getNumCasesWithPerfectFitness());
						gap.setValue(s.getValue());

					}
				}
			}
		}
		int numTrans = getController().getIPN().getselectedPetriNetMap().getPetriNet().getTransitions().size();

		//finally add - if more than 15, use buckets for int/floats 
		for (Object o : valMap.keySet()) {
			GlobalAttributeMapL gap = valMap.get(o);
			getTableModelConfig().addRow(new Object[] { o, new Integer(gap.getNumCases()),
					gap.getAllFitness() / gap.getNumCases(),
					//gap.getTotFitActs() / (gap.getTotFitActs() + gap.getTotNonFitActs()),
					//(gap.getTotActsCovered() / gap.getNumCases()) / numTrans });
					gap.getNumCasesWithPerfectFitness() });
		}

	}
	public String getAttributeChosen() {
		return attributeChosen;
	}

	public void setAttributeChosen(String attributeChosen) {
		this.attributeChosen = attributeChosen;
	}

	public float getMinPrecision() {
		return minPrecision;
	}

	public void setMinPrecision(float minPrecision) {
		this.minPrecision = minPrecision;
	}

	public float getMinFitness() {
		return minFitness;
	}

	public void setMinFitness(float minFitness) {
		this.minFitness = minFitness;
	}

	public int getFitWeight() {
		return fitWeight;
	}

	public void setFitWeight(int fitWeight) {
		this.fitWeight = fitWeight;
	}

	public int getPreWeight() {
		return preWeight;
	}

	public void setPreWeight(int preWeight) {
		this.preWeight = preWeight;
	}


}
