package org.processmining.incrementaldiscovery;

import java.util.Iterator;
import java.util.List;
import java.util.Set;

import org.processmining.humbertogarciacaballero.models.controller.PNetVisualizerControllerL;
import org.processmining.humbertogarciacaballero.models.gui.controllers.NodeHighlightControlPMDUtilsL;
import org.processmining.incrementaldiscovery.models.IncrementalModelL;
import org.processmining.prabhakardixit.alignments.UpdateProjectedAveragedAlignmentsL;
import org.processmining.reduction.ReductionTrackerL;

import prefuse.data.Node;
import prefuse.visual.EdgeItem;
import prefuse.visual.NodeItem;
import prefuse.visual.VisualItem;

public class ExpandUsingTrackerForVizAndNetL {

	/**
	 * @param controller
	 * @param m
	 * @param calcSynth
	 *            : decides if the addition is temporary or permanent -
	 *            calculate the SS or not
	 */
	public static void expandUsingTracker(PNetVisualizerControllerL controller, IncrementalModelL m, boolean calcSynth) {

		List<ReductionTrackerL> synthesis = m.getTrackSynthesis();
		controller.getMainView().getDisplay().requestFocus();
		int i = 0;
		try {
			synchronized (controller.getData()) {
				for (ReductionTrackerL rt : synthesis) {
					i++;
					switch (rt.getMethodName()) {
						case ABSTRACTION :
							addAbstraction(controller, rt, calcSynth, synthesis.size() == i);
							break;
						case LDPLACE :
							addLinearDependency(controller, rt, true, calcSynth, synthesis.size() == i);
							break;
						case LDTRAN :
							addLinearDependency(controller, rt, false, calcSynth, synthesis.size() == i);
							break;
						case RENAME :
							renameActivity(controller, rt, calcSynth);
							break;
						default :
							break;
					}
				}
			}
		} catch (IllegalArgumentException | IllegalStateException ex2) {
			System.out.println("DAMN!");
		}
	
		System.out.println(UpdateProjectedAveragedAlignmentsL.times);
	}

	private static void renameActivity(PNetVisualizerControllerL controller, ReductionTrackerL rt, boolean calcSynth) {
		int tranId = rt.getNewTranID();
		Iterator iter = controller.getVisualization().getVisualGroup("graph.nodes").tuples();
		NodeItem item = null;
		while (iter.hasNext()) {
			VisualItem n = (VisualItem) iter.next();
			if (n instanceof NodeItem && n.get("type") == "TRANSITION" && n.getInt("ID") == rt.getNewTranID()) {
				item = (NodeItem) n;
			}
		}
		if (item != null)
			NodeHighlightControlPMDUtilsL.renameActivity(item, controller.getMainView().getNodeHighlightControlPMD(),
					rt.isSilent() ? "" : rt.getTranName(), calcSynth);
	}

	/**
	 * @param controller
	 * @param rt
	 * @param parallePlace
	 *            -is new place
	 * @param runLayout
	 */
	private static void addLinearDependency(PNetVisualizerControllerL controller, ReductionTrackerL rt,
			boolean parallePlace, boolean calcSynth, boolean runLayout) {
		//		controller.setActivity(rt.getName());
		Iterator iter = controller.getVisualization().getVisualGroup("graph.nodes").tuples();
		//		TupleSet inputOutput = controller.getMainView().getNodeHighlightControlPMD().getInputOutputSet();
		Set<Integer> selectedInput = controller.getMainView().getNodeHighlightControlPMD().getSelectedInput();
		Set<Integer> selectedOutput = controller.getMainView().getNodeHighlightControlPMD().getSelectedOutput();
		selectedInput.clear();
		selectedOutput.clear();

		NodeItem outputNodeItem = null;
		while (iter.hasNext()) {
			VisualItem n = (VisualItem) iter.next();
			if (n instanceof NodeItem) {
				if (parallePlace) {
					if (n.get("type") == "TRANSITION") {
						//						if ((rt.getAtt1().contains(n.get("ID")) || rt.getAtt1().equals(n.get("ID")))
						//								&& (rt.getAtt2().contains(n.get("ID")) || rt.getAtt2().equals(n.get("ID"))))
						//							inputOutput.addTuple(n.getSourceTuple());
						//						else 
						if (rt.getAtt1().contains(n.get("ID")) || rt.getAtt1().equals(n.get("ID")))
							selectedInput.add((int) n.getInt("ID"));
						if ((rt.getAtt2().contains(n.get("ID")) || rt.getAtt2().equals(n.get("ID")))) {
							selectedOutput.add((int) n.getInt("ID"));
							outputNodeItem = (NodeItem) n;
						}
					}
				} else {
					if (n.get("type") == "PLACE") {
						if (rt.getAtt1().contains(n.get("ID")) || rt.getAtt1().equals(n.get("ID")))
							selectedInput.add((int) n.getInt("ID"));
						if ((rt.getAtt2().contains(n.get("ID")) || rt.getAtt2().equals(n.get("ID")))) {
							selectedOutput.add((int) n.getInt("ID"));
							outputNodeItem = (NodeItem) n;
						}
					}
				}

			}
		}

		//set new ids
		if (parallePlace)
			NodeHighlightControlPMDUtilsL.addLinearlyDependentNode(outputNodeItem, controller.getMainView()
					.getNodeHighlightControlPMD(), rt.getNewPlaceID(), rt.getTranName(), calcSynth, runLayout,!calcSynth);
		else
			NodeHighlightControlPMDUtilsL.addLinearlyDependentNode(outputNodeItem, controller.getMainView()
					.getNodeHighlightControlPMD(), rt.getNewTranID(), rt.isSilent() ? "" : rt.getTranName(), calcSynth,
					runLayout,!calcSynth);
	}

	private static void addAbstraction(PNetVisualizerControllerL controller, ReductionTrackerL rt, boolean calcSynth,
			boolean runLayout) {
		//		controller.setActivity(rt.getName());

		//get all edges from graph first

		Set<VisualItem> selectedEdges = controller.getMainView().getNodeHighlightControlPMD().getSelectedEdges();
		selectedEdges.clear();
		Iterator iter = controller.getVisualization().getVisualGroup("graph.edges").tuples();

		while (iter.hasNext()) {
			VisualItem ed = (VisualItem) iter.next();
			if (ed instanceof EdgeItem) {
				Node nSource = ((EdgeItem) ed).getSourceItem();
				Node nTarget = ((EdgeItem) ed).getTargetItem();
				if (nSource.get("type") == "TRANSITION") {
					if ((rt.getAtt1().contains(nSource.get("ID")) || rt.getAtt1().equals(nSource.get("ID")))
							&& (rt.getAtt2().contains(nTarget.get("ID")) || rt.getAtt2().equals(nTarget.get("ID"))))
						selectedEdges.add(ed);
				}

			}
		}

		//set new ids
		//		controller.getMainView().getNodeHighlightControlPMD().setNewPlaceId(rt.getNewPlaceID());
		//		controller.getMainView().getNodeHighlightControlPMD().setNewTramId(rt.getNewTranID());

		//raise event for adding tran in seq ('raise enter')
		NodeHighlightControlPMDUtilsL.addTransAndPlace(controller.getMainView().getNodeHighlightControlPMD(),
				rt.getNewPlaceID(), rt.getNewTranID(), rt.getTranName(), calcSynth, runLayout,!calcSynth);
		selectedEdges.clear();
	}
}
