package org.processmining.incrementaldiscovery;

import java.util.ArrayList;
import java.util.Collection;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Set;
import java.util.TreeMap;

import org.deckfour.xes.classification.XEventClass;
import org.deckfour.xes.extension.std.XConceptExtension;
import org.deckfour.xes.info.XLogInfo;
import org.deckfour.xes.info.XLogInfoFactory;
import org.deckfour.xes.model.XEvent;
import org.deckfour.xes.model.XLog;
import org.deckfour.xes.model.XTrace;
import org.processmining.acceptingpetrinet.models.AcceptingPetriNetArray;
import org.processmining.acceptingpetrinet.models.impl.AcceptingPetriNetArrayFactory;
import org.processmining.acceptingpetrinet.models.impl.AcceptingPetriNetImpl;
import org.processmining.contexts.uitopia.UIPluginContext;
import org.processmining.decomposelogs.AggregatedHeuristicsL;
import org.processmining.decomposelogs.Log2DeclareL;
import org.processmining.incrementaldiscovery.models.IncrementalModelL;
import org.processmining.models.graphbased.directed.petrinet.Petrinet;
import org.processmining.models.graphbased.directed.petrinet.PetrinetEdge;
import org.processmining.models.graphbased.directed.petrinet.PetrinetNode;
import org.processmining.models.graphbased.directed.petrinet.elements.Place;
import org.processmining.models.graphbased.directed.petrinet.elements.Transition;
import org.processmining.plugins.InductiveMiner.efficienttree.EfficientTreeReduce.ReductionFailedException;
import org.processmining.prabhakardixit.alignments.CombinationUtilsL;
import org.processmining.prabhakardixit.fcnets.expansion.ExpandUsingTrackerL;
import org.processmining.prabhakardixit.fcnets.matrix.MatrixUtilsL;
import org.processmining.prabhakardixit.fcnets.models.InteractiveMatrixNetL;
import org.processmining.prabhakardixit.fcnets.models.InteractivePetriNetL;
import org.processmining.prabhakardixit.fcnets.models.LDSetsL;
import org.processmining.prabhakardixit.fcnets.synthesis.CreateIPNL;
import org.processmining.prabhakardixit.utils.PetrinetUtilsL;
import org.processmining.projectedrecallandprecision.result.ProjectedRecallPrecisionResult.ProjectedMeasuresFailedException;

import com.google.common.collect.Lists;

public class IncrementalDiscoveryGuidedDFAlgorithmL {
	int k = 0;
	XLog log;
	UIPluginContext context;
	Set<String> addedActs = new HashSet<String>();
	boolean proj;
	boolean calcSynthSpace = true;
	Map<String, Integer> directlyFollows = new HashMap<String, Integer>();
	Map<String, Integer> directlyPrecedes = new HashMap<String, Integer>();
	Set<Integer> prePlaces = new HashSet<Integer>();
	Set<Integer> postPlaces = new HashSet<Integer>();
	Set<Integer> preTrans = new HashSet<Integer>();
	Set<Integer> postTrans = new HashSet<Integer>();
	Set<Integer> inputForPP = new HashSet<Integer>();
	Set<Integer> outputForPP = new HashSet<Integer>();

	public TreeMap<Float, IncrementalModelL> generatePetriNetForSelectedActivity(XLog log, UIPluginContext context,
			InteractivePetriNetL ipn, int nbrOfRep, float minFit, float minPre, float fitWeight, float preWeight,
			boolean projectAlignments) throws ProjectedMeasuresFailedException, ReductionFailedException {
		k = 0;
		this.log = log;
		this.context = context;
		proj = projectAlignments;

		String selectedAct = ipn.getSelectedActivity();

		//DO NOT calc synth space initially
		calcSynthSpace = false;

		//create activity set
		Set<String> activities = ipn.getLogInfo().getActivities();

		//get Aggregated Log Heuristics from ipn
		AggregatedHeuristicsL ahl = ipn.getAggregatedHeuristics();

		//create all possible combinations
		Set<IncrementalModelL> allModels = new HashSet<IncrementalModelL>();

		InteractiveMatrixNetL imn = ipn.getInteractiveMatNet();
		Set<IncrementalModelL> models = new HashSet<IncrementalModelL>();
		models.add(new IncrementalModelL(imn));
		boolean init = ipn.getselectedPetriNetMap().getPetriNet().getTransitions().size() < 3;
		int i = 0;
		long msecs = System.currentTimeMillis();

		for (Transition t : ipn.getselectedPetriNetMap().getPetriNet().getTransitions()) {
			if (!t.isInvisible())
				addedActs.add(t.getLabel());
		}
		if (init) { //init
			preTrans.add((int) 0);
			preTrans.add((int) 1);
			prePlaces.add((int) 0);
			prePlaces.add((int) 1);
			prePlaces.add((int) 2);
		} else
			getDirectlyFollows(log, selectedAct);
		addedActs.add(selectedAct);
		allModels.clear();

		for (IncrementalModelL m : models) {
			if (!init)
				usePlacesTrans(log, m);
			else
				init = false;
			allModels.addAll(createIncrementalModels(m.getImn(), ahl, selectedAct, nbrOfRep));
		}

		//update pn and imn by selecting the best one
		float fit = 0f;
		float pre = 0f;
		//		for (IncrementalModel m : allModels) {
		//			if (m.getFitness() + m.getPrecision() > fitPreScore) {
		//				fit = m.getFitness();
		//				pre = m.getPrecision();
		//				fitPreScore = fit + pre;
		//			}
		//		}
		//		models.clear();
		//		for (IncrementalModel m : allModels) {
		//			if (m.getFitness() + m.getPrecision() == fitPreScore) {
		//				Alignments.getProjectedModel2ModelComparison(context, m, models); //add to models-if similar model isnt already present, replace if the curr model is better than similar if possible
		//			}
		//
		//		}

		Set<IncrementalModelL> removeModels = new HashSet<IncrementalModelL>();
		for (IncrementalModelL m : allModels) {
			for (IncrementalModelL m2 : allModels) {
				if (!m.equals(m2) && !removeModels.contains(m)) {
					if (m.getFitness() == m2.getFitness() && m.getPrecision() == m2.getPrecision()) {
						if (m.getNet().getNodes().size() > m2.getNet().getNodes().size())
							removeModels.add(m);
						else {
							if (m.getNet().getNodes().size() == m2.getNet().getNodes().size()
									&& m.getNet().getEdges().size() > m2.getNet().getEdges().size()) {
								removeModels.add(m);
							} else {
								if (isNextTranSilent(m, selectedAct))
									removeModels.add(m);
								//								else
								//									removeModels.add(m2);

							}
						}
					}
				}
			}
		}
				allModels.removeAll(removeModels); //remove models with same fit and pre values
		TreeMap<Float, IncrementalModelL> sortedMap = new TreeMap<>();

		for (IncrementalModelL m : allModels) {
			if (m.getFitness() >= minFit && m.getPrecision() >= minPre) {
				int diffInNodes = m.getNet().getTransitions().size()
						- ipn.getselectedPetriNetMap().getPetriNet().getTransitions().size();
				float key = m.getFitness() * fitWeight + m.getPrecision() * preWeight + 0.01f * 1 / diffInNodes;
				if (sortedMap.containsKey(key)) {
					boolean add = false;

					while (!add) {
						if (sortedMap.containsKey(key))
							key = key - 0.01f;
						else
							add = true;
					}
					sortedMap.put(key, m);
				} else
					sortedMap.put(key, m);
			}
		}

		//		for (float f : sortedMap.keySet()) {
		//			if (sortedMap.get(f).getFitness() > 0.76f)
		//				if (fcnetsUtils.isFCNet(sortedMap.get(f).getNet())){
		//					System.out.println("Flaot : " + f + " FC ");
		//					context.getProvidedObjectManager().createProvidedObject(
		//							"Extracted Petrinet", sortedMap.get(f).getNet(), Petrinet.class, context);
		//				}
		//			//			else
		//			//				System.out.println("Flaot : " + f + " NON FC ");
		//		}

		//		Set<Float> removeKeys = new HashSet<Float>();
		//		int id = 0;
		//		for (float f : sortedMap.keySet()) {
		//			if (id++ > 5)
		//				removeKeys.add(f);
		//		}
		//		for (float f : removeKeys)
		//			sortedMap.remove(f);

		msecs -= System.currentTimeMillis();
		System.out.println("Time to add activity : " + selectedAct + msecs);
		msecs = System.currentTimeMillis();

		return sortedMap;
	}

	private boolean isNextTranSilent(IncrementalModelL m, String selectedAct) {
		for (Transition t : m.getNet().getTransitions()) {
			if (!t.isInvisible() && t.getLabel().equals(selectedAct)) {
				for (PetrinetEdge<? extends PetrinetNode, ? extends PetrinetNode> eOut : m.getNet().getOutEdges(t)) {
					if (m.getNet().getOutEdges(eOut.getTarget()).size() > 1)
						for (PetrinetEdge<? extends PetrinetNode, ? extends PetrinetNode> eOut2 : m.getNet()
								.getOutEdges(eOut.getTarget())) {
							if (((Transition) eOut2.getTarget()).isInvisible())
								return true;
						}
				}
			}
		}

		return false;
	}

	public AcceptingPetriNetArray generatePetriNet(XLog log, UIPluginContext context)
			throws ProjectedMeasuresFailedException, ReductionFailedException {
		k = 0;
		this.log = log;
		this.context = context;
		proj = true;

		calcSynthSpace = false;

		AggregatedHeuristicsL ahl = Log2DeclareL.getAllAggPairHeuristics(log);

		//init create ipn to generate the basic structure of the net
		InteractivePetriNetL ipn = CreateIPNL.createIPN(log, null, context, false, false, true, false);
		MatrixUtilsL.initiate(ipn);
		Petrinet pni = ipn.getselectedPetriNetMap().getPetriNet();

		//create activity set
		Set<String> activities = new HashSet<String>();
		XLogInfo logInfo = XLogInfoFactory.createLogInfo(log);
		for (XEventClass evt : logInfo.getEventClasses().getClasses()) {
			activities.add(evt.getId().replace("+complete", ""));
		}

		//create all possible combinations
		Set<IncrementalModelL> allModels = new HashSet<IncrementalModelL>();

		InteractiveMatrixNetL imn = ipn.getInteractiveMatNet();
		Set<IncrementalModelL> models = new HashSet<IncrementalModelL>();
		models.add(new IncrementalModelL(imn));
		boolean init = true;
		int i = 0;
		long msecs = System.currentTimeMillis();
		for (String a : activities) {
			//			if (a.equals("a") || a.equals("b") || a.equals("c") ) {
			//				for (int j = 0; j < 1; j++) {

			if (i++ < 25) {
				if (addedActs.isEmpty()) { //init
					preTrans.add((int) 0);
					preTrans.add((int) 1);
					prePlaces.add((int) 0);
					prePlaces.add((int) 1);
					prePlaces.add((int) 2);
				} else
					getDirectlyFollows(log, a);

				allModels.clear();
				addedActs.add(a);
				boolean changed = false;
				for (IncrementalModelL m : models) { //assume only ONE model is selected

					if (!init) {
						usePlacesTrans(log, m);
						if (!changed) {
							ExpandUsingTrackerL.expanUsingTracker(Lists.reverse(m.getTrackSynthesis()), ipn);
							changed = true;
						}
					} else
						init = false;
					allModels.addAll(createIncrementalModels(ipn.getInteractiveMatNet(), ahl, a, 1));
				}

				//update pn and imn by selecting the best one
				float fit = 0f;
				float pre = 0f;
				float fitPreScore = 0f;
				for (IncrementalModelL m : allModels) {
					if (m.getFitness() + m.getPrecision() > fitPreScore) {
						fit = m.getFitness();
						pre = m.getPrecision();
						fitPreScore = fit + pre;
					}
				}
				models.clear();
				for (IncrementalModelL m : allModels) {
					if (m.getFitness() + m.getPrecision() == fitPreScore) {
						AlignmentsL.getProjectedModel2ModelComparison(context, m, models); //add to models-if similar model isnt already present, replace if the curr model is better than similar if possible
					}

				}
				msecs -= System.currentTimeMillis();
				System.out.println("Time to add activity : " + a + msecs);
				msecs = System.currentTimeMillis();
				//				}
				//			}
			}
		}
		float fit = 0f;
		float pre = 0f;
		Set<IncrementalModelL> keep = new HashSet<IncrementalModelL>();

		//		allModels.retainAll(keep);
		AcceptingPetriNetArray apn = AcceptingPetriNetArrayFactory.createAcceptingPetriNetArray();

		for (IncrementalModelL m : models) {
			if (i++ % 1 == 0)
				apn.addNet(new AcceptingPetriNetImpl(m.getNet()));
		}

		return apn;
	}

	private Set<IncrementalModelL> createIncrementalModels(InteractiveMatrixNetL interactiveMatNet,
			AggregatedHeuristicsL ahl, String label, int maxNumOfIter) throws ProjectedMeasuresFailedException {
		List<Set<IncrementalModelL>> models = new ArrayList<Set<IncrementalModelL>>(); //one set for each iter
		//first create the obj for each iter
		for (int i = 0; i < maxNumOfIter + 1; i++) {
			models.add(new HashSet<IncrementalModelL>());
		}
		models.get(0).add(new IncrementalModelL(interactiveMatNet));

		for (int i = 0; i < maxNumOfIter; i++) {
			Set<IncrementalModelL> nextModels = models.get(i + 1);
			for (IncrementalModelL m : models.get(i)) {
				//				if(m.getImn().getInputOutPlaces().size() !=4 && i!=0)
				//					continue;
				//add all sequences
				nextModels.addAll(getAllSequenceModels(m, ahl, label));

				//add all choices
				nextModels.addAll(getAllChoiceModels(m, ahl, label));

				//add all parallel places (incl the seq tran adding the new place)
				nextModels.addAll(getAllParallelModels(m, ahl, label, log));
			}
		}
		Set<IncrementalModelL> allModels = new HashSet<IncrementalModelL>();
		for (int i = 1; i < maxNumOfIter + 1; i++)
			allModels.addAll(models.get(i));

		return allModels;
	}

	private Collection<? extends IncrementalModelL> getAllParallelModels(IncrementalModelL m,
			AggregatedHeuristicsL ahl, String label, XLog log) throws ProjectedMeasuresFailedException {
		Set<IncrementalModelL> allModels = new HashSet<IncrementalModelL>();

		for (LDSetsL lds : m.getImn().getplaceLDNonSiphons()) {
			//						if ((!Collections.disjoint(lds.getPositiveNodes(m.getImn()), inputForPP) && !Collections.disjoint(
			//								lds.getNegativeNodes(m.getImn()), outputForPP))) {
			//			if ((inputForPP.containsAll(lds.getPositiveNodes(m.getImn())) && outputForPP.containsAll(lds
			//					.getNegativeNodes(m.getImn())))) {
			if (inputForPP.containsAll(lds.getPositiveNodes(m.getImn()))) {
				Set<Integer> startTranSet = new HashSet<Integer>();
				Set<Integer> endTranSet = new HashSet<Integer>();
				startTranSet.addAll(lds.getPositiveNodes(m.getImn()));
				endTranSet.addAll(lds.getNegativeNodes(m.getImn()));
				IncrementalModelL mnew = new IncrementalModelL(m);
				IncrementalDiscoveryUtilsL
						.addParallelPlaceAndSeq(mnew, startTranSet, endTranSet, label, calcSynthSpace);
				allModels.add(mnew);
				//				context.getProvidedObjectManager().createProvidedObject("Petrinet" + " ", mnew.getNet(),
				//						Petrinet.class, context);
				//			context.getProvidedObjectManager().createProvidedObject("Petrinet" + ac, apn, AcceptingPetriNet.class,
				//					context);
				//alignments
				float[] fitPre = new float[2];
				//				if (proj)
				//				fitPre = Alignments.getProjectedFitnessAndPrecision(
				//						ProjectedAveragedAlignments.createMergedLog(log, addedActs, false), mnew);
				//				else
				//					fitPre = Alignments.getTraditionalFitnessAndPrecision(context,
				//							ProjectedAveragedAlignments.createMergedLog(log, addedActs, false), mnew);
				fitPre = AlignmentsL.getProjectedFitnessAndPrecisionUsingPatterns(context, ahl, mnew, log);

				mnew.setFitness(fitPre[0]);
				mnew.setPrecision(fitPre[1]);
				//				System.out.println(k++);

				addSkippableTrans(m, mnew, ahl, allModels); //adds new models with silent trans (loop and skip)
			}
		}
		return allModels;
	}

	private Collection<? extends IncrementalModelL> getAllChoiceModels(IncrementalModelL m, AggregatedHeuristicsL ahl,
			String label) throws ProjectedMeasuresFailedException {
		Set<IncrementalModelL> allModels = new HashSet<IncrementalModelL>();
		Petrinet pn = m.getNet();

		for (LDSetsL lds : m.getImn().getAllTranLDFCs()) {
			if (lds.getAllNodes(m.getImn()).contains(m.getImn().getPlaceId(PetrinetUtilsL.getInitialPlace(pn)))
					|| lds.getAllNodes(m.getImn()).contains(m.getImn().getPlaceId(PetrinetUtilsL.getFinalPlace(pn))))
				continue;
			if (!java.util.Collections.disjoint(lds.getNegativeNodes(m.getImn()), prePlaces)
					&& !java.util.Collections.disjoint(lds.getPositiveNodes(m.getImn()), postPlaces)) {
				Set<Integer> startPlaceSet = new HashSet<Integer>();
				Set<Integer> endPlaceSet = new HashSet<Integer>();
				startPlaceSet.addAll(lds.getNegativeNodes(m.getImn()));
				endPlaceSet.addAll(lds.getPositiveNodes(m.getImn()));
				IncrementalModelL mnew = new IncrementalModelL(m);
				IncrementalDiscoveryUtilsL.addChoiceTransition(mnew, startPlaceSet, endPlaceSet, label, calcSynthSpace);
				allModels.add(mnew);
				//alignments
				float[] fitPre = new float[2];
				//				if (proj)
				//				fitPre = Alignments.getProjectedFitnessAndPrecision(
				//						ProjectedAveragedAlignments.createMergedLog(log, addedActs, false), mnew);
				//				else
				//					fitPre = Alignments.getTraditionalFitnessAndPrecision(context,
				//							ProjectedAveragedAlignments.createMergedLog(log, addedActs, false), mnew);

				fitPre = AlignmentsL.getProjectedFitnessAndPrecisionUsingPatterns(context, ahl, mnew, log);
				mnew.setFitness(fitPre[0]);
				mnew.setPrecision(fitPre[1]);

				//				System.out.println(k++);
			}
		}
		return allModels;
	}

	private Collection<? extends IncrementalModelL> getAllSequenceModels(IncrementalModelL m,
			AggregatedHeuristicsL ahl, String label) throws ProjectedMeasuresFailedException {
		Set<IncrementalModelL> allModels = new HashSet<IncrementalModelL>();
		Petrinet pn = m.getNet();
		for (Transition t : pn.getTransitions()) {
			if (!preTrans.contains(m.getImn().getTranId(t)))
				continue;
			Set<Integer> outPlaces = new HashSet<Integer>();
			for (PetrinetEdge<? extends PetrinetNode, ? extends PetrinetNode> e : pn.getOutEdges(t))
				if (e.getTarget() != PetrinetUtilsL.getFinalPlace(pn)) {
					outPlaces.add(m.getImn().getPlaceId((Place) e.getTarget()));
				}

			for (int i = 1; i < outPlaces.size() + 1; i++) {
				Set<Set<Integer>> allComb = CombinationUtilsL.getAllCombs(outPlaces, i);
				for (Set<Integer> placeId : allComb) {
					Set<Integer> tranId = new HashSet<Integer>();
					tranId.add(m.getImn().getTranId(t));
					//add new tran between this tran and the outgoing place 
					IncrementalModelL mnew = new IncrementalModelL(m);
					IncrementalDiscoveryUtilsL.addNewTransition(mnew, tranId, placeId, label, calcSynthSpace);
					allModels.add(mnew);
					//alignments
					float[] fitPre = new float[2];
					//					if (proj)
					//					fitPre = Alignments.getProjectedFitnessAndPrecision(
					//							ProjectedAveragedAlignments.createMergedLog(log, addedActs, false), mnew);
					//					else
					//						fitPre = Alignments.getTraditionalFitnessAndPrecision(context,
					//								ProjectedAveragedAlignments.createMergedLog(log, addedActs, false), mnew);
					fitPre = AlignmentsL.getProjectedFitnessAndPrecisionUsingPatterns(context, ahl, mnew, log);
					mnew.setFitness(fitPre[0]);
					mnew.setPrecision(fitPre[1]);
					//					System.out.println(k++);

					addSkippableTrans(m, mnew, ahl, allModels); //adds new models with silent trans (loop and skip)
				}
			}
		}

		//for places with more than one incoming arcs, add trans too on the combinations
		for (Place p : pn.getPlaces()) {
			boolean containsTrans = false;
			Set<Integer> inTransitions = new HashSet<Integer>();
			for (PetrinetEdge<? extends PetrinetNode, ? extends PetrinetNode> e : pn.getInEdges(p))
				if (e.getTarget() != PetrinetUtilsL.getFinalPlace(pn)) {
					inTransitions.add(m.getImn().getTranId((Transition) e.getSource()));
					if (preTrans.contains(m.getImn().getTranId((Transition) e.getSource())))
						containsTrans = true;
					else
						containsTrans = false;
				}
			if (containsTrans)
				for (int i = 2; i < inTransitions.size() + 1; i++) {
					Set<Set<Integer>> allComb = CombinationUtilsL.getAllCombs(inTransitions, i);
					for (Set<Integer> tranId : allComb) {
						Set<Integer> placeId = new HashSet<Integer>();
						placeId.add(m.getImn().getPlaceId(p));
						IncrementalModelL mnew = new IncrementalModelL(m);
						IncrementalDiscoveryUtilsL.addNewTransition(mnew, tranId, placeId, label, calcSynthSpace);
						allModels.add(mnew);
						//alignments
						float[] fitPre = new float[2];
						//						if (proj)
						//						fitPre = Alignments.getProjectedFitnessAndPrecision(
						//								ProjectedAveragedAlignments.createMergedLog(log, addedActs, false), mnew);
						//						else
						//							fitPre = Alignments.getTraditionalFitnessAndPrecision(context,
						//									ProjectedAveragedAlignments.createMergedLog(log, addedActs, false), mnew);
						fitPre = AlignmentsL.getProjectedFitnessAndPrecisionUsingPatterns(context, ahl, mnew, log);
						mnew.setFitness(fitPre[0]);
						mnew.setPrecision(fitPre[1]);
						//						System.out.println(k++);

						addSkippableTrans(m, mnew, ahl, allModels); //adds new models with silent trans (loop and skip)
					}
				}
		}

		//finally, add multi tran to multi places combinations
		//		System.out.println();
		Set<Set<Transition>> commonOutTrans = new HashSet<Set<Transition>>();
		Set<Transition> evaluatedTrans = new HashSet<Transition>();
		for (Transition t : pn.getTransitions()) {
			if (pn.getOutEdges(t).size() > 1) {
				if (evaluatedTrans.contains(t))
					continue;
				Set<Place> outPlaces = new HashSet<Place>();
				for (PetrinetEdge<? extends PetrinetNode, ? extends PetrinetNode> e : pn.getOutEdges(t))
					outPlaces.add((Place) e.getTarget());
				Set<Transition> commonOutTran = new HashSet<Transition>();
				commonOutTran.add(t);
				//check if any other tran has the SAME outputs
				for (Transition t2 : pn.getTransitions()) {
					if (!t.equals(t2) && pn.getOutEdges(t).size() == pn.getOutEdges(t2).size()) {
						boolean sameOp = true;
						for (PetrinetEdge<? extends PetrinetNode, ? extends PetrinetNode> e : pn.getOutEdges(t2))
							if (!outPlaces.contains((Place) e.getTarget()))
								sameOp = false;
						if (sameOp)
							commonOutTran.add(t2);

					}
				}
				if (commonOutTran.size() > 1) {
					commonOutTrans.add(commonOutTran);
					evaluatedTrans.addAll(commonOutTran);
				}
			}
		}
		for (Set<Transition> set : commonOutTrans) {
			Set<Integer> tranIds = new HashSet<Integer>();
			Set<Integer> tempPreTrans = new HashSet<Integer>(preTrans);
			Transition tTemp = null;
			for (Transition t : set) {
				tranIds.add(m.getImn().getTranId(t));
				tTemp = t;
				if (t.isInvisible())
					tempPreTrans.add(m.getImn().getTranId(t));
			}

			//get place ids
			Set<Integer> placeId = new HashSet<Integer>();
			for (PetrinetEdge<? extends PetrinetNode, ? extends PetrinetNode> e : pn.getOutEdges(tTemp)) {
				placeId.add(m.getImn().getPlaceId((Place) e.getTarget()));
			}
			if (tempPreTrans.containsAll(tranIds)) {
				for (int i = 2; i < tranIds.size() + 1; i++) {
					Set<Set<Integer>> allComb = CombinationUtilsL.getAllCombs(tranIds, i);
					for (Set<Integer> tranId : allComb) {
						IncrementalModelL mnew = new IncrementalModelL(m);
						IncrementalDiscoveryUtilsL.addNewTransition(mnew, tranId, placeId, label, calcSynthSpace);
						allModels.add(mnew);
						//alignments
						float[] fitPre = new float[2];
						//						if (proj)
						//						fitPre = Alignments.getProjectedFitnessAndPrecision(
						//								ProjectedAveragedAlignments.createMergedLog(log, addedActs, false), mnew);
						//						else
						//							fitPre = Alignments.getTraditionalFitnessAndPrecision(context,
						//									ProjectedAveragedAlignments.createMergedLog(log, addedActs, false), mnew);
						fitPre = AlignmentsL.getProjectedFitnessAndPrecisionUsingPatterns(context, ahl, mnew, log);
						mnew.setFitness(fitPre[0]);
						mnew.setPrecision(fitPre[1]);
						//						System.out.println(k++);

						addSkippableTrans(m, mnew, ahl, allModels); //adds new models with silent trans (loop and skip)
					}
				}
			}

		}

		return allModels;
	}

	/**
	 * @param m
	 *            - old incremental model
	 * @param mnew
	 *            - new incremental model
	 * @param allModels
	 *            - newly created models will be added to this set
	 * @throws ProjectedMeasuresFailedException
	 */
	private void addSkippableTrans(IncrementalModelL m, IncrementalModelL mnew, AggregatedHeuristicsL ahl,
			Set<IncrementalModelL> allModels) throws ProjectedMeasuresFailedException {
		//add a skippable transition for the corresponding seq tran (both choice and loop)
		IncrementalModelL mnew1 = new IncrementalModelL(mnew);
		IncrementalModelL mnew2 = new IncrementalModelL(mnew);
		//get the new tran id
		int newTranId = -1;
		for (Transition t1 : mnew.getImn().getAllTranId().keySet()) { //search the newly added tran- use this to add new silent tran
			boolean present = false;
			for (Transition t2 : m.getImn().getAllTranId().keySet()) {
				if (m.getImn().getTranId(t2).equals(mnew.getImn().getTranId(t1)))
					present = true;
			}
			if (!present && !t1.equals(PetrinetUtilsL.getFinalTranistion(mnew.getNet()))) //if not present and not the final tran
				newTranId = mnew.getImn().getTranId(t1);
		}
		IncrementalDiscoveryUtilsL.addSilentTransitions(mnew1, mnew2, newTranId, calcSynthSpace);
		float[] fitPre = new float[2];
		//		if (proj)
		//			fitPre = Alignments.getProjectedFitnessAndPrecision(
		//					ProjectedAveragedAlignments.createMergedLog(log, addedActs, false), mnew1);
		//		else
		//			fitPre = Alignments.getTraditionalFitnessAndPrecision(context,
		//					ProjectedAveragedAlignments.createMergedLog(log, addedActs, false), mnew1);
		//		fitPre = Alignments.getProjectedFitnessAndPrecision(
		//				ProjectedAveragedAlignments.createMergedLog(log, addedActs, false), mnew1);
		fitPre = AlignmentsL.getProjectedFitnessAndPrecisionUsingPatterns(context, ahl, mnew1, log);

		mnew1.setFitness(fitPre[0]);
		mnew1.setPrecision(fitPre[1]);
		allModels.add(mnew1);
		//		if (proj)
		//		fitPre = Alignments.getProjectedFitnessAndPrecision(
		//				ProjectedAveragedAlignments.createMergedLog(log, addedActs, false), mnew2);
		//		else
		//			fitPre = Alignments.getTraditionalFitnessAndPrecision(context,
		//					ProjectedAveragedAlignments.createMergedLog(log, addedActs, false), mnew2);
		//		System.out.println(k++);
		//		System.out.println(k++);
		fitPre = AlignmentsL.getProjectedFitnessAndPrecisionUsingPatterns(context, ahl, mnew2, log);
		mnew2.setFitness(fitPre[0]);
		mnew2.setPrecision(fitPre[1]);
		allModels.add(mnew2);
	}

	private void usePlacesTrans(XLog log, IncrementalModelL m) {
		prePlaces.clear();
		postPlaces.clear();
		preTrans.clear();
		postTrans.clear();
		inputForPP.clear();
		outputForPP.clear();
		Set<Integer> silentPostTrans = new HashSet<Integer>(); //incoming silents of post trans
		Set<Integer> silentPreTrans = new HashSet<Integer>(); //outgoing silents of pre trans
		for (Transition t : m.getNet().getTransitions())
			if (!t.isInvisible()) {
				if (directlyFollows.keySet().contains(t.getLabel())) {
					postTrans.add(m.getImn().getTranId(t));
					updateAllSilentPostTrans(t, m, silentPostTrans,
							new HashSet<PetrinetEdge<PetrinetNode, PetrinetNode>>());
					for (PetrinetEdge<? extends PetrinetNode, ? extends PetrinetNode> e : m.getNet().getOutEdges(t)) {
						postPlaces.add(m.getImn().getPlaceId((Place) e.getTarget()));
						for (PetrinetEdge<? extends PetrinetNode, ? extends PetrinetNode> e2 : m.getNet().getOutEdges(
								e.getTarget())) {
							postTrans.add(m.getImn().getTranId((Transition) e2.getTarget()));
						}
					}
				}
				if (directlyPrecedes.keySet().contains(t.getLabel())) {
					preTrans.add(m.getImn().getTranId(t));
					//add all the following subsequent transitions (which are silent as pre trans too)
					updateAllSilentPreTrans(t, m, silentPreTrans,
							new HashSet<PetrinetEdge<PetrinetNode, PetrinetNode>>());
					for (PetrinetEdge<? extends PetrinetNode, ? extends PetrinetNode> e : m.getNet().getInEdges(t)) {
						prePlaces.add(m.getImn().getPlaceId((Place) e.getSource()));
						for (PetrinetEdge<? extends PetrinetNode, ? extends PetrinetNode> e2 : m.getNet().getInEdges(
								e.getSource())) {
							postTrans.add(m.getImn().getTranId((Transition) e2.getSource()));
						}
					}
				}

			}

		Set<Integer> temp = new HashSet<Integer>(prePlaces);
		prePlaces.removeAll(postPlaces);
		postPlaces.removeAll(temp);
		inputForPP = new HashSet<Integer>(preTrans);
		//		inputForPP.removeAll(postTrans);
		outputForPP = new HashSet<Integer>(postTrans);
		//		outputForPP.removeAll(preTrans);

		if (preTrans.isEmpty()) {
			preTrans.add((int) 1);
			preTrans.add((int) 0);
		}

		//		if (inputForPP.isEmpty()) {
		//			for (int i : preTrans) {
		//				Transition t = m.getImn().getTranBasedOnID(i);
		//				for (PetrinetEdge<? extends PetrinetNode, ? extends PetrinetNode> e : m.getNet().getInEdges(t)) {
		//					for (PetrinetEdge<? extends PetrinetNode, ? extends PetrinetNode> e2 : m.getNet().getInEdges(
		//							e.getSource())) {
		//						inputForPP.add(m.getImn().getTranId((Transition) e2.getSource()));
		//					}
		//				}
		//			}
		//		}
		//		//		if (outputForPP.isEmpty()) {
		//		for (int i : postTrans) {
		//			Transition t = m.getImn().getTranBasedOnID(i);
		//			for (PetrinetEdge<? extends PetrinetNode, ? extends PetrinetNode> e : m.getNet().getOutEdges(t)) {
		//				for (PetrinetEdge<? extends PetrinetNode, ? extends PetrinetNode> e2 : m.getNet().getOutEdges(
		//						e.getTarget())) {
		//					outputForPP.add(m.getImn().getTranId((Transition) e2.getTarget()));
		//				}
		//			}
		//			//			}
		//		}
		//
		//		temp = new HashSet<Integer>(preTrans);
		//		temp.retainAll(postTrans);
		//		inputForPP.removeAll(temp);
		//		outputForPP.removeAll(temp);
		//		inputForPP.addAll(silentPreTrans);
		//		outputForPP.addAll(silentPostTrans);
		//
		//		//TODO - fix this - temp : finally add all the silent trans as input and output 
		for (Transition t : m.getNet().getTransitions())
			if (t.isInvisible()) {
				outputForPP.add(m.getImn().getTranId(t));
				inputForPP.add(m.getImn().getTranId(t));
				preTrans.add(m.getImn().getTranId(t));
				postTrans.add(m.getImn().getTranId(t));
			}
		//
		//		preTrans.addAll(silentPreTrans);

		//TEMP STARTS
//		for (Transition t : m.getNet().getTransitions()) {
//			outputForPP.add(m.getImn().getTranId(t));
//			inputForPP.add(m.getImn().getTranId(t));
//			preTrans.add(m.getImn().getTranId(t));
//			postTrans.add(m.getImn().getTranId(t));
//		}

//		for (Place p : m.getNet().getPlaces()) {
//			prePlaces.add(m.getImn().getPlaceId(p));
//			postPlaces.add(m.getImn().getPlaceId(p));
//		}
		//TEMP ENDS

	}

	/**
	 * Recursively add all the incoming trans of postTran which are silent to
	 * postTran
	 * 
	 * @param t
	 * @param m
	 * @param silentPostTrans
	 */
	private void updateAllSilentPostTrans(Transition t, IncrementalModelL m, Set<Integer> silentPostTrans,
			Set<PetrinetEdge<PetrinetNode, PetrinetNode>> edges) {
		Petrinet pn = m.getNet();
		for (PetrinetEdge<? extends PetrinetNode, ? extends PetrinetNode> e : pn.getInEdges(t)) {
			postPlaces.add(m.getImn().getPlaceId((Place) e.getSource()));
			for (PetrinetEdge<? extends PetrinetNode, ? extends PetrinetNode> e2 : pn.getInEdges(e.getSource())) {
				if (((Transition) e2.getSource()).isInvisible()) {
					if (!postTrans.contains(m.getImn().getTranId((Transition) e2.getSource()))
							&& edges.add((PetrinetEdge<PetrinetNode, PetrinetNode>) e2)) {
						postTrans.add(m.getImn().getTranId((Transition) e2.getSource()));
						silentPostTrans.add(m.getImn().getTranId((Transition) e2.getSource()));
						updateAllSilentPostTrans((Transition) e2.getSource(), m, silentPostTrans, edges);
					}
				}
			}
		}
	}

	/**
	 * Recursively add all the outgoing trans of preTran which are silent to
	 * preTran
	 * 
	 * @param t
	 * @param m
	 * @param silentPreTrans
	 */
	private void updateAllSilentPreTrans(Transition t, IncrementalModelL m, Set<Integer> silentPreTrans,
			Set<PetrinetEdge<PetrinetNode, PetrinetNode>> edges) {
		Petrinet pn = m.getNet();
		for (PetrinetEdge<? extends PetrinetNode, ? extends PetrinetNode> e : pn.getOutEdges(t)) {
			prePlaces.add(m.getImn().getPlaceId((Place) e.getTarget()));
			for (PetrinetEdge<? extends PetrinetNode, ? extends PetrinetNode> e2 : pn.getOutEdges(e.getTarget())) {
				if (((Transition) e2.getTarget()).isInvisible()) {
					if (!preTrans.contains(m.getImn().getTranId((Transition) e2.getTarget()))
							&& !postTrans.contains(m.getImn().getTranId((Transition) e2.getTarget()))
							&& edges.add((PetrinetEdge<PetrinetNode, PetrinetNode>) e2)) {
						preTrans.add(m.getImn().getTranId((Transition) e2.getTarget()));
						silentPreTrans.add(m.getImn().getTranId((Transition) e2.getTarget()));
						updateAllSilentPreTrans((Transition) e2.getTarget(), m, silentPreTrans, edges);
					}
				}
			}
		}
	}

	private void getDirectlyFollows(XLog log, String selectedAct) {
		int maxCount = 0;
		int minCount = 100;
		int numCases = 0;

		directlyFollows.clear();
		directlyPrecedes.clear();

		for (XTrace trace : log) {
			int count = 0;
			String prevEvt = "";
			//			String nextEvt = "";
			boolean addNext = false;
			//			boolean actPresent = false;
			for (XEvent event : trace) {
				String currEvt = XConceptExtension.instance().extractName(event);
				if (!addedActs.contains(currEvt) && !selectedAct.equals(currEvt))
					continue;
				if (currEvt.equals(selectedAct)) {
					count++;
					//					actPresent = true;
					addNext = true;
				}
				if (!selectedAct.equals(currEvt)) {
					if (addNext) {
						addNext = false;
						int val = directlyFollows.containsKey(currEvt) ? directlyFollows.get(currEvt) + 1 : 1;
						directlyFollows.put(currEvt, val);
					}

				}
				if (currEvt.equals(selectedAct)) { //TODO add anyway for eventually follows
					int val = directlyPrecedes.containsKey(prevEvt) ? directlyPrecedes.get(prevEvt) + 1 : 1;
					directlyPrecedes.put(prevEvt, val);
				}

				prevEvt = currEvt;
			}

			maxCount = maxCount > count ? maxCount : count;
			minCount = minCount < count ? minCount : count;

		}

		//remove noisy relations
		Set<String> remove = new HashSet<String>();
		int maxVal = 0;
		for (String s : directlyPrecedes.keySet())
			if (maxVal < directlyPrecedes.get(s))
				maxVal = directlyPrecedes.get(s);

		for (String s : directlyPrecedes.keySet()) {
			if (directlyPrecedes.get(s) < 0.15 * maxVal) {
				remove.add(s);
			}
		}
		
		//Threshld check invalidated for now
		for (String s : remove)
			directlyPrecedes.remove(s);

		remove.clear();
		maxVal = 0;
		for (String s : directlyFollows.keySet())
			if (maxVal < directlyFollows.get(s))
				maxVal = directlyFollows.get(s);
		for (String s : directlyFollows.keySet())
			if (directlyFollows.get(s) < 0.15 * maxVal) {
				remove.add(s);
			}
		for (String s : remove)
			directlyFollows.remove(s);

		//		controller.getIPN().getLogInfo().setMaxOcc(maxCount);
		//		controller.getIPN().getLogInfo().setMinOcc(minCount);
		//		controller.getIPN().getLogInfo().setCasePerct((float) numCases / log.size() * 100);
	}

}
