package org.processmining.incrementaldiscovery;

import java.util.HashSet;
import java.util.Set;

import org.processmining.incrementaldiscovery.models.IncrementalModelL;
import org.processmining.models.graphbased.directed.petrinet.Petrinet;
import org.processmining.models.graphbased.directed.petrinet.PetrinetEdge;
import org.processmining.models.graphbased.directed.petrinet.PetrinetNode;
import org.processmining.models.graphbased.directed.petrinet.elements.Place;
import org.processmining.models.graphbased.directed.petrinet.elements.Transition;
import org.processmining.prabhakardixit.fcnets.expansion.AddNewTranBwTranAndPlaceL;
import org.processmining.prabhakardixit.fcnets.expansion.ChoiceTransitionL;
import org.processmining.prabhakardixit.fcnets.expansion.ParallelPlaceL;
import org.processmining.prabhakardixit.utils.PetrinetUtilsL;
import org.processmining.reduction.ReductionTrackerL;
import org.processmining.reduction.ReductionTrackerL.Rule;

public class IncrementalDiscoveryUtilsL {
	public static void addNewTransition(IncrementalModelL m, int t, int p, String newLabel, int newPlaceId,
			int newTranId, boolean calcSynthSpace) {
		Set<Transition> tranSet = new HashSet<Transition>();
		Set<Place> placeSet = new HashSet<Place>();
		tranSet.add(m.getImn().getTranBasedOnID(t));
		placeSet.add(m.getImn().getPlaceBasedOnID(p));
		addNewTransition(m, tranSet, placeSet, newLabel, newLabel == "", calcSynthSpace);
	}

	public static void addNewTransition(IncrementalModelL m, Set<Integer> tSet, Set<Integer> pSet, String newLabel,
			boolean calcSynthSpace) {
		Set<Transition> tranSet = new HashSet<Transition>();
		Set<Place> placeSet = new HashSet<Place>();
		for (int t : tSet)
			tranSet.add(m.getImn().getTranBasedOnID(t));
		for (int p : pSet)
			placeSet.add(m.getImn().getPlaceBasedOnID(p));
		addNewTransition(m, tranSet, placeSet, newLabel, newLabel == "", calcSynthSpace);
	}

	public static Transition addNewTransition(IncrementalModelL m, Set<Transition> tranSet, Set<Place> placeSet,
			String newLabel, boolean isSilent, boolean calcSynthSpace) {
		Petrinet pn = m.getNet();
		Transition tNew = pn.addTransition(newLabel);
		if (isSilent)
			tNew.setInvisible(true);
		Place pNew = pn.addPlace(newLabel);
		m.getImn().setPlaceId(pNew, (int) pNew.hashCode());
		m.getImn().setTranId(tNew, (int) tNew.hashCode());

		for (Transition t : tranSet)
			pn.addArc(t, pNew);
		pn.addArc(pNew, tNew);
		for (Place p : placeSet) {
			pn.addArc(tNew, p);
			for (Transition t : tranSet)
				for (PetrinetEdge<? extends PetrinetNode, ? extends PetrinetNode> e : pn.getOutEdges(t))
					if (e.getTarget().equals(p))
						pn.removeEdge(e);
		}

		Set<Integer> placeIDs = new HashSet<Integer>();
		Set<Integer> tranIDs = new HashSet<Integer>();
		//		newPlaceId = m.getImn().getPlaceId(pNew);
		//		newTranId = m.getImn().getTranId(tNew);
		for (Place p : placeSet)
			placeIDs.add(m.getImn().getPlaceId(p));
		for (Transition t : tranSet)
			tranIDs.add(m.getImn().getTranId(t));

		//update TranLD
		//		MatrixUtils.updatePlaceLD(m.getImn(), pn, tranIDs, newCol);
		//		MatrixUtils.updateTranLD(m.getImn(), pn, placeIDs, newRow);

		//update red tracker
		m.addTrackSynthesis(new ReductionTrackerL(Rule.ABSTRACTION, tranIDs, placeIDs, tNew.getLabel(), newLabel,
				isSilent, m.getImn().getPlaceId(pNew), m.getImn().getTranId(tNew)));

		if (calcSynthSpace)
			AddNewTranBwTranAndPlaceL.updateLDSets(placeSet, tranSet, m.getImn(), pNew, tNew);

		return tNew;
	}

	public static void addChoiceTransition(IncrementalModelL m, Set<Integer> source, Set<Integer> target, String label,
			boolean calcSynthSpace) {
		Petrinet pn = m.getNet();

		Set<Transition> fcSet = new HashSet<Transition>();
		Set<Place> pStartSet = new HashSet<Place>();
		Set<Place> pEndSet = new HashSet<Place>();
		for (int p : source)
			pStartSet.add(m.getImn().getPlaceBasedOnID(p));
		for (int p : target)
			pEndSet.add(m.getImn().getPlaceBasedOnID(p));

		//add an escape silent tran (if not already present) before adding the choice tran 
		for (Place p : pStartSet)
			for (PetrinetEdge<? extends PetrinetNode, ? extends PetrinetNode> e : pn.getOutEdges(p)) {
				Transition t = (Transition) e.getTarget();
				Set<Integer> sourceTrans = new HashSet<Integer>();
				Set<Integer> targetPlaces = new HashSet<Integer>();
				sourceTrans.add(m.getImn().getTranId(t));
				if (!t.isInvisible()) {
					//make this tran invisible, and create a new transition in seq
					t.setInvisible(true);
					m.getKeepTrans().add(m.getImn().getTranId(t));
					//update reduction
					m.addTrackSynthesis(new ReductionTrackerL(Rule.RENAME, source, target, t.getLabel(), p.getLabel(), t
							.isInvisible(), (int) 0, m.getImn().getTranId(t)));

					for (PetrinetEdge<? extends PetrinetNode, ? extends PetrinetNode> e2 : pn.getOutEdges(t)) {
						targetPlaces.add(m.getImn().getPlaceId((Place) e2.getTarget()));
					}
					addNewTransition(m, sourceTrans, targetPlaces, t.getLabel(), calcSynthSpace);
				} else {
					//also verify if t is bot, in that case add new tran b/w t and end
					for (PetrinetEdge<? extends PetrinetNode, ? extends PetrinetNode> e2 : pn.getOutEdges(t)) {
						targetPlaces.add(m.getImn().getPlaceId((Place) e2.getTarget()));
						if (PetrinetUtilsL.getFinalPlace(pn).equals(e2.getTarget())) {
							addNewTransition(m, sourceTrans, targetPlaces, "", calcSynthSpace);
						}
					}
				}
			}

		Transition tNew = pn.addTransition(label);
		if (label == "")
			tNew.setInvisible(true);
		m.getImn().setTranId(tNew, (int) tNew.hashCode());

		for (Place pStart : pStartSet) {
			for (PetrinetEdge<? extends PetrinetNode, ? extends PetrinetNode> e : pn.getOutEdges(pStart)) {
				//add every start trans as freeChoice
				fcSet.add((Transition) e.getTarget());
			}
			pn.addArc(pStart, tNew);
		}

		for (Place pEnd : pEndSet)
			pn.addArc(tNew, pEnd);
		try {
			//update reduction
			m.addTrackSynthesis(new ReductionTrackerL(Rule.LDTRAN, source, target, tNew.getLabel(), label, tNew
					.isInvisible(), (int) 0, m.getImn().getTranId(tNew)));
			if (calcSynthSpace)
				ChoiceTransitionL.updateLDSets(pStartSet, pEndSet, m.getImn(), tNew);
		} catch (NullPointerException e) {

		}
		//		MatrixUtils.updatePlaceLDForNewTran(m.getImn(), pn, m.getImn().getTranId(tNew));

	}

	public static void addParallelPlaceAndSeq(IncrementalModelL m, Set<Integer> source, Set<Integer> target, String label,
			boolean calcSynthSpace) {
		Petrinet pn = m.getNet();
		Place pNew = pn.addPlace(label);
		m.getImn().setPlaceId(pNew, (int) pNew.hashCode());
		Set<Transition> tStartSet = new HashSet<Transition>();
		Set<Transition> tEndSet = new HashSet<Transition>();
		for (int t : source)
			tStartSet.add(m.getImn().getTranBasedOnID(t));
		for (int t : target)
			tEndSet.add(m.getImn().getTranBasedOnID(t));

		for (Transition t : tStartSet)
			pn.addArc(t, pNew);

		for (Transition t : tEndSet)
			pn.addArc(pNew, t);

		//update reduction tracker
		m.addTrackSynthesis(new ReductionTrackerL(Rule.LDPLACE, source, target, "", pNew.getLabel(), false, m.getImn()
				.getPlaceId(pNew), (int) 0));

		//		newPlaceId1 = m.getImn().getPlaceId(pNew);		//use the id in main calling fun
		if (calcSynthSpace)
			ParallelPlaceL.updateLDSets(tStartSet, tEndSet, m.getImn(), pNew);

		//		MatrixUtils.updateTranLDForNewPlace(m.getImn(), pn, m.getImn().getPlaceId(pNew));

		//add seq between tStart and newly added place - with the current label
		Set<Integer> pSet = new HashSet<Integer>();
		pSet.add(m.getImn().getPlaceId(pNew));
		addNewTransition(m, source, pSet, label, calcSynthSpace);

		//if tEndSet contains last (bot) transition, then make a new bot transition, and replace both with new tran
		if (tEndSet.contains(PetrinetUtilsL.getFinalTranistion(m.getNet()))) {
			//					Transition bot = PetrinetUtils.getFinalTranistion(m.getNet());
			//					Set<Place> places = new HashSet<Place>();
			//					Set<Transition> trans = new HashSet<Transition>();
			//					for (PetrinetEdge<? extends PetrinetNode, ? extends PetrinetNode> eIn : m.getNet().getInEdges(bot)) {
			//						places.add((Place) eIn.getSource());
			//						for (PetrinetEdge<? extends PetrinetNode, ? extends PetrinetNode> eInTran : m.getNet().getInEdges(
			//								eIn.getSource())) {
			//							trans.add((Transition) eInTran.getSource());
			//						}
			//					}
			//					Transition tNew = addNewTransition(m, trans, places, label, true, calcSynthSpace); //use this new tran as the end tran
			//					
			//					//use tNew for end
			//					tEndSet.remove(bot);
			//					tEndSet.add(tNew);
			//					target.remove(m.getImn().getTranId(bot));
			//					target.add(m.getImn().getTranId(tNew));

			//add a new end Tran
			Set<Integer> tran = new HashSet<Integer>();
			Set<Integer> place = new HashSet<Integer>();
			for (Transition t : tEndSet) {
				tran.add(m.getImn().getTranId(t));
				for (PetrinetEdge<? extends PetrinetNode, ? extends PetrinetNode> eOut : m.getNet().getOutEdges(t))
					place.add(m.getImn().getPlaceId((Place) eOut.getTarget()));
			}
			addNewTransition(m, tran, place, "", calcSynthSpace);
		}

	}

	//create a forward silent tran in m1 and backward tran in m2, for the newly added tran
	public static void addSilentTransitions(IncrementalModelL m1, IncrementalModelL m2, int tranId,
			boolean calcSynthSpace) {
		Set<Integer> source = new HashSet<Integer>();
		Set<Integer> target = new HashSet<Integer>();
		Place targetPlace = null;
		for (PetrinetEdge<? extends PetrinetNode, ? extends PetrinetNode> e : m1.getNet().getInEdges(
				m1.getImn().getTranBasedOnID(tranId)))
			source.add(m1.getImn().getPlaceId((Place) e.getSource()));
		for (PetrinetEdge<? extends PetrinetNode, ? extends PetrinetNode> e : m1.getNet().getOutEdges(
				m1.getImn().getTranBasedOnID(tranId))) {
			target.add(m1.getImn().getPlaceId((Place) e.getTarget()));
			targetPlace = m2.getImn().getPlaceBasedOnID(m1.getImn().getPlaceId((Place) e.getTarget()));
		}
		//		targetPlace = (Place) e.getTarget();

		//		target.add(m2.getImn().getPlaceId(targetPlace)); //use m2 here as we use the second place node later on for loops (the ids are same for m1 & m2)
		addChoiceTransition(m1, source, new HashSet<Integer>(target), "", calcSynthSpace);

		Set<Integer> tSet = new HashSet<Integer>();
		tSet.add(tranId);

		if (m2.getNet().getOutEdges(targetPlace).size() > 1 || m2.getNet().getInEdges(targetPlace).size() > 1) { //if join or split then add an escaping tran

			addNewTransition(m2, tSet, target, "", calcSynthSpace); //escaping tran

			//find the new starting place for looping back
			int newPlaceId = -1;
			for (Place p1 : m2.getImn().getAllPlaceId().keySet()) { //search the newly added tran- use this to add new silent tran
				boolean present = false;
				for (Place p2 : m1.getImn().getAllPlaceId().keySet()) {
					if (m2.getImn().getPlaceId(p1).equals(m1.getImn().getPlaceId(p2))) //check if the place is newly add seq place
						present = true;
				}
				if (!present)
					newPlaceId = m2.getImn().getPlaceId(p1);
			}
			target.clear();
			target.add(newPlaceId); //use the newly created place from seq as the start place of loop back
		}

		addNewTransition(m2, tSet, target, "", calcSynthSpace); //escaping tran within the loop

		//		//check and add additional outgoing trans that may be present in target
		//		for (PetrinetEdge<? extends PetrinetNode, ? extends PetrinetNode> e : m2.getNet().getOutEdges(targetPlace)) {
		//			for (PetrinetEdge<? extends PetrinetNode, ? extends PetrinetNode> e2 : m2.getNet()
		//					.getInEdges(e.getTarget())) {
		//				if (target.add(m2.getImn().getPlaceId((Place) e2.getSource()))) //if not already present in target
		//					source.add(m2.getImn().getPlaceId((Place) e2.getSource()));
		//
		//			}
		//		}

		//add a new escape transition between new tran n new place (this is the escape tran)
		int tNewId = -1;

		addChoiceTransition(m2, target, source, "", calcSynthSpace);

	}

}
