package org.processmining.incrementaldiscovery.models;

import java.util.ArrayList;
import java.util.HashSet;
import java.util.List;
import java.util.Set;

import org.processmining.models.graphbased.directed.petrinet.Petrinet;
import org.processmining.prabhakardixit.fcnets.models.InteractiveMatrixNetL;
import org.processmining.reduction.ReductionTrackerL;

public class IncrementalModelL {
	private InteractiveMatrixNetL imn;
	private Petrinet net;
	private float fitness;
	private float precision;
	private Set<Integer> keepTrans = new HashSet<Integer>(); //these transitions are escape routes to loops or choices, never use these in new place 
	private List<ReductionTrackerL> trackSynthesis; //expanded using a different function (not expand using tracker from plugin)

	public IncrementalModelL(InteractiveMatrixNetL imnOld) {
		//use old imn to create the new imn corresponding to this model
		imn = new InteractiveMatrixNetL(imnOld, true);
	}

	public IncrementalModelL(IncrementalModelL m) {
		//use old imn to create the new imn corresponding to this model
		imn = new InteractiveMatrixNetL(m.getImn(), true);
		if (m.getTrackSynthesis() != null) {
			trackSynthesis = new ArrayList<ReductionTrackerL>();
			for (int i = 0; i < m.getTrackSynthesis().size(); i++)
				trackSynthesis.add(m.getTrackSynthesis().get(i));
		}
	}

	public InteractiveMatrixNetL getImn() {
		return imn;
	}

	public void setImn(InteractiveMatrixNetL imn) {
		this.imn = imn;
	}

	public float getFitness() {
		if (fitness > 0.99)
			return fitness;
		else
			return round(fitness, 2);
	}

	public void setFitness(float fitness) {
		this.fitness = fitness;
	}

	public float getPrecision() {
		if (precision > 0.99)
			return precision;
		else
			return round(precision, 2);
	}

	public void setPrecision(float precision) {
		this.precision = precision;
	}

	public Petrinet getNet() {
		return imn.getselectedPetriNetMap().getPetriNet();
	}

	public void setNet(Petrinet net) {
		this.net = net;
	}

	public Set<Integer> getKeepTrans() {
		return keepTrans;
	}

	public void setKeepTrans(Set<Integer> keepTrans) {
		this.keepTrans = keepTrans;
	}

	public List<ReductionTrackerL> getTrackSynthesis() {
		return trackSynthesis;
	}

	public void addTrackSynthesis(ReductionTrackerL tracker) {
		if (this.trackSynthesis != null)
			trackSynthesis.add(tracker);
		else {
			trackSynthesis = new ArrayList<ReductionTrackerL>();
			trackSynthesis.add(tracker);
		}
	}

	private static float round(float value, int places) {
		if (places < 0)
			throw new IllegalArgumentException();

		long factor = (long) Math.pow(10, places);
		value = value * factor;
		long tmp = Math.round(value);
		return (float) tmp / factor;
	}

}
