package org.processmining.layouter;

import java.util.Iterator;

import org.processmining.acceptingpetrinet.models.AcceptingPetriNet;
import org.processmining.acceptingpetrinet.models.impl.AcceptingPetriNetImpl;
import org.processmining.acceptingpetrinetreductionlayouter.parameters.LayoutParameters;
import org.processmining.models.graphbased.directed.petrinet.Petrinet;
import org.processmining.models.graphbased.directed.petrinet.PetrinetNode;
import org.processmining.models.graphbased.directed.petrinet.elements.Place;
import org.processmining.models.graphbased.directed.petrinet.elements.Transition;
import org.processmining.models.graphbased.directed.petrinet.impl.PetrinetImpl;

import prefuse.Visualization;
import prefuse.data.Edge;
import prefuse.data.Node;

public class PetrinetBoxStructureLayouterL {

	/**
	 * Returns layout result with each node labeled as unique ID from IPM
	 * 
	 * @return
	 */
	public static LayoutAlgorithmReductionRulesL getLayout(Visualization visualization) {

		Petrinet pn = getPetriNetFromData(visualization);

		LayoutAlgorithmReductionRulesL lar = new LayoutAlgorithmReductionRulesL();
		LayoutParameters parameters = new LayoutParameters();
		AcceptingPetriNet net = new AcceptingPetriNetImpl(pn);
		parameters.setHorizontalSpace(0);
		parameters.setVerticalSpace(60);
		lar.apply(net, parameters);
		return lar;
	}

	private static Petrinet getPetriNetFromData(Visualization visualization) {
		Iterator iter = visualization.getGroup("graph.nodes").tuples();
		Petrinet pn = new PetrinetImpl("net");
		while (iter.hasNext()) {
			//first rename all nodes with unique ids
			Node n = (Node) iter.next();
			if (n.get("type") == "TRANSITION")
				pn.addTransition(n.get("ID").toString());
			else
				pn.addPlace(n.get("ID").toString());
		}

		Iterator iter2 = visualization.getGroup("graph.edges").tuples();
		while (iter2.hasNext()) {
			Edge e = (Edge) iter2.next();
			Node nSource = e.getSourceNode();
			Node nTarget = e.getTargetNode();
			PetrinetNode source = null;
			PetrinetNode target = null;
			for (Transition t : pn.getTransitions())
				if (nSource.get("type") == "TRANSITION" && Integer.valueOf(t.getLabel()).equals(nSource.get("ID")))
					source = t;
				else if (nTarget.get("type") == "TRANSITION" && Integer.valueOf(t.getLabel()).equals(nTarget.get("ID")))
					target = t;
			for (Place p : pn.getPlaces())
				if (nSource.get("type") == "PLACE" && Integer.valueOf(p.getLabel()).equals(nSource.getInt("ID")))
					source = p;
				else if (nTarget.get("type") == "PLACE" && Integer.valueOf(p.getLabel()).equals(nTarget.get("ID")))
					target = p;
			if (nSource.get("type") == "TRANSITION")
				pn.addArc((Transition) source, (Place) target);
			else
				pn.addArc((Place) source, (Transition) target);

		}
		return pn;
	}
}
