package org.processmining.prabhakardixit.alignments;

import java.util.HashMap;
import java.util.HashSet;
import java.util.Map;
import java.util.Set;
import java.util.TreeSet;

import org.deckfour.xes.model.XLog;
import org.processmining.models.graphbased.directed.petrinet.elements.Transition;
import org.processmining.plugins.petrinet.replayresult.PNRepResult;
import org.processmining.prabhakardixit.fcnets.models.PetriNetMapL;
import org.processmining.projectedrecallandprecision.helperclasses.EfficientLog;

import com.google.common.collect.HashMultiset;
import com.google.common.collect.Multiset;

public class TransPairAlignmentL {
	Set<String> activities = new HashSet<String>();

	//	String activity2;
	PetriNetMapL pnmap;
	XLog log; //TODO set this up
	EfficientLog efficientLog;		//use this log for projected fitness and pre
	float fitness;
	float precision;
	PNRepResult replayResult;
	Map<Transition, Set<Integer>> modelMoveMap = new HashMap<Transition, Set<Integer>>(); //map model moves to trace ids used for choice and loop trans
	Multiset<Transition> syncMoves = HashMultiset.create();
	Multiset<Transition> modelMoves = HashMultiset.create();
	Multiset<String> logMoves = HashMultiset.create();

	//new alignments
	Map<Transition, Integer> syncMoveMap = new HashMap<Transition, Integer>();
	Map<Transition, Integer> modelMove = new HashMap<Transition, Integer>();
	Map<String, Integer> logMoveMap = new HashMap<String, Integer>();

	//new alignments

	public double getPrecision() {
		return AlignmentUtils.round(precision, 2);
	}

	public void setPrecision(double precision) {
		this.precision = (float) precision;
	}

	public double getFitness() {
		if (fitness > 0.99)
			return fitness;
		else
			return AlignmentUtils.round(fitness, 2);
	}

	public void setFitness(double fitness) {
		this.fitness = (float) fitness;
	}

	public XLog getLog() {
		return log;
	}

	public void setLog(XLog log) {
		this.log = log;
	}

	public PetriNetMapL getPetriNetMap() {
		return pnmap;
	}

	public void setPetriNetMap(PetriNetMapL petriNet) {
		this.pnmap = petriNet;
	}

	public void put(String activity) {

		this.activities.add(activity);
		//		setAct2(t2);
	}

	public Set<String> getActivities() {
		return activities;
	}

	public void setActivities(Set<String> activities) {
		this.activities = activities;
	}

	//	public String getAct2() {
	//		return activity2;
	//	}
	//
	//	public void setAct2(String activity2) {
	//		this.activity2 = activity2;
	//	}

	public PNRepResult getReplayResult() {
		return replayResult;
	}

	public void setReplayResult(PNRepResult replayResult) {
				this.replayResult = replayResult;
	}

	public boolean containsActivity(String act) {
		if (this.activities.contains(act))
			return true;
		return false;
	}

	public Transition getMappedTransition(Transition t2) {
		for (Transition t : getPetriNetMap().getMapTransitions().keySet()) {
			if (getPetriNetMap().getMapTransitions().get(t).equals(t2))
				return t;

		}
		return null;
	}

	public Map<Transition, Set<Integer>> getModelMoveMap() {
		return modelMoveMap;
	}

	public void setModelMoveMap(Map<Transition, Set<Integer>> modelMoveMap) {
		this.modelMoveMap = modelMoveMap;
	}

	public void setModelMoveMap(Transition t, Set<Integer> ids) {
		if (this.modelMoveMap.containsKey(t))
			modelMoveMap.get(t).addAll(ids);
		else
			modelMoveMap.put(t, new HashSet<Integer>(ids));
	}

	public Multiset<Transition> getSyncMoves() {
		return syncMoves;
	}

	public void setSyncMoves(Multiset<Transition> syncMoves) {
		this.syncMoves = syncMoves;
	}

	public Multiset<Transition> getModelMoves() {
		return modelMoves;
	}

	public void setModelMoves(Multiset<Transition> modelMoves) {
		this.modelMoves = modelMoves;
	}

	public Multiset<String> getLogMoves() {
		return logMoves;
	}

	public void setLogMoves(Multiset<String> logMoves) {
		this.logMoves = logMoves;
	}

	public void setLogMoves(String logMove) {
		this.logMoves.add(logMove);

	}

	public Integer getNumLogMoves(String act) {
		return logMoves.count(act);
	}

	public Set<String> getSortedActivities() {
		return new TreeSet<>(activities);
		//		return sortedAct;
	}

	private static class AlignmentUtils {
		private static double round(double value, int places) {
			if (places < 0)
				throw new IllegalArgumentException();

			long factor = (long) Math.pow(10, places);
			value = value * factor;
			long tmp = Math.round(value);
			return (double) tmp / factor;
		}
	}

	public Map<String, Integer> getLogMoveMap() {
		return logMoveMap;
	}

	public void setlogMoveMap(Map<String, Integer> logMoveMap) {
		this.logMoveMap = logMoveMap;
	}

	public Map<Transition, Integer> getModelMove() {
		return modelMove;
	}

	public void setModelMove(Map<Transition, Integer> modelMove) {
		this.modelMove = modelMove;
	}

	public void incrementModelMove(Transition t, int val) {
		if (this.modelMove.containsKey(t))
			this.modelMove.put(t, this.modelMove.get(t) + val);
		else
			this.modelMove.put(t, val);
	}

	public void incrementLogMoveMap(String t, int val) {
		if (this.logMoveMap.containsKey(t))
			this.logMoveMap.put(t, this.logMoveMap.get(t) + val);
		else
			this.logMoveMap.put(t, val);
	}

	public Map<Transition, Integer> getSyncMoveMap() {
		return syncMoveMap;
	}

	public void setSyncMoveMap(Map<Transition, Integer> syncMoveMap) {
		this.syncMoveMap = syncMoveMap;
	}

	public void incrementSyncMoveMap(Transition t, int val) {
		if (this.syncMoveMap.containsKey(t))
			this.syncMoveMap.put(t, this.syncMoveMap.get(t) + val);
		else
			this.syncMoveMap.put(t, val);
	}


	public EfficientLog getEfficientLog() {
		return efficientLog;
	}

	public void setEfficientLog(EfficientLog efficientLog) {
		this.efficientLog = efficientLog;
	}

}
