package org.processmining.prabhakardixit.fcnets.matrix;

import java.util.ArrayList;
import java.util.Collection;
import java.util.Collections;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Set;

import org.apache.commons.math3.optim.MaxIter;
import org.apache.commons.math3.optim.PointValuePair;
import org.apache.commons.math3.optim.linear.LinearConstraint;
import org.apache.commons.math3.optim.linear.LinearConstraintSet;
import org.apache.commons.math3.optim.linear.LinearObjectiveFunction;
import org.apache.commons.math3.optim.linear.NoFeasibleSolutionException;
import org.apache.commons.math3.optim.linear.NonNegativeConstraint;
import org.apache.commons.math3.optim.linear.Relationship;
import org.apache.commons.math3.optim.linear.SimplexSolver;
import org.apache.commons.math3.optim.nonlinear.scalar.GoalType;
import org.ejml.simple.SimpleMatrix;
import org.processmining.models.graphbased.directed.petrinet.Petrinet;
import org.processmining.models.graphbased.directed.petrinet.PetrinetEdge;
import org.processmining.models.graphbased.directed.petrinet.PetrinetNode;
import org.processmining.models.graphbased.directed.petrinet.elements.Place;
import org.processmining.models.graphbased.directed.petrinet.elements.Transition;
import org.processmining.prabhakardixit.alignments.CombinationUtilsL;
import org.processmining.prabhakardixit.fcnets.models.InteractiveMatrixNetL;
import org.processmining.prabhakardixit.fcnets.models.InteractivePetriNetL;
import org.processmining.prabhakardixit.fcnets.models.LDSetsL;
import org.processmining.prabhakardixit.fcnets.models.PetriNetMapL;
import org.processmining.prabhakardixit.utils.PetrinetUtilsL;
import org.processmining.prabhakardixit.utils.FCnetsUtilsL;

public class MatrixUtilsL {

	public static void initiate(InteractivePetriNetL ipn) {
		PetriNetMapL pnmap = ipn.getselectedPetriNetMap();
		Place pi = PetrinetUtilsL.getInitialPlace(pnmap.getPetriNet());
		Place pf = PetrinetUtilsL.getFinalPlace(pnmap.getPetriNet());
		for (Place p : pnmap.getPetriNet().getPlaces())
			if (p.equals(pi))
				ipn.setPlaceId(p, (Integer) 0);
			else if (p.equals(pf))
				ipn.setPlaceId(p, (Integer) 2);
			else
				ipn.setPlaceId(p, (Integer) 1);
		int i = 0;
		for (Transition t : pnmap.getPetriNet().getTransitions())
			ipn.setTranId(t, (Integer) i++);

		//update incidence matrix
		//		incidenceMatrix = new SimpleMatrix(0, 0);
		double[][] inci = new double[pnmap.getPetriNet().getPlaces().size()][pnmap.getPetriNet().getTransitions()
				.size()];
		//perf starts
		Set<Integer> negativeTransForNewLD = new HashSet<Integer>();
		Set<Integer> positiveTransForNewLD = new HashSet<Integer>();
		negativeTransForNewLD.add((Integer) 1);
		positiveTransForNewLD.add((Integer) 0);
		if (!ipn.getInputOutTransitions().contains(negativeTransForNewLD))
			ipn.getInputOutTransitions().add(negativeTransForNewLD);
		if (!ipn.getInputOutTransitions().contains(positiveTransForNewLD))
			ipn.getInputOutTransitions().add(positiveTransForNewLD);
		ipn.addLdPlaceInfo(0, ipn.getInputOutTransitions().indexOf(negativeTransForNewLD), ipn.getInputOutTransitions()
				.indexOf(positiveTransForNewLD)); //performance3
		//perf ends

		int id = 0;

		int placeSize = ipn.getselectedPetriNetMap().getPetriNet().getPlaces().size();
		InteractiveMatrixNetL imn = ipn.getInteractiveMatNet();
		for (Integer row = 0; row < placeSize; row++)
			addSelfLoop(imn, row);

		//also add new tran starting with mid and ending with end
		initiateMidToEndTran(imn);
		updatePlaceLDNonSiphons(imn, ipn.getselectedPetriNetMap().getPetriNet());
		updateTranLDFreeChoices(imn, ipn.getselectedPetriNetMap().getPetriNet(), null);

		for (LDSetsL l : imn.getplaceLDNonSiphons()) {
			l.getNegativeNodes(imn);
			l.getPositiveNodes(imn);
		}
		ipn.getAllTranLDFCs();

	}

	//	private static double[][] toArray(DenseMatrix64F matrix) {
	//		double array[][] = new double[matrix.getNumRows()][matrix.getNumCols()];
	//		for (int r = 0; r < matrix.getNumRows(); r++) {
	//			System.out.println(" ");
	//			for (int c = 0; c < matrix.getNumCols(); c++) {
	//				array[r][c] = matrix.get(r, c);
	//				System.out.print(array[r][c] + " ");
	//			}
	//		}
	//		return array;
	//	}

	public static void updateTranLD(InteractiveMatrixNetL imn, Petrinet pn, Set<Integer> oldPlaceIDs, Integer newRow) {
		//		InteractivePetriNet ipn=null;
		Set<LDSetsL> ldTrans = new HashSet<LDSetsL>();
		Set<LDSetsL> ldTransNew = new HashSet<LDSetsL>();
		for (int place : oldPlaceIDs)
			ldTrans.addAll(imn.getLdPlaceMapForTranLD(place));

		for (LDSetsL ldt : ldTrans) {
			boolean allOne = false;
			boolean allNegativeOne = false;
			if (ldt.getPositiveNodes(imn).containsAll(oldPlaceIDs) || ldt.getPositiveNodes(imn).equals(oldPlaceIDs))
				allOne = true;

			if (ldt.getNegativeNodes(imn).containsAll(oldPlaceIDs) || ldt.getNegativeNodes(imn).equals(oldPlaceIDs))
				allNegativeOne = true;

			Set<Integer> negativeRow = new HashSet<Integer>(ldt.getNegativeNodes(imn));
			Set<Integer> positiveRow = new HashSet<Integer>(ldt.getPositiveNodes(imn));
			if (allNegativeOne || allOne) { //copy LD's to pNew wrt to oldP's
				if (allOne) {
					positiveRow.removeAll(oldPlaceIDs);
					positiveRow.add(newRow);
					if (!imn.getInputOutPlaces().contains(positiveRow))
						imn.getInputOutPlaces().add(positiveRow);
					if (!imn.getInputOutPlaces().contains(negativeRow))
						imn.getInputOutPlaces().add(negativeRow);
					ldTransNew.add(updateTranLDMatrix(imn, imn.getInputOutPlaces().indexOf(positiveRow), imn
							.getInputOutPlaces().indexOf(negativeRow)));
				}
				if (allNegativeOne) {
					positiveRow = new HashSet<Integer>(ldt.getPositiveNodes(imn));
					negativeRow.removeAll(oldPlaceIDs);
					negativeRow.add(newRow);
					if (!imn.getInputOutPlaces().contains(positiveRow))
						imn.getInputOutPlaces().add(positiveRow);
					if (!imn.getInputOutPlaces().contains(negativeRow))
						imn.getInputOutPlaces().add(negativeRow);
					ldTransNew.add(updateTranLDMatrix(imn, imn.getInputOutPlaces().indexOf(positiveRow), imn
							.getInputOutPlaces().indexOf(negativeRow)));
				}
				if (allOne && allNegativeOne) {
					positiveRow.removeAll(oldPlaceIDs);
					positiveRow.add(newRow); //negative is already taken care of above^^
					ldTransNew.add(updateTranLDMatrix(imn, imn.getInputOutPlaces().indexOf(positiveRow), imn
							.getInputOutPlaces().indexOf(negativeRow)));
				}
			}
		}
		updateTranLDFreeChoices(imn, pn, ldTransNew);
	}

	private static LDSetsL updateTranLDMatrix(InteractiveMatrixNetL imn, Integer posId, Integer negId) {
		LDSetsL ldset = new LDSetsL(imn.getLdTranInfo().size(), negId, posId, false);
		ldset.setID(ldset.hashCode());
		for (Integer i : imn.getInputOutPlaces().get(negId))
			imn.setLdPlaceMapForTranLD(i, ldset);
		for (Integer i : imn.getInputOutPlaces().get(posId))
			imn.setLdPlaceMapForTranLD(i, ldset);
		imn.setLdTranInfo(ldset); //performance
		//		updateTranLDFreeChoices(ipn, ldset);
		return ldset;
	}

	//	private static void updateTranLDFreeChoices1(InteractivePetriNet ipn, LDSets ldt) {//**change this to use only indices
	//		Set<LDSets> selfLoops = new HashSet<LDSets>();
	//
	//		Set<Integer> tranSplitPlaces = fcnetsUtils.getAllPostPlacesAfterTranSplit(ipn); //contains all the first places after tran split
	//		Set<LDSets> tranLDsWithSplitPlacesAsStart = new HashSet<LDSets>(); //LD tran with all tran split places as start
	//
	//		Set<Integer> startSet = new HashSet<Integer>(ldt.getNegativeNodes(ipn)); //places
	//		Set<Place> startPlaces = new HashSet<Place>();
	//		for (int i : startSet) {
	//			startPlaces.add(ipn.getPlaceBasedOnID(i));
	//		}
	//		boolean fcStartTrans = true;
	//		//		Set<Integer> allOneCols = new HashSet<Integer>();
	//
	//		if (startSet.size() > 1) {//if more than one, then check FCs
	//
	//			//if the first places after tran split then we keep it or future 
	//			if (tranSplitPlaces.containsAll(startSet))
	//				tranLDsWithSplitPlacesAsStart.add(ldt);
	//
	//			fcStartTrans = fcnetsUtils.checkFCForLDTrans(ipn, startPlaces);
	//		} else {//check if any output tran of start place has more than one input => invalid and (self loop)
	//			fcStartTrans = fcnetsUtils.checkSelfLoopLD(startPlaces, ipn.getselectedPetriNetMap().getPetriNet());
	//			if (!fcStartTrans)
	//				selfLoops.add(ldt); //add as self loop if the post tran of any place has more than one inputs
	//
	//		}
	//		if (fcStartTrans || selfLoops.contains(ldt) || tranLDsWithSplitPlacesAsStart.contains(ldt)) {
	//			if (fcStartTrans) {
	//				ipn.getTranLDFCs().add(ldt.hashCode());
	//			} else {
	//				ipn.getTranLDFCs().remove(ldt.hashCode());
	//			}
	//		} else {
	//			ipn.getAllLdTranInfo().remove(ldt);
	//			for (int key : ipn.getAllLdPlaceMapForTranLD().keySet())
	//				if (ipn.getLdPlaceMapForTranLD(key).contains(ldt))
	//					ipn.getLdPlaceMapForTranLD(key).remove(ldt);
	//
	//		}
	//	}

	//create incidenceMat and the map of the trans and places in inciMat
	public static SimpleMatrix createInciMap(Map<Transition, Integer> tranInciMap, Map<Place, Integer> placeInciMap,
			Petrinet pn) {
		SimpleMatrix inciMat = new SimpleMatrix(0, 0);

		int count = 0;
		//first fix the position of transitions in the inciMat
		for (Transition t : pn.getTransitions()) {
			tranInciMap.put(t, count++);
		}
		count = 0;

		//create inciMat
		for (Place p : pn.getPlaces()) {
			placeInciMap.put(p, count++);
			Collection<PetrinetEdge<? extends PetrinetNode, ? extends PetrinetNode>> inEdges = pn.getInEdges(p);
			Collection<PetrinetEdge<? extends PetrinetNode, ? extends PetrinetNode>> outEdges = pn.getOutEdges(p);
			SimpleMatrix newRow = new SimpleMatrix(1, pn.getTransitions().size());

			for (PetrinetEdge<? extends PetrinetNode, ? extends PetrinetNode> e : inEdges) {
				newRow.set(0, tranInciMap.get(e.getSource()), 1);
			}
			for (PetrinetEdge<? extends PetrinetNode, ? extends PetrinetNode> e : outEdges) {
				newRow.set(0, tranInciMap.get(e.getTarget()), newRow.get(0, tranInciMap.get(e.getTarget())) - 1);
			}
			inciMat = inciMat.combine(inciMat.numRows(), 0, newRow);
		}
		return inciMat;
	}

	//incidence matrix excluding some transitions
	public static SimpleMatrix createInciMapWithoutSomeTrans(Map<Transition, Integer> tranInciMap,
			Map<Place, Integer> placeInciMap, Set<Transition> exclTrans, Petrinet pn) {
		SimpleMatrix inciMat = new SimpleMatrix(0, 0);

		int count = 0;
		//first fix the position of transitions in the inciMat
		for (Transition t : pn.getTransitions()) {
			if (!exclTrans.contains(t))
				tranInciMap.put(t, count++);
		}
		count = 0;

		//create inciMat
		for (Place p : pn.getPlaces()) {
			placeInciMap.put(p, count++);
			Collection<PetrinetEdge<? extends PetrinetNode, ? extends PetrinetNode>> inEdges = pn.getInEdges(p);
			Collection<PetrinetEdge<? extends PetrinetNode, ? extends PetrinetNode>> outEdges = pn.getOutEdges(p);
			SimpleMatrix newRow = new SimpleMatrix(1, pn.getTransitions().size());

			for (PetrinetEdge<? extends PetrinetNode, ? extends PetrinetNode> e : inEdges) {
				if (tranInciMap.containsKey(e.getSource()))
					newRow.set(0, tranInciMap.get(e.getSource()), 1);
			}
			for (PetrinetEdge<? extends PetrinetNode, ? extends PetrinetNode> e : outEdges) {
				if (tranInciMap.containsKey(e.getTarget()))
					newRow.set(0, tranInciMap.get(e.getTarget()), newRow.get(0, tranInciMap.get(e.getTarget())) - 1);
			}
			inciMat = inciMat.combine(inciMat.numRows(), 0, newRow);
		}
		return inciMat;
	}

	//this is for abstration rule only
	public static void updatePlaceLD(InteractiveMatrixNetL imn, Petrinet pn, Set<Integer> oldTranIDs, Integer newCol) {
		//performance starts
		List<LDSetsL> ldPlaces = imn.getLdPlaceInfo();
		int ldpSize = ldPlaces.size();
		//performance ends

		for (int ldp = 0; ldp < ldpSize; ldp++) {
			int r = ldPlaces.get(ldp).getID();
			boolean allOne = false;
			boolean allNegativeOne = false;
			if (ldPlaces.get(ldp).getPositiveNodes(imn).containsAll(oldTranIDs)
					|| ldPlaces.get(ldp).getPositiveNodes(imn).equals(oldTranIDs))
				allOne = true;

			if (ldPlaces.get(ldp).getNegativeNodes(imn).containsAll(oldTranIDs)
					|| ldPlaces.get(ldp).getNegativeNodes(imn).equals(oldTranIDs))
				allNegativeOne = true;

			if (allNegativeOne || allOne) {
				updateNewPairs(imn, oldTranIDs, ldPlaces.get(ldp).getPositiveNodes(imn), ldPlaces.get(ldp)
						.getNegativeNodes(imn), newCol, allOne, allNegativeOne);
			} else {
				//covers for "0" valued cells (i.e. possibly both input and output)
				//check if there is some overlap between input/output and R (oldTranIds)
				if (!Collections.disjoint(ldPlaces.get(ldp).getPositiveNodes(imn), oldTranIDs)
						|| !Collections.disjoint(ldPlaces.get(ldp).getNegativeNodes(imn), oldTranIDs)) {
					Set<Integer> remainingNodes = new HashSet<Integer>(oldTranIDs);
					remainingNodes.removeAll(ldPlaces.get(ldp).getPositiveNodes(imn));
					remainingNodes.removeAll(ldPlaces.get(ldp).getNegativeNodes(imn));
					if (!remainingNodes.isEmpty()) {
						Set<Set<Integer>> allCombs = CombinationUtilsL.getAllCombs(remainingNodes);
						for (Set<Integer> allComb : allCombs) {
							if (!allComb.isEmpty()) {
								Set<Integer> newPos = new HashSet<Integer>(ldPlaces.get(ldp).getPositiveNodes(imn));
								Set<Integer> newNeg = new HashSet<Integer>(ldPlaces.get(ldp).getNegativeNodes(imn));
								newPos.addAll(allComb);
								newNeg.addAll(allComb);
								allOne = false;
								allNegativeOne = false;
								if (newPos.containsAll(oldTranIDs) || newPos.equals(oldTranIDs))
									allOne = true;

								if (newNeg.containsAll(oldTranIDs) || newNeg.equals(oldTranIDs))
									allNegativeOne = true;
								updateNewPairs(imn, oldTranIDs, newPos, newNeg, newCol, allOne, allNegativeOne);
							}
						}
					}

				}

				//check if it is possible to add the tran/place to input
			}
		}
		//new tran as follows with pNew and pPrev and vice versa
		Set<Integer> positiveTransForNewLD = new HashSet<Integer>(); //perf
		Set<Integer> negativeTransForNewLD = new HashSet<Integer>(); //perf
		for (Integer col : oldTranIDs) {
			negativeTransForNewLD.add(col);
		}
		positiveTransForNewLD.add(newCol);
		if (!imn.getInputOutTransitions().contains(positiveTransForNewLD))
			imn.getInputOutTransitions().add(positiveTransForNewLD);
		if (!imn.getInputOutTransitions().contains(negativeTransForNewLD))
			imn.getInputOutTransitions().add(negativeTransForNewLD);
		imn.addLdPlaceInfo(imn.getLdPlaceInfo().size(), imn.getInputOutTransitions().indexOf(negativeTransForNewLD),
				imn.getInputOutTransitions().indexOf(positiveTransForNewLD)); //performance

		Set<Integer> positiveTransForNewLD2 = new HashSet<Integer>(); //perf
		Set<Integer> negativeTransForNewLD2 = new HashSet<Integer>(); //perf
		for (Integer col : oldTranIDs) {
			positiveTransForNewLD2.add(col);
		}
		negativeTransForNewLD2.add(newCol);
		if (!imn.getInputOutTransitions().contains(positiveTransForNewLD2))
			imn.getInputOutTransitions().add(positiveTransForNewLD2);
		if (!imn.getInputOutTransitions().contains(negativeTransForNewLD2))
			imn.getInputOutTransitions().add(negativeTransForNewLD2);
		imn.addLdPlaceInfo(imn.getLdPlaceInfo().size(), imn.getInputOutTransitions().indexOf(negativeTransForNewLD2),
				imn.getInputOutTransitions().indexOf(positiveTransForNewLD2)); //performance
		updatePlaceLDNonSiphons(imn, pn);
	}

	private static void updateNewPairs(InteractiveMatrixNetL imn, Set<Integer> oldTranIDs, Set<Integer> positiveNodes,
			Set<Integer> negativeNodes, int newCol, boolean allOne, boolean allNegativeOne) {
		Set<Integer> positiveTransForNewLD = new HashSet<Integer>();
		Set<Integer> negativeTransForNewLD = new HashSet<Integer>();
		for (Integer c : positiveNodes)
			if (!(oldTranIDs.contains(c) && allOne)) {
				positiveTransForNewLD.add(c);
			}
		for (Integer c : negativeNodes)
			if (!(oldTranIDs.contains(c) && allNegativeOne)) {
				negativeTransForNewLD.add(c);
			}
		if (allOne) {
			positiveTransForNewLD.add(newCol);
		}
		if (allNegativeOne) {
			negativeTransForNewLD.add(newCol);
		}
		if (!imn.getInputOutTransitions().contains(positiveTransForNewLD))
			imn.getInputOutTransitions().add(positiveTransForNewLD);
		if (!imn.getInputOutTransitions().contains(negativeTransForNewLD))
			imn.getInputOutTransitions().add(negativeTransForNewLD);
		imn.addLdPlaceInfo(imn.getLdPlaceInfo().size(), imn.getInputOutTransitions().indexOf(negativeTransForNewLD),
				imn.getInputOutTransitions().indexOf(positiveTransForNewLD)); //performance

	}

	public static void updatePlaceLDNonSiphons(InteractiveMatrixNetL imn, Petrinet pn) {//**update this wrt indices
		imn.getPlaceLDNonSiphonIndex().clear();
		for (LDSetsL lds : imn.getLdPlaceInfo()) {
			Set<Transition> tStart = new HashSet<Transition>();
			Set<Transition> tEnd = new HashSet<Transition>();
			for (int c : lds.getPositiveNodes(imn))
				tStart.add(imn.getTranBasedOnID(c));
			for (int c : lds.getNegativeNodes(imn))
				tEnd.add(imn.getTranBasedOnID(c));
			boolean fcStartTrans = FCnetsUtilsL.checkFCForLDPlaces(pn, tStart, tEnd);

			if (fcStartTrans && !FCnetsUtilsL.isPlaceSiphon(pn, tStart, tEnd)) //add non siphon places
				imn.getPlaceLDNonSiphonIndex().add((Integer) lds.getID());
		}
	}

	//also keeps self loops and choices wrt first places after splits
	private static void updateTranLDFreeChoices(InteractiveMatrixNetL imn, Petrinet pn, Set<LDSetsL> allSets2) {
		Set<LDSetsL> allSets = new HashSet<LDSetsL>(allSets2 != null ? allSets2 : imn.getLdTranInfo());
		List<LDSetsL> ldTranInfo = imn.getLdTranInfo();
		List<Set<Integer>> nodeInputOutput = new ArrayList<Set<Integer>>(imn.getInputOutPlaces());
		Set<Integer> validFCs = new HashSet<Integer>();
		Set<Integer> removeLDT = new HashSet<Integer>();
		Set<Integer> calculatedNodes = new HashSet<Integer>();
		for (LDSetsL ldt : allSets) {
			if (!calculatedNodes.contains(ldt.getNegativeId())) {
				//				for (Set<Integer> startSet : nodeInputOutput) {
				Set<Integer> startSet = nodeInputOutput.get((Integer) ldt.getNegativeId());
				calculatedNodes.add((Integer) ldt.getNegativeId());
				boolean selfLoops = false;

				Set<Integer> tranSplitPlaces = FCnetsUtilsL.getAllPostPlacesAfterTranSplit(pn, imn); //contains all the first places after tran split
				boolean tranLDsWithSplitPlacesAsStart = false; //LD tran with all tran split places as start

				Set<Place> startPlaces = new HashSet<Place>();
				for (int i : startSet) {
					startPlaces.add(imn.getPlaceBasedOnID(i));
				}
				boolean fcStartTrans = true;

				if (startSet.size() > 1) {//if more than one, then check FCs

					//if the first places after tran split then we keep it or future 
					if (tranSplitPlaces.containsAll(startSet))
						tranLDsWithSplitPlacesAsStart = true;

					fcStartTrans = FCnetsUtilsL.checkFCForLDTrans(pn, startPlaces);
				} else {//check if any output tran of start place has more than one input => invalid and (self loop)
					fcStartTrans = FCnetsUtilsL.checkSelfLoopLD(startPlaces, pn);
					if (!fcStartTrans)
						selfLoops = true; //add as self loop if the post tran of any place has more than one inputs

				}
				if (fcStartTrans || selfLoops || tranLDsWithSplitPlacesAsStart) {
					if (fcStartTrans)
						validFCs.add((Integer) nodeInputOutput.indexOf(startSet));

				} else {
					removeLDT.add((Integer) nodeInputOutput.indexOf(startSet));
				}
				//				}
			}

			if (removeLDT.contains(ldt.getNegativeId())) {
				imn.getLdTranInfo().remove(ldt);
				for (int key : imn.getAllLdPlaceMapForTranLD().keySet())
					if (imn.getLdPlaceMapForTranLD(key).contains(ldt.getID()))
						imn.getLdPlaceMapForTranLD(key).remove(ldt.getID());
				imn.getTranLDFCs().remove(ldt.getID());
			} else if (validFCs.contains(ldt.getNegativeId()))
				imn.getTranLDFCs().add((Integer) ldt.getID());
			else
				imn.getTranLDFCs().remove(ldt.getID());
		}

		//		for(Set<Integer> k: nodeInputOutput)
		//			if(removeLDT.contains(nodeInputOutput.indexOf(k)) || removeLDT.equals(k)){
		//				System.out.println("places : ");
		//				for(int i : k)
		//					System.out.println(imn.getPlaceBasedOnID(i));
		//			}

	}

	public static void printMatrix(SimpleMatrix matrix) {
		System.out.println("---");
		for (int r = 0; r < matrix.numRows(); r++) {
			System.out.println(" ");
			for (int c = 0; c < matrix.numCols(); c++) {
				String space = " ";
				//				if (!String.valueOf(matrix.get(r, c)).contains("-"))
				//					space = " ";
				System.out.print(space + matrix.get(r, c));
			}
		}
	}

	public static void updateTranLDForNewPlace(InteractiveMatrixNetL imn, Petrinet pn, Integer pNewID) {
		//		SimpleMatrix incidenceMatrixInput = ipn.getIncidenceMatrixInput();
		//		SimpleMatrix incidenceMatrixOutput = ipn.getIncidenceMatrixOutput();

		//		System.out.println("Size of All : " + imn.getAllTranLDFCs().size() + " Size of LD Only : "
		//				+ imn.getTranLDFCs().size());
		Map<Transition, Integer> tranInciMap = new HashMap<Transition, Integer>();
		Map<Place, Integer> placeInciMap = new HashMap<Place, Integer>();
		SimpleMatrix inciMat = createInciMap(tranInciMap, placeInciMap, pn); //alsoAugmented TOD

		SimpleMatrix augMat3 = getSCIncidenceMatrix2(inciMat, placeInciMap, pn);

		int inciIndexOfpNew = placeInciMap.get(imn.getPlaceBasedOnID(pNewID));
		SimpleMatrix inciMatWithoutNewLD = new SimpleMatrix(augMat3.numCols(), augMat3.numRows() - 1);
		SimpleMatrix newLDPlace = new SimpleMatrix(augMat3.numCols(), 1);
		splitInciMat(augMat3.transpose(), inciMatWithoutNewLD, newLDPlace, inciIndexOfpNew);
		PointValuePair lam = getNodesInBetween(inciMatWithoutNewLD, newLDPlace, inciIndexOfpNew);
		//performance starts
		Set<Integer> allParallelPlacesWrtPNew = new HashSet<Integer>();
		for (Place p : placeInciMap.keySet()) {
			if (!imn.getPlaceId(p).equals(pNewID)) {
				double val = lam.getKey()[placeInciMap.get(p)];
				if (val != 0.)
					allParallelPlacesWrtPNew.add(imn.getPlaceId(p)); //get the original tran id
			}
		}
		Set<LDSetsL> ldTrans = new HashSet<LDSetsL>();
		Set<LDSetsL> ldTransNew = new HashSet<LDSetsL>();
		for (int i : allParallelPlacesWrtPNew)
			ldTrans.addAll(imn.getLdPlaceMapForTranLD(i));

		//		List<LDSets> ldTrans = ipn.getAllLdTranInfo();
		//		int ldtSize = ldTrans.size();
		//performance ends

		for (LDSetsL ldt : ldTrans) {
			int ind = 0; //default - i.e. parallel palces are completely distinct - dont update anything
			for (int id : ldt.getPositiveNodes(imn)) {
				if (allParallelPlacesWrtPNew.contains(id))
					ind += lam.getKey()[placeInciMap.get(imn.getPlaceBasedOnID(id))];
			}
			for (int id : ldt.getNegativeNodes(imn))
				if (allParallelPlacesWrtPNew.contains(id))
					ind -= lam.getKey()[placeInciMap.get(imn.getPlaceBasedOnID(id))];

			Set<Integer> negativePlacesForNewLD = new HashSet<Integer>(ldt.getNegativeNodes(imn));
			Set<Integer> positivePlacesForNewLD = new HashSet<Integer>(ldt.getPositiveNodes(imn));
			switch (ind) {
				case 0 :

					if (!imn.getInputOutPlaces().contains(negativePlacesForNewLD))
						imn.getInputOutPlaces().add(negativePlacesForNewLD);
					if (!imn.getInputOutPlaces().contains(positivePlacesForNewLD))
						imn.getInputOutPlaces().add(positivePlacesForNewLD);
					LDSetsL ldt2 = new LDSetsL(imn.getLdTranInfo().size(), imn.getInputOutPlaces().indexOf(
							negativePlacesForNewLD), imn.getInputOutPlaces().indexOf(positivePlacesForNewLD), false);
					ldt2.setID(ldt2.hashCode());
					for (int i : negativePlacesForNewLD)
						imn.setLdPlaceMapForTranLD((Integer) i, ldt2);
					for (int i : positivePlacesForNewLD)
						imn.setLdPlaceMapForTranLD((Integer) i, ldt2);
					//					updateTranLDFreeChoices(ipn, ldt2);
					ldTransNew.add(ldt2);
					imn.setLdTranInfo(ldt2); //performance
					negativePlacesForNewLD.add(pNewID);
					positivePlacesForNewLD.add(pNewID);
					if (!imn.getInputOutPlaces().contains(negativePlacesForNewLD))
						imn.getInputOutPlaces().add(negativePlacesForNewLD);
					if (!imn.getInputOutPlaces().contains(positivePlacesForNewLD))
						imn.getInputOutPlaces().add(positivePlacesForNewLD);
					ldt.setNegativeID(imn.getInputOutPlaces().indexOf(negativePlacesForNewLD));
					ldt.setPositiveID(imn.getInputOutPlaces().indexOf(positivePlacesForNewLD));
					break;
				case 1 :
					positivePlacesForNewLD.add(pNewID);
					if (!imn.getInputOutPlaces().contains(positivePlacesForNewLD))
						imn.getInputOutPlaces().add(positivePlacesForNewLD);
					ldt.setPositiveID(imn.getInputOutPlaces().indexOf(positivePlacesForNewLD));
					break;
				case -1 :
					negativePlacesForNewLD.add(pNewID);
					if (!imn.getInputOutPlaces().contains(negativePlacesForNewLD))
						imn.getInputOutPlaces().add(negativePlacesForNewLD);
					ldt.setNegativeID(imn.getInputOutPlaces().indexOf(negativePlacesForNewLD));
				default :
					break;
			}
			imn.setLdPlaceMapForTranLD(pNewID, ldt);
			//			updateTranLDFreeChoices(ipn, ldt);
			ldTransNew.add(ldt);
		}
		Set<Integer> added = new HashSet<Integer>();
		for (LDSetsL lds : imn.getAllTranLDFCs())
			if (!added.add(lds.getID())) {
				//				System.out.println("duplicated" + lds.getID());
			}
		//		System.out.println("Size of All : " + imn.getAllTranLDFCs().size() + " Size of LD Only : "
		//				+ imn.getTranLDFCs().size());
		ldTransNew.add(addSelfLoop(imn, pNewID));
		//		System.out.println("Size of All : " + imn.getAllTranLDFCs().size() + " Size of LD Only : "
		//				+ imn.getTranLDFCs().size());
		updateTranLDFreeChoices(imn, pn, ldTransNew);
		//		System.out.println("Size of All : " + imn.getAllTranLDFCs().size() + " Size of LD Only : "
		//				+ imn.getTranLDFCs().size());
		updatePlaceLDNonSiphons(imn, pn);
		//		System.out.println("Size of All : " + imn.getAllTranLDFCs().size() + " Size of LD Only : "
		//				+ imn.getTranLDFCs().size());

	}

	public static void splitInciMat(SimpleMatrix inciMat, SimpleMatrix withoutLDCol, SimpleMatrix ldCol,
			int colToExclude) {
		for (int r = 0; r < inciMat.numRows(); r++) {
			int c2 = 0;
			for (int c = 0; c < inciMat.numCols(); c++) {
				if (c == colToExclude)
					ldCol.set(r, 0, inciMat.get(r, c));
				else
					withoutLDCol.set(r, c2++, inciMat.get(r, c));
			}
		}
	}

	public static SimpleMatrix getSCIncidenceMatrix2(SimpleMatrix incidenceMatrix, Map<Place, Integer> placeInciMap,
			Petrinet pn) {
		//get Integer circuited incidence matrix
		int initialPlace = placeInciMap.get(PetrinetUtilsL.getInitialPlace(pn));
		int finalPlace = placeInciMap.get(PetrinetUtilsL.getFinalPlace(pn));
		SimpleMatrix newCol = new SimpleMatrix(incidenceMatrix.numRows(), 1);
		newCol.set(initialPlace, 0, 1);
		newCol.set(finalPlace, 0, -1);
		incidenceMatrix = incidenceMatrix.combine(0, incidenceMatrix.numCols(), newCol);
		return incidenceMatrix;
	}

	private static boolean isAugMatLD(InteractiveMatrixNetL imn, SimpleMatrix augMat, LDSetsL ldSets, int addNum) {
		//		SimpleMatrix augMat2 = new SimpleMatrix(0, 0);
		SimpleMatrix ldCol = new SimpleMatrix(0, 0);

		if (ldSets.getNegativeNodes(imn).equals(ldSets.getPositiveNodes(imn)))
			return true;
		SimpleMatrix newCol = new SimpleMatrix(augMat.numRows(), 1);
		if (addNum != 1) {
			for (int id : ldSets.getPositiveNodes(imn)) {
				if (!ldSets.getNegativeNodes(imn).contains(id)) {
					newCol.set(id + addNum, 0, 1);
				}
			}
			for (int id : ldSets.getNegativeNodes(imn)) {
				if (!ldSets.getPositiveNodes(imn).contains(id)) {
					newCol.set(id + addNum, 0, -1);
				}
			}
		} else {
			for (int id : ldSets.getPositiveNodes(imn)) {
				if (!ldSets.getNegativeNodes(imn).contains(id)) {
					newCol.set(id + addNum, 0, -1);
				}
			}
			for (int id : ldSets.getNegativeNodes(imn)) {
				if (!ldSets.getPositiveNodes(imn).contains(id)) {
					newCol.set(id + addNum, 0, 1);
				}
			}

		}
		//			augMat2 = augMat.transpose();
		ldCol = ldCol.combine(0, 0, newCol);
		//			printMatrix(augMat2);
		//		printMatrix(ldCol);

		//		printMatrix(augMat2);
		return isAugMatLD(augMat, ldCol);
	}

	public static void updatePlaceLDForNewTran(InteractiveMatrixNetL imn, Petrinet pn, Integer tNewID) {
		//		SimpleMatrix incidenceMatrixInput = ipn.getIncidenceMatrixInput();
		//		SimpleMatrix incidenceMatrixOutput = ipn.getIncidenceMatrixOutput();

		Map<Transition, Integer> tranInciMap = new HashMap<Transition, Integer>();
		Map<Place, Integer> placeInciMap = new HashMap<Place, Integer>();
		SimpleMatrix inciMat = createInciMap(tranInciMap, placeInciMap, pn);
		SimpleMatrix augMat3 = getSCIncidenceMatrix2(inciMat, placeInciMap, pn);

		Transition tNew = imn.getTranBasedOnID(tNewID);
		int inciIndexOftNew = tranInciMap.get(tNew);

		SimpleMatrix inciMatWithoutNewLD = new SimpleMatrix(augMat3.numRows(), augMat3.numCols() - 1);
		SimpleMatrix newLDCol = new SimpleMatrix(augMat3.numRows(), 1);
		//				getArrayWithoutLastCol(incidenceMatrixInput.plus(incidenceMatrixOutput), withoutLastCol, lastCol);

		splitInciMat(augMat3, inciMatWithoutNewLD, newLDCol, inciIndexOftNew);
		PointValuePair mu = getNodesInBetween(inciMatWithoutNewLD, newLDCol, inciIndexOftNew);
		//for self loops -add all trans as choices
		//perf ends

		//performance starts
		Set<Integer> allChoiceTransWrtTNew = new HashSet<Integer>();
		for (Transition t : tranInciMap.keySet()) {
			double val = mu.getKey()[tranInciMap.get(t)];
			if (val != 0.)
				//			if (allChoiceTransWrtTNewInciMat.contains(tranInciMap.get(t)))
				allChoiceTransWrtTNew.add(imn.getTranId(t)); //get the original tran id
		}
		//performance starts

		List<LDSetsL> ldPlaces = imn.getLdPlaceInfo();
		int ldpSize = ldPlaces.size();
		//performance ends
		for (int ldp = 0; ldp < ldpSize; ldp++) {
			int r = ldPlaces.get(ldp).getID();
			Set<Transition> tStart = new HashSet<Transition>();
			Set<Transition> tEnd = new HashSet<Transition>();

			for (int c : ldPlaces.get(ldp).getPositiveNodes(imn))
				tStart.add(imn.getTranBasedOnID(c));
			for (int c : ldPlaces.get(ldp).getNegativeNodes(imn))
				tEnd.add(imn.getTranBasedOnID(c));

			//Perf changes start
			int ind = 0;
			for (int id : ldPlaces.get(ldp).getPositiveNodes(imn)) {
				if (allChoiceTransWrtTNew.contains(id))
					ind += mu.getKey()[tranInciMap.get(imn.getTranBasedOnID(id))];
			}
			for (int id : ldPlaces.get(ldp).getNegativeNodes(imn)) {
				if (allChoiceTransWrtTNew.contains(id))
					ind -= mu.getKey()[tranInciMap.get(imn.getTranBasedOnID(id))];
			}

			Set<Integer> negativeTransForNewLD = new HashSet<Integer>(ldPlaces.get(ldp).getNegativeNodes(imn));
			Set<Integer> positiveTransForNewLD = new HashSet<Integer>(ldPlaces.get(ldp).getPositiveNodes(imn));
			//			switch (ind) {
			if (ind == 0) {
				//if zero is LD, check if its also FC, if not, add self loop wrt new tran
				if (!FCnetsUtilsL.checkFCForLDPlaces(pn, tStart, tEnd)) {
					tStart.add(imn.getTranBasedOnID(tNewID));
					tEnd.add(imn.getTranBasedOnID(tNewID));
					if (FCnetsUtilsL.checkFCForLDPlaces(pn, tStart, tEnd)) {
						negativeTransForNewLD.add(tNewID); //perf
						positiveTransForNewLD.add(tNewID); //perf
						if (!imn.getInputOutTransitions().contains(negativeTransForNewLD))
							imn.getInputOutTransitions().add(negativeTransForNewLD);
						if (!imn.getInputOutTransitions().contains(positiveTransForNewLD))
							imn.getInputOutTransitions().add(positiveTransForNewLD);
						imn.addLdPlaceInfo(imn.getLdPlaceInfo().size(),
								imn.getInputOutTransitions().indexOf(negativeTransForNewLD), imn
										.getInputOutTransitions().indexOf(positiveTransForNewLD)); //performance
					}
				}
				//Also keep the prev LD non-FC place - for future 
				//			break;
			}
			if (ind > 0) {
				positiveTransForNewLD.add(tNewID);
				if (!imn.getInputOutTransitions().contains(positiveTransForNewLD))
					imn.getInputOutTransitions().add(positiveTransForNewLD);
				ldPlaces.get(ldp).updatePositiveId(imn.getInputOutTransitions().indexOf(positiveTransForNewLD)); //perf
				//			break;
			}
			if (ind < 0) {
				negativeTransForNewLD.add(tNewID);
				if (!imn.getInputOutTransitions().contains(negativeTransForNewLD))
					imn.getInputOutTransitions().add(negativeTransForNewLD);
				ldPlaces.get(ldp).updateNegativeId(imn.getInputOutTransitions().indexOf(negativeTransForNewLD)); //perf
			}
			//			break;
			//				default :
			//					break;
			//			}

			//Perf changes end
		}

		updatePlaceLDNonSiphons(imn, pn);
		updateTranLDFreeChoices(imn, pn, null);
	}

	//	private static Set<Integer> getAllLDNodes(SimpleMatrix augMat, Set<Integer> rrefCols, int subNum) {
	//		Map<Integer, Integer> mapIndexMap = new HashMap<Integer, Integer>();
	//		SimpleMatrix inciMat = keepOnlyRREFVals(augMat, rrefCols, mapIndexMap);
	//
	//		Map<Integer, Set<Integer>> mapNodeRRefVals = new HashMap<Integer, Set<Integer>>();
	//		for (int c = 0; c < augMat.numCols(); c++) {
	//			if (!rrefCols.contains(c)) {
	//				SimpleMatrix newCol = new SimpleMatrix(inciMat.numRows(), 1);
	//				for (int r = 0; r < augMat.numRows(); r++) {
	//					newCol.set(r, 0, augMat.get(r, c));
	//				}
	//				mapNodeRRefVals.put(c, new HashSet<Integer>(getNodesInBetween(inciMat, newCol, 0)));
	//			}
	//		}
	//		Set<Integer> allChoices = new HashSet<Integer>();
	//		int lastCol = augMat.numCols() - 1;
	//		for (Integer key : mapNodeRRefVals.keySet()) {
	//			if (key == lastCol)
	//				for (Integer id : mapNodeRRefVals.get(key)) {
	//					allChoices.add(mapIndexMap.get(id) - subNum);
	//				}
	//			else if ((mapNodeRRefVals.get(lastCol).containsAll(mapNodeRRefVals.get(key)) || mapNodeRRefVals
	//					.get(lastCol).equals(mapNodeRRefVals.get(key)))) {
	//				allChoices.add(key - subNum);
	//			}
	//
	//		}
	//		return allChoices;
	//	}

	private static SimpleMatrix keepOnlyRREFVals(SimpleMatrix inciMat, Set<Integer> rrefCols,
			Map<Integer, Integer> matIndexMap) {
		//		SimpleMatrix inciMat = new SimpleMatrix(withoutLastCol2);
		SimpleMatrix inciMat2 = new SimpleMatrix(0, 0);
		for (int c = 0; c < inciMat.numCols(); c++) {
			if (rrefCols.contains(c)) {
				SimpleMatrix newCol = new SimpleMatrix(inciMat.numRows(), 1);
				matIndexMap.put(inciMat2.numCols(), c);
				for (int r = 0; r < inciMat.numRows(); r++) {
					newCol.set(r, 0, inciMat.get(r, c));
				}
				inciMat2 = inciMat2.combine(0, inciMat2.numCols(), newCol);
			}
		}
		return inciMat2;
	}

	public static LDSetsL addSelfLoop(InteractiveMatrixNetL imn, Integer rowID) {
		//		SimpleMatrix tranLDInput = ipn.getTranLDInput();
		//		SimpleMatrix tranLDOutput = ipn.getTranLDOutput();
		//		SimpleMatrix newColInput = new SimpleMatrix(tranLDInput.numRows(), 1);
		//		SimpleMatrix newColOutput = new SimpleMatrix(tranLDOutput.numRows(), 1);
		//		newColInput.set(rowID, 0, 1);
		//		newColOutput.set(rowID, 0, -1);
		//perf
		Set<Integer> negativePlacesForNewLD = new HashSet<Integer>();
		Set<Integer> positivePlacesForNewLD = new HashSet<Integer>();
		negativePlacesForNewLD.add(rowID);
		positivePlacesForNewLD.add(rowID);
		if (!imn.getInputOutPlaces().contains(negativePlacesForNewLD))
			imn.getInputOutPlaces().add(negativePlacesForNewLD);
		LDSetsL ldt = new LDSetsL(imn.getLdTranInfo().size(), imn.getInputOutPlaces().indexOf(negativePlacesForNewLD),
				imn.getInputOutPlaces().indexOf(positivePlacesForNewLD), false);
		ldt.setID(ldt.hashCode());
		imn.setLdPlaceMapForTranLD(rowID, ldt);
		imn.setLdTranInfo(ldt); //performance
		//		updateTranLDFreeChoices(ipn, ldt);
		//perf
		//		tranLDInput = tranLDInput.combine(0, tranLDInput.numCols(), newColInput);
		//		tranLDOutput = tranLDOutput.combine(0, tranLDOutput.numCols(), newColOutput);
		//		ipn.setTranLDMatrixInput(tranLDInput);
		//		ipn.setTranLDMatrixOutput(tranLDOutput);
		//		printMatrix(tranLDInput);
		//		printMatrix(tranLDOutput);
		return ldt;
	}

	public static void initiateMidToEndTran(InteractiveMatrixNetL imn) {
		//perf
		Set<Integer> negativePlacesForNewLD = new HashSet<Integer>();
		Set<Integer> positivePlacesForNewLD = new HashSet<Integer>();
		negativePlacesForNewLD.add(1);
		positivePlacesForNewLD.add(2);
		LDSetsL ldt = new LDSetsL(imn.getLdTranInfo().size(), imn.getInputOutPlaces().indexOf(negativePlacesForNewLD),
				imn.getInputOutPlaces().indexOf(positivePlacesForNewLD), false);
		ldt.setID(ldt.hashCode());
		imn.setLdPlaceMapForTranLD((Integer) 1, ldt);
		imn.setLdPlaceMapForTranLD((Integer) 2, ldt);
		imn.setLdTranInfo(ldt); //performance

		//reverse
		LDSetsL ldt2 = new LDSetsL(imn.getLdTranInfo().size(), imn.getInputOutPlaces().indexOf(positivePlacesForNewLD),
				imn.getInputOutPlaces().indexOf(negativePlacesForNewLD), false);
		ldt2.setID(ldt2.hashCode());
		imn.setLdPlaceMapForTranLD((Integer) 1, ldt2);
		imn.setLdPlaceMapForTranLD((Integer) 2, ldt2);
		imn.setLdTranInfo(ldt2); //performance

	}

	//Ax=b : A- inci, b- linearly dependent 
	public static PointValuePair getNodesInBetween(SimpleMatrix inciMatrixWithoutLastCol, SimpleMatrix linDepMatrix,
			int indexOfNewNodeInInciMat) {
		Set<Integer> ldCols = new HashSet<Integer>();
		double[] coeff = new double[inciMatrixWithoutLastCol.numCols()];
		//		double[] coeff = getCoeff(inciMatrixWithoutLastCol,subNum);
		LinearObjectiveFunction f = new LinearObjectiveFunction(coeff, 0);
		Collection<LinearConstraint> constraints = new ArrayList<LinearConstraint>();
		//		Collection<LinearConstraint> constraints_Loop = new ArrayList<LinearConstraint>();
		//System.out.println("x");
		//		printMatrix(inciMatrixWithoutLastCol);
		//		printMatrix(linDepMatrix);
		for (int r = 0; r < inciMatrixWithoutLastCol.numRows(); r++) {
			double[] inciMat = new double[inciMatrixWithoutLastCol.numCols()];
			for (int c = 0; c < inciMatrixWithoutLastCol.numCols(); c++) {
				inciMat[c] = inciMatrixWithoutLastCol.get(r, c);
			}
			constraints.add(new LinearConstraint(inciMat, Relationship.EQ, linDepMatrix.get(r, 0)));
			//			constraints_Loop.add(new LinearConstraint(inciMat, Relationship.EQ, -linDepMatrix.get(r, 0)));
		}
		SimplexSolver solver = new SimplexSolver();
		PointValuePair solution;

		solution = solver.optimize(new MaxIter(1000), f, new LinearConstraintSet(constraints), GoalType.MINIMIZE,
				new NonNegativeConstraint(true));

		//		 solution = solution.getPoint();
		//		System.out.println(solution.getPoint());

		for (int c = 0; c < solution.getKey().length; c++) {
			double val = solution.getKey()[c];
			if (val != 0.)
				//				if (val != 1.)
				//					System.out.println("Error!!");
				//				else {
				if (c >= indexOfNewNodeInInciMat) {
					ldCols.add(c + 1); //since ld col is removed from position 'indexOfNewNodeInInciMat', add 1 to compensate 
				} else {
					ldCols.add(c);
				}
			//				}
		}
		return solution;
	}

	public static boolean isAugMatLD(SimpleMatrix inciMatrixWithoutLastCol, SimpleMatrix linDepMatrix) {
		SimpleMatrix inciMatrix = new SimpleMatrix(inciMatrixWithoutLastCol);
		double[] coeff = new double[inciMatrix.numCols()];
		LinearObjectiveFunction f = new LinearObjectiveFunction(coeff, 0);
		Collection<LinearConstraint> constraints = new ArrayList<LinearConstraint>();
		boolean isLD = true;
		for (int r = 0; r < inciMatrix.numRows(); r++) {
			double[] inciMat = new double[inciMatrix.numCols()];
			for (int c = 0; c < inciMatrix.numCols(); c++) {
				inciMat[c] = inciMatrix.get(r, c);
			}
			constraints.add(new LinearConstraint(inciMat, Relationship.EQ, linDepMatrix.get(r, 0)));
		}
		SimplexSolver solver = new SimplexSolver();
		PointValuePair solution = null;
		try {
			solution = solver.optimize(new MaxIter(100), f, new LinearConstraintSet(constraints), GoalType.MINIMIZE,
					new NonNegativeConstraint(true));
		} catch (NoFeasibleSolutionException ex) {
			isLD = false;
		}
		return isLD;
	}

	public static boolean isAugMatLDNeg(SimpleMatrix inciMatrixWithoutLastCol, SimpleMatrix linDepMatrix) {
		SimpleMatrix inciMatrix = new SimpleMatrix(inciMatrixWithoutLastCol);
		double[] coeff = new double[inciMatrix.numCols()];
		LinearObjectiveFunction f = new LinearObjectiveFunction(coeff, 0);
		Collection<LinearConstraint> constraints = new ArrayList<LinearConstraint>();
		boolean isLD = true;
		for (int r = 0; r < inciMatrix.numRows(); r++) {
			double[] inciMat = new double[inciMatrix.numCols()];
			for (int c = 0; c < inciMatrix.numCols(); c++) {
				inciMat[c] = inciMatrix.get(r, c);
			}
			constraints.add(new LinearConstraint(inciMat, Relationship.EQ, linDepMatrix.get(r, 0)));
		}
		SimplexSolver solver = new SimplexSolver();
		PointValuePair solution = null;
		try {
			solution = solver.optimize(new MaxIter(50), f, new LinearConstraintSet(constraints), GoalType.MINIMIZE,
					new NonNegativeConstraint(false));
		} catch (NoFeasibleSolutionException ex) {
			isLD = false;
		}
		return isLD;
	}

	public static void updateHistoryTab(InteractivePetriNetL ipn) {
		if (ipn.isKeepHistory()) {
			int maxInd = ipn.getImnIndex() - 2;
			if (maxInd < 1)
				return;
			for (int i = 0; i < maxInd; i++) {
				ipn.getImnList().get(i).getAllTranLDFCs().clear();
				ipn.getImnList().get(i).getAllLdPlaceMapForTranLD().clear();
				ipn.getImnList().get(i).getAllPlaceId().clear();
				ipn.getImnList().get(i).getAllTranId().clear();

			}
		}

	}

	public static SimpleMatrix createSequence(Map<Transition, Integer> tranInciMap, Set<Transition> trans,
			SimpleMatrix im) {

		SimpleMatrix seq = new SimpleMatrix(tranInciMap.size(), 1);
		return seq;
	}

	public static SimpleMatrix createInitialAnOrFinalMarking(Map<Place, Integer> placeInciMap, Petrinet pn,
			Set<Place> marking) {
		SimpleMatrix M = new SimpleMatrix(placeInciMap.size(), 1);
		for (Place p : marking)
			M.set(placeInciMap.get(p), 0, 1);
		return M;
	}

	//transposed incidence matrix, transposed mf-mi matrix, and coeff values
	public static boolean canTraceBeReplayed(SimpleMatrix t_Inci, SimpleMatrix t_MfMinusMi, Map<Integer, Integer> coeffs) {
		SimpleMatrix inciMatrix = new SimpleMatrix(t_Inci);
		double[] coeff = new double[inciMatrix.numRows()];
		//		for (int id : coeffs.keySet())
		//			coeff[id] = coeffs.get(id);

		System.out.println(t_Inci);
		System.out.println(t_MfMinusMi);

		LinearObjectiveFunction f = new LinearObjectiveFunction(coeff, 1);
		Collection<LinearConstraint> constraints = new ArrayList<LinearConstraint>();
		boolean isLD = true;
		for (int c = 0; c < inciMatrix.numCols(); c++) {

			double[] inciMat = new double[inciMatrix.numRows()];
			for (int r = 0; r < inciMatrix.numRows(); r++) {
				inciMat[r] = inciMatrix.get(r, c) * coeffs.get(r); //multiply by the coefficient directly here - the zero columns would nullified
			}
			constraints.add(new LinearConstraint(inciMat, Relationship.GEQ, t_MfMinusMi.get(0, c)));
			System.out.println(inciMat);
			System.out.println(t_MfMinusMi.get(0, c));

		}
		SimplexSolver solver = new SimplexSolver();
		PointValuePair solution = null;

		try {
			solution = solver.optimize(new MaxIter(50), f, new LinearConstraintSet(constraints), GoalType.MINIMIZE,
					new NonNegativeConstraint(true));
			for (int c = 0; c < solution.getKey().length; c++) {
				double val = solution.getKey()[c];
				//				System.out.println(val);
			}

		} catch (NoFeasibleSolutionException ex) {
			isLD = false;
		}
		return isLD;
	}

}
