package org.processmining.prabhakardixit.fcnets.models;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Set;

import org.processmining.models.graphbased.directed.petrinet.elements.Place;
import org.processmining.models.graphbased.directed.petrinet.elements.Transition;
import org.processmining.plugins.petrinet.replayresult.PNRepResult;
import org.processmining.projectedrecallandprecision.result.ProjectedRecallPrecisionResult;
import org.processmining.reduction.ReductionTrackerL;

public class InteractiveMatrixNetL {
	private Map<Place, Integer> placeId = new HashMap<Place, Integer>(); //id corresponding node
	private Map<Transition, Integer> tranId = new HashMap<Transition, Integer>();
	private Set<Integer> placeLDNonSiphonIndex = new HashSet<Integer>(); //Indices of non siphon valid places
	private Set<Integer> tranLDFCs = new HashSet<Integer>(); //Indices of FC valid Trans

	private List<Set<Integer>> inputOutPlaces = new ArrayList<Set<Integer>>(); // all the possible places combinations - use these to reference from LDSets
	private List<Set<Integer>> inputOutTrans = new ArrayList<Set<Integer>>(); // all the possible transitions combinations - use these to reference from LDSets

	private List<LDSetsL> ldPlaceInfo = new ArrayList<LDSetsL>(); //performance
	private List<LDSetsL> ldTranInfo = new ArrayList<LDSetsL>(); //performance
	private PetriNetMapL selectedPetriNetMap; //Petri net object corresponding to this imn
	//	private Graph graph; //Prefuse graph obj

	private Map<Integer, Set<Integer>> ldPlaceMapForTranLD = new HashMap<Integer, Set<Integer>>(); //performance for icnremental disc

	private ReductionTrackerL currTrack = new ReductionTrackerL(); //re-do current to go back
	private ReductionTrackerL nextTrack = new ReductionTrackerL(); //if gone back, press next to go forward

	//	private Map<Set<Integer>, Set<Integer>> mapLDTrans = new HashMap<Set<Integer>, Set<Integer>>();

	//aggregagted alignment starts
	private Map<Transition, Integer> aggrSyncMoves = new HashMap<Transition, Integer>();
	private Map<Transition, Integer> aggrModelMoves = new HashMap<Transition, Integer>();
	private Map<String, Integer> aggrLogMoves = new HashMap<String, Integer>();

	//replay object - predictor mode
	private PNRepResult replayReult;
	private ProjectedRecallPrecisionResult projectedResult;
	float[] fitPre = new float[2];

	//aggregagted alignment ends	
	//	Set<LDSets> tranLDFCs = new HashSet<LDSets>(); //Indices of FC valid Trans

	//	Set<TransPairAlignment> tranPairs = new HashSet<TransPairAlignment>(); 
	//	List<TraceLevelPairwiseInfo> taceLevelPairwiseInfo = new ArrayList<TraceLevelPairwiseInfo>();// pairwise alignment info obj		

	//	public Graph getGraph() {
	//		return graph;
	//	}
	//
	//	public void setGraph(Graph graph) {
	//
	//		this.graph = graph;
	//	}

	Map<Transition, Set<Transition>> tranPostTran = new HashMap<Transition, Set<Transition>>();

	public InteractiveMatrixNetL() {

	}

	public InteractiveMatrixNetL(InteractiveMatrixNetL imn, boolean createNewPn) {
		PetriNetMapL pnmapOrg = imn.getselectedPetriNetMap();
		if (createNewPn) {
			selectedPetriNetMap = new PetriNetMapL(pnmapOrg);
			for (Place p : pnmapOrg.getPetriNet().getPlaces()) {
				placeId.put(selectedPetriNetMap.getMapPlaces().get(p), imn.getPlaceId(p));
			}
			for (Transition t : pnmapOrg.getPetriNet().getTransitions()) {
				tranId.put(selectedPetriNetMap.getMapTransitions().get(t), imn.getTranId(t));
			}
		} else {
			//change to not update the petrinet map starts - align
			selectedPetriNetMap = pnmapOrg;
			for (Place p : pnmapOrg.getPetriNet().getPlaces()) {
				placeId.put(p, imn.getPlaceId(p));
			}
			for (Transition t : pnmapOrg.getPetriNet().getTransitions()) {
				tranId.put(t, imn.getTranId(t));
			}
		}
		//change to not update the petrinet map ends - align
		placeLDNonSiphonIndex = new HashSet<Integer>(imn.getPlaceLDNonSiphonIndex());
		tranLDFCs = new HashSet<Integer>(imn.getTranLDFCs());
		//		mapLDTrans = new HashMap<Set<Integer>, Set<Integer>>(imn.getMapLDTrans());
		for (Integer key : imn.ldPlaceMapForTranLD.keySet())
			ldPlaceMapForTranLD.put(key, new HashSet<Integer>(imn.getLdPlaceMapForTranLDIndex(key)));
		if (!createNewPn) { //if not automated discovery
			ldPlaceInfo = new ArrayList<LDSetsL>(imn.getLdPlaceInfo());
			ldTranInfo = new ArrayList<LDSetsL>(imn.getLdTranInfo());

		} else {
			ldPlaceInfo = new ArrayList<LDSetsL>();
			for (int i = 0; i < imn.getLdPlaceInfo().size(); i++) {
				LDSetsL lds = imn.getLdPlaceInfo().get(i);
				LDSetsL newlds = new LDSetsL(lds.getID(), lds.getNegativeId(), lds.getPositiveId(), lds.LDplaces);
				ldPlaceInfo.add(newlds);
			}
			ldTranInfo = new ArrayList<LDSetsL>();
			for (int i = 0; i < imn.getLdTranInfo().size(); i++) {
				LDSetsL lds = imn.getLdTranInfo().get(i);
				ldTranInfo.add(new LDSetsL(lds.getID(), lds.getNegativeId(), lds.getPositiveId(), lds.LDplaces));
			}
			//			for (Integer key : imn.ldPlaceMapForTranLD.keySet()) {
			//				//TODO check if this needs to be updated
			//				//				ldPlaceMapForTranLD.put(key, new HashSet<LDSets>(imn.getLdPlaceMapForTranLD(key)));
			//				Set<LDSets> allSets = new HashSet<LDSets>();
			//				for (int i = 0; i < imn.getLdPlaceMapForTranLD(key).size(); i++) {
			//					for (LDSets lds : imn.getLdPlaceMapForTranLD(key))
			//						allSets.add(new LDSets(lds.getID(), lds.getNegativeId(), lds.getPositiveId(), lds.LDplaces));
			//				}
			//				ldPlaceMapForTranLD.put(key, allSets);
			//			}
		}

		//		setGraph(imn.getGraph());

		//aggr alignment
		aggrSyncMoves = new HashMap<Transition, Integer>(imn.getAggrSyncMoves());
		aggrLogMoves = new HashMap<String, Integer>(imn.getAggrLogMoves());

		inputOutPlaces = new ArrayList<Set<Integer>>(imn.getInputOutPlaces());
		inputOutTrans = new ArrayList<Set<Integer>>(imn.getInputOutTransitions());
	}

	public Set<Integer> getPlaceLDNonSiphonIndex() {
		return placeLDNonSiphonIndex;
	}

	public Set<Integer> getTranLDFCs() {
		return tranLDFCs;
	}

	public List<LDSetsL> getplaceLDNonSiphons() {
		List<LDSetsL> placeLDAll = this.getLdPlaceInfo();
		List<LDSetsL> placeLDNonSiphons = new ArrayList<LDSetsL>();
		for (int r = 0; r < placeLDAll.size(); r++) {
			if (this.placeLDNonSiphonIndex.contains(placeLDAll.get(r).getID())) {
				placeLDNonSiphons.add(placeLDAll.get(r));
			}
		}
		return placeLDNonSiphons;
	}

	public Set<Integer> getTranLDFreeChoice() {
		return tranLDFCs;
	}

	public Integer getPlaceId(Place p) {
		return placeId.get(p);
	}

	public void setPlaceId(Place p, Integer i) {
		placeId.put(p, i);
	}

	public Map<Place, Integer> getAllPlaceId() {
		return placeId;
	}

	public Integer getTranId(Transition t) {
		return tranId.get(t);
	}

	public void setTranId(Transition t, Integer i) {
		tranId.put(t, i);
	}

	public Map<Transition, Integer> getAllTranId() {
		return tranId;
	}

	public Place getPlaceBasedOnID(int id) {
		for (Place p : placeId.keySet()) {
			System.out.println(placeId.get(p));
			if (placeId.get(p).equals(id))
				return p;
		}
		return null;
	}

	//tranId in viz is tran+1000
	public Transition getTranBasedOnID(int id) {
		for (Transition t : tranId.keySet()) {
			if (tranId.get(t).equals(id))
				return t;
		}
		return null;
	}

	public PetriNetMapL getselectedPetriNetMap() {
		return selectedPetriNetMap;
	}

	public void setselectedPetriNetMap(PetriNetMapL pnMap) {
		this.selectedPetriNetMap = pnMap;
		//		pnMap.mapNodes();
		//PetrinetUtils.createLoopChoices(pnMap);

	}

	public Map<Transition, Set<Transition>> getTranPostTran() {
		return tranPostTran;
	}

	public Set<Transition> getTranPostTran(Transition t) {
		return tranPostTran.get(t);
	}

	public Set<String> getTranPostTranString(Transition t) {
		Set<String> postTranStr = new HashSet<String>();
		for (Transition tr : tranPostTran.get(t)) {
			postTranStr.add(tr.toString());
		}
		return postTranStr;
	}

	public void setTranPostTran(Map<Transition, Set<Transition>> tranPostTran) {
		this.tranPostTran = tranPostTran;
	}

	public void addTranPostTran(Transition t, Set<Transition> tSet) {
		this.tranPostTran.put(t, tSet);
	}

	public List<LDSetsL> getLdPlaceInfo() {
		return ldPlaceInfo;
	}

	public void setLdPlaceInfo(List<LDSetsL> ldPlaceInfo) {
		this.ldPlaceInfo = ldPlaceInfo;
	}

	//	public void addLdPlaceInfo(int rowId, Set<Integer> negativeTrans, Set<Integer> positiveTrans) {
	//		this.ldPlaceInfo.add(new LDSets(rowId, negativeTrans, positiveTrans));
	//	}
	public void addLdPlaceInfo(int rowId, int negativeid, int positiveid) {
		this.ldPlaceInfo.add(new LDSetsL(rowId, negativeid, positiveid, true));
	}

	public List<LDSetsL> getLdTranInfo() {
		return ldTranInfo;
	}

	public void setLdTranInfo(List<LDSetsL> ldTranInfo) {
		this.ldTranInfo = ldTranInfo;
	}

	public void addLdTranInfo(int colId, int negativeid, int positiveid) {
		this.ldTranInfo.add(new LDSetsL(colId, negativeid, positiveid, false));
	}

	public Set<LDSetsL> getLdPlaceMapForTranLD(Integer key) {
		Set<LDSetsL> ldSets = new HashSet<LDSetsL>();
		for (LDSetsL lds : ldTranInfo) {
			if (ldPlaceMapForTranLD.get(key).contains(lds.getID()))
				ldSets.add(lds);
		}
		return ldSets;
	}

	public Set<Integer> getLdPlaceMapForTranLDIndex(Integer key) {
		return ldPlaceMapForTranLD.get(key);
	}

	public void setLdPlaceMapForTranLD(Integer key, LDSetsL e) {
		if (ldPlaceMapForTranLD.containsKey(key)) {
			ldPlaceMapForTranLD.get(key).add((Integer) e.getID());
		} else {
			Set<Integer> lds = new HashSet<Integer>();
			lds.add((Integer) (e.getID()));
			ldPlaceMapForTranLD.put(key, lds);
		}
	}

	public void removeLdPlaceMapForTranLD(Integer key, LDSetsL lds) {
		//		ldPlaceMapForTranLD.remove(key, lds); /* Requires Java 8 */
		if (ldPlaceMapForTranLD.containsKey(key) && (ldPlaceMapForTranLD.get(key).contains(lds.getID()))) {
			ldPlaceMapForTranLD.remove(key);
		}
	}

	public void setLdTranInfo(LDSetsL ldset) {
		ldTranInfo.add(ldset);
	}

	public Map<Integer, Set<Integer>> getAllLdPlaceMapForTranLD() {
		return ldPlaceMapForTranLD;
	}

	public Set<LDSetsL> getAllTranLDFCs() {
		Set<LDSetsL> ldset = new HashSet<LDSetsL>();
		for (LDSetsL lds : ldTranInfo) {
			if (tranLDFCs.contains(lds.getID()))
				ldset.add(lds);
		}
		return ldset;
	}

	public ReductionTrackerL getCurrentRedTracker() {
		return currTrack;
	}

	public ReductionTrackerL getNextRedTracker() {
		return nextTrack;
	}

	public void setCurrentRedTracker(ReductionTrackerL redTrack) {
		this.currTrack = redTrack;
	}

	public void setNextRedTracker(ReductionTrackerL redTrack) {
		this.nextTrack = redTrack;
	}

	//	public Map<Set<Integer>, Set<Integer>> getMapLDTrans() {
	//		return mapLDTrans;
	//	}

	//aggr alignment starts
	public Map<Transition, Integer> getAggrSyncMoves() {
		return aggrSyncMoves;
	}

	public void setAggrSyncMoves(Map<Transition, Integer> aggrSyncMoves) {
		this.aggrSyncMoves = aggrSyncMoves;
	}

	public Map<String, Integer> getAggrLogMoves() {
		return aggrLogMoves;
	}

	public void setAggrLogMoves(Map<String, Integer> aggrLogMoves) {
		this.aggrLogMoves = aggrLogMoves;
	}

	public Map<Transition, Integer> getAggrModelMoves() {
		return aggrModelMoves;
	}

	public void setAggrModelMoves(Map<Transition, Integer> aggrModelMoves) {
		this.aggrModelMoves = aggrModelMoves;
	}

	//aggr alignment ends

	public List<Set<Integer>> getInputOutPlaces() {
		return inputOutPlaces;
	}

	public void setInputOutPlaces(List<Set<Integer>> inputOutPlaces) {
		this.inputOutPlaces = inputOutPlaces;
	}

	public List<Set<Integer>> getInputOutTransitions() {
		return inputOutTrans;
	}

	public void setInputOutTransitions(List<Set<Integer>> inputOutTrans) {
		this.inputOutTrans = inputOutTrans;
	}

	public PNRepResult getReplayReult() {
		return replayReult;
	}

	public void setReplayReult(PNRepResult replayReult) {
		this.replayReult = replayReult;
	}

	public ProjectedRecallPrecisionResult getProjectedResult() {
		return projectedResult;
	}

	public void setProjectedResult(ProjectedRecallPrecisionResult projectedResult) {
		this.projectedResult = projectedResult;
	}

	public float getFitness() {
		return fitPre[0];
	}

	public float getPre() {
		return fitPre[1];
	}

	public void setFitPre(float[] pair) {
		fitPre[0] = pair[0];
		fitPre[1] = pair[1];
	}

}
