package org.processmining.prabhakardixit.fcnets.synthesis;

import org.deckfour.xes.model.XLog;
import org.processmining.contexts.uitopia.UIPluginContext;
import org.processmining.decomposelogs.AggregatedHeuristicsL;
import org.processmining.models.graphbased.directed.petrinet.Petrinet;
import org.processmining.models.graphbased.directed.petrinet.PetrinetEdge;
import org.processmining.models.graphbased.directed.petrinet.PetrinetNode;
import org.processmining.models.graphbased.directed.petrinet.elements.Place;
import org.processmining.models.graphbased.directed.petrinet.elements.Transition;
import org.processmining.models.graphbased.directed.petrinet.impl.PetrinetImpl;
import org.processmining.prabhakardixit.fcnets.models.InteractivePetriNetL;
import org.processmining.prabhakardixit.fcnets.models.PetriNetMapL;
import org.processmining.prabhakardixit.utils.PetrinetUtilsL;

public class CreateIPNL {

	public static InteractivePetriNetL createIPNDataQuality(XLog log, UIPluginContext context, boolean alignmentPlugin,
			boolean predictorPlugin, boolean keepHistory, boolean dqPlugin) {
		InteractivePetriNetL ipn = createIPN(log, null, context, alignmentPlugin, predictorPlugin, keepHistory, false);
		ipn.setDataQuality(dqPlugin);

		return ipn;
	}

	//call when results needs to be checked exhaustively after each iteration
	public static InteractivePetriNetL createIPNBruteForce(XLog log, UIPluginContext context, boolean alignmentPlugin,
			boolean predictorPlugin, boolean keepHistory) {
		InteractivePetriNetL ipn = createIPN(log, null, context, alignmentPlugin, predictorPlugin, keepHistory, false);
		ipn.setBruteForceCalc(true);

		return ipn;
	}

	/**
	 * Create ipn and set alignmentMode, predictor mode etc
	 * 
	 * @param log
	 * @param context
	 * @param alignmentPlugin
	 * @param predictorPlugin
	 * @param keepHistoryPlugin
	 * @param decomposeLog
	 *            declare like
	 * @return
	 */
	public static InteractivePetriNetL createIPN(XLog log, AggregatedHeuristicsL ahl, UIPluginContext context,
			boolean alignmentPlugin, boolean predictorPlugin, boolean keepHistory, boolean declareAlignments) {
		Petrinet pnet = new PetrinetImpl("ipn");
		Place p1 = pnet.addPlace("start");
		Transition t1 = pnet.addTransition("DUMMY");
		Place p2 = pnet.addPlace("mid");
		Transition t2 = pnet.addTransition("DUMMY");
		Place p3 = pnet.addPlace("end");
		t1.setInvisible(true);
		t2.setInvisible(true);
		pnet.addArc(p1, t1);
		pnet.addArc(t1, p2);
		pnet.addArc(p2, t2);
		pnet.addArc(t2, p3);

		PetriNetMapL pn = new PetriNetMapL(pnet);
		for (Place p : pn.getPetriNet().getPlaces())
			if (!PetrinetUtilsL.initialOrFinalPlace(p, pn.getPetriNet())) {
				Place pI = PetrinetUtilsL.getInitialPlace(pn.getPetriNet());
				Place pO = PetrinetUtilsL.getFinalPlace(pn.getPetriNet());
			}
		Transition tNew1 = null;
		Transition tNew2 = null;
		for (Transition t : pn.getPetriNet().getTransitions()) {
			for (PetrinetEdge<? extends PetrinetNode, ? extends PetrinetNode> e : pn.getPetriNet().getInEdges(t)) {
				if (e.getSource().equals(PetrinetUtilsL.getInitialPlace(pn.getPetriNet())))
					tNew1 = t;
				else
					tNew2 = t;
			}
		}
		InteractivePetriNetL ipn = new InteractivePetriNetL(pn, log, ahl, context);
		ipn.setAlignmentMode(alignmentPlugin);
		ipn.setPredictorMode(predictorPlugin);
		ipn.setKeepHistory(keepHistory);
		ipn.setCalculateAlignments(alignmentPlugin);
		ipn.setDeclareAlignments(declareAlignments);

		return ipn;

	}
}
