package org.processmining.prabhakardixit.utils;

import java.util.HashMap;
import java.util.Iterator;
import java.util.Map;

import org.processmining.acceptingpetrinet.models.AcceptingPetriNet;
import org.processmining.framework.plugin.PluginContext;
import org.processmining.humbertogarciacaballero.models.model.NodeTypeL;
import org.processmining.humbertogarciacaballero.models.model.PNConstantsL;
import org.processmining.humbertogarciacaballero.models.model.SoundnessInformationL;
import org.processmining.humbertogarciacaballero.models.util.UtilsL;
import org.processmining.models.graphbased.AttributeMap;
import org.processmining.models.graphbased.directed.petrinet.Petrinet;
import org.processmining.models.graphbased.directed.petrinet.PetrinetEdge;
import org.processmining.models.graphbased.directed.petrinet.PetrinetGraph;
import org.processmining.models.graphbased.directed.petrinet.PetrinetNode;
import org.processmining.models.graphbased.directed.petrinet.elements.InhibitorArc;
import org.processmining.models.graphbased.directed.petrinet.elements.Place;
import org.processmining.models.graphbased.directed.petrinet.elements.ResetArc;
import org.processmining.models.graphbased.directed.petrinet.elements.Transition;
import org.processmining.models.semantics.petrinet.Marking;
import org.processmining.prabhakardixit.fcnets.models.InteractivePetriNetL;
import org.processmining.prabhakardixit.fcnets.synthesis.VizGraphParametersUtilsL;

import prefuse.data.Edge;
import prefuse.data.Graph;
import prefuse.data.Node;
import prefuse.data.Tuple;
import prefuse.data.expression.ColumnExpression;
import prefuse.data.expression.ComparisonPredicate;
import prefuse.data.expression.ObjectLiteral;

public class GraphUtilsL {

	@SuppressWarnings("unchecked")
	public static Graph netToGraph(PluginContext context, AcceptingPetriNet n, SoundnessInformationL information,
			Map<PetrinetNode, Integer> idMap) {
		Graph graph = netToGraph(context, n.getNet(), null, idMap);
		Iterator<Place> iter = n.getInitialMarking().iterator();
		while (iter.hasNext()) {
			Place p = iter.next();
			Iterator<Tuple> iter2 = graph.getNodes().tuples(
					new ComparisonPredicate(ComparisonPredicate.EQ, new ColumnExpression(PNConstantsL.NODE_ID_KEY),
							new ObjectLiteral(p.getLocalID().toString())));

			if (iter2.hasNext()) {
				iter2.next().set(PNConstantsL.NODE_STARTING_KEY, true);
			}
		}

		Iterator<Marking> markings = n.getFinalMarkings().iterator();
		while (markings.hasNext()) {
			iter = markings.next().iterator();
			while (iter.hasNext()) {
				Place p = iter.next();
				Iterator<Tuple> iter2 = graph.getNodes().tuples(
						new ComparisonPredicate(ComparisonPredicate.EQ, new ColumnExpression(PNConstantsL.NODE_ID_KEY),
								new ObjectLiteral(p.getLocalID().toString())));

				if (iter2.hasNext()) {
					iter2.next().set(PNConstantsL.NODE_ENDING_KEY, true);
				}
			}
		}
		return graph;
	}

	//PMD IPN Starts
	public static Graph netToGraph(PluginContext context, PetrinetGraph d, Marking m, Map<PetrinetNode, Integer> idMap) {
		return netToGraph(context, d, m, null, idMap);
	}

	//PMD IPN Ends

	public static Graph netToGraph(PluginContext context, PetrinetGraph d, Marking m, InteractivePetriNetL ipn,
			Map<PetrinetNode, Integer> idMap) {
		Graph graph = new Graph(true);
		Map<PetrinetNode, Integer> nodeMap = new HashMap<>();
		UtilsL.initSchema(graph);
		Map<String, String> labels = preprocessDiagnosisPetrinet(context, (Petrinet) d);

		int V = d.getNodes().size();
		PetrinetNode[] nodes = d.getNodes().toArray(new PetrinetNode[V]);

		for (PetrinetNode node : nodes) {
			Node n = graph.addNode();
			nodeMap.put(node, n.getRow());
			if (node instanceof Place) {
				n.set(PNConstantsL.NODE_TYPE_KEY, NodeTypeL.PLACE.toString());
				n.set(PNConstantsL.NODE_LABEL_KEY, labels.get(node.getId().toString()));

				//PMD starts
				if (idMap != null) { //if id map provided, then use id from the idmap
					n.set("ID", idMap.get(node));
				} else {
					n.set("ID", ipn.getPlaceId((Place) node));
					n.set("InputOutputNodes",
							VizGraphParametersUtilsL.addHighlightPlacesBasedOnLDMatrix((Place) node, ipn));
				}
				//PMD ends

			} else {
				//PMD starts
				if (idMap != null) { //if id map provided, then use id from the idmap
					n.set("ID", idMap.get(node));
				} else {
					n.set("ID", (ipn.getTranId((Transition) node)));
					n.set("InputOutputNodes",
							VizGraphParametersUtilsL.addHighlightTransBasedOnLDMatrix((Transition) node, ipn));
				}
				//PMD ends
				n.set(PNConstantsL.NODE_TYPE_KEY, NodeTypeL.TRANSITION.toString());
				boolean tau;
				n.set(PNConstantsL.NODE_TAU_TRANSITION_KEY, (tau = ((Transition) node).isInvisible()));

				if (tau) {
					n.set(PNConstantsL.NODE_LABEL_KEY, "TAU");
				} else {
					n.set(PNConstantsL.NODE_LABEL_KEY, labels.get(node.getId().toString()));
				}

			}
			//PMD Starts
			if (ipn != null)
				if (node instanceof Place) {
					String str = "";
					n.set(PNConstantsL.NODE_TYPE_KEY, NodeTypeL.PLACE.toString());

				} else {
					//add Follows Precedes Coccors relations
					String strtemp = "";
					System.out.println(node);
					n.set(PNConstantsL.NODE_TYPE_KEY, NodeTypeL.TRANSITION.toString());
				}
			//PMD Ends
			n.setString(PNConstantsL.NODE_ID_KEY, node.getLocalID().localNodeID().toString());

		}

		for (PetrinetEdge<? extends PetrinetNode, ? extends PetrinetNode> edge : d.getEdges()) {
			int source = nodeMap.get(edge.getSource());
			int target = nodeMap.get(edge.getTarget());
			Edge e = graph.getEdge(graph.addEdge(source, target));
			if (edge instanceof ResetArc) {
				e.setBoolean(PNConstantsL.EDGE_RESET_KEY, true);
			} else if (edge instanceof InhibitorArc) {
				e.setBoolean(PNConstantsL.EDGE_INHIBITOR_KEY, true);
			}
			if (d.getArc(edge.getTarget(), edge.getSource()) != null) {
				e.setBoolean(PNConstantsL.EDGE_BIDIRECTIONAL, true);
				if (graph.getEdge(target, source) != -1) {
					e.setBoolean(PNConstantsL.EDGE_BIDIRECTIONAL_DRAW, false);
				}
			}
		}

		postprocessDiagnosisPetrinet(context, (Petrinet) d, labels);

		// Adding extra information
		if (m != null) {
			for (Place p : m) {
				graph.getNode(nodeMap.get(p)).setBoolean(PNConstantsL.NODE_STARTING_KEY, true);
			}
			markEndingNodes(graph);
		} else {
			markStartingNodes(graph);
			markEndingNodes(graph);
		}

		return graph;
	}

	private static void markEndingNodes(Graph g) {
		Iterator<Node> iter = UtilsL.getSinkNodes(g);
		while (iter.hasNext()) {
			Node node = iter.next();
			if (NodeTypeL.valueOf(node.getString(PNConstantsL.NODE_TYPE_KEY)) == NodeTypeL.PLACE)
				node.setBoolean(PNConstantsL.NODE_ENDING_KEY, true);
		}
	}

	private static void markStartingNodes(Graph g) {
		Iterator<Node> iter = UtilsL.getRootNodes(g);
		while (iter.hasNext()) {
			Node node = iter.next();
			if (NodeTypeL.valueOf(node.getString(PNConstantsL.NODE_TYPE_KEY)) == NodeTypeL.PLACE)
				node.setBoolean(PNConstantsL.NODE_STARTING_KEY, true);
		}
	}

	private static Map<String, String> preprocessDiagnosisPetrinet(PluginContext context, Petrinet net) {
		Map<String, String> labels = new HashMap<>();

		int label = 0;
		for (PetrinetNode node : net.getNodes()) {
			labels.put(node.getId().toString(), node.getLabel());
			node.getAttributeMap().put(AttributeMap.LABEL, String.valueOf(label++));
		}

		return labels;
	}

	private static void postprocessDiagnosisPetrinet(PluginContext context, Petrinet net, Map<String, String> labels) {
		for (PetrinetNode node : net.getNodes()) {
			node.getAttributeMap().put(AttributeMap.LABEL, String.valueOf(labels.get(node.getId().toString())));
		}
	}
}
