package org.processmining.prabhakardixit.utils;

import java.util.Collection;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Map;
import java.util.Set;

import org.processmining.models.graphbased.directed.petrinet.Petrinet;
import org.processmining.models.graphbased.directed.petrinet.PetrinetEdge;
import org.processmining.models.graphbased.directed.petrinet.PetrinetNode;
import org.processmining.models.graphbased.directed.petrinet.elements.Place;
import org.processmining.models.graphbased.directed.petrinet.elements.Transition;
import org.processmining.models.graphbased.directed.petrinet.impl.PetrinetImpl;
import org.processmining.models.semantics.petrinet.Marking;

import com.google.common.collect.HashMultiset;
import com.google.common.collect.Multiset;

public class PetrinetUtilsL {

	public static boolean containsEvent(Petrinet petrinet, String eventName) {
		boolean containsEvent = false;
		for (Transition t : petrinet.getTransitions()) {
			if (eventName.equals(t.toString()))
				containsEvent = true;
		}
		return containsEvent;
	}

	public static Petrinet duplicatePN(Petrinet pn) {
		Map<Transition, Transition> transitions = new HashMap<Transition, Transition>();
		Map<Place, Place> places = new HashMap<Place, Place>();
		return duplicatePN(pn, places, transitions);
	}

	public static Petrinet duplicatePN(Petrinet pn, Map<Place, Place> places, Map<Transition, Transition> transitions) {
		//Map<Transition, Transition> transitions = new HashMap<Transition, Transition>();
		Petrinet pnNew = new PetrinetImpl(pn.getLabel());
		for (Transition t : pn.getTransitions()) {
			Transition tNew = pnNew.addTransition(t.getLabel());
			if (t.isInvisible())
				tNew.setInvisible(true);
			transitions.put(t, tNew); //update transitions map
		}
		for (Place p : pn.getPlaces()) {
			Place pNew = pnNew.addPlace(p.getLabel());
			places.put(p, pNew); //update places map
			for (PetrinetEdge<? extends PetrinetNode, ? extends PetrinetNode> outEdges : pn.getOutEdges(p)) {
				Transition t = (Transition) outEdges.getTarget();
				pnNew.addArc(pNew, transitions.get(t));
			}
			for (PetrinetEdge<? extends PetrinetNode, ? extends PetrinetNode> inEdges : pn.getInEdges(p)) {
				Transition t = (Transition) inEdges.getSource();
				pnNew.addArc(transitions.get(t), pNew);
			}
		}
		return pnNew;
	}

	public static Set<String> getTransitionNames(Petrinet petrinet) {
		Set<String> transitionNames = new HashSet<String>();
		for (Transition transition : petrinet.getTransitions()) {
			if (transition.isInvisible())
				continue;
			else {
				transitionNames.add(transition.toString());
			}

		}
		return transitionNames;
	}

	public static boolean initialOrFinalPlace(Place p, Petrinet pn) {
		if (!initialPlace(p, pn) && !finalPlace(p, pn))
			return false;
		else
			return true;
	}

	private static boolean finalPlace(Place p, Petrinet pn) {
		//		if (p.getLabel().equals("o"))
		//			return true;
		//		else {
		Collection<PetrinetEdge<? extends PetrinetNode, ? extends PetrinetNode>> e = pn.getOutEdges(p);
		if (e.isEmpty())
			return true;
		else
			return false;
		//		}
	}

	public static Place getInitialPlace(Petrinet pn) {
		for (Place p : pn.getPlaces())
			if (initialPlace(p, pn))
				return p;
		return null;
	}

	public static Place getFinalPlace(Petrinet pn) {
		for (Place p : pn.getPlaces())
			if (finalPlace(p, pn))
				return p;
		return null;
	}

	public static Transition getInitialTranistion(Petrinet pn) {
		Transition t = null;
		Place pi = getInitialPlace(pn);
		for (PetrinetEdge<? extends PetrinetNode, ? extends PetrinetNode> e : pn.getOutEdges(pi)) {
			t = (Transition) e.getTarget();
		}
		return t;
	}

	public static Transition getFinalTranistion(Petrinet pn) {
		Transition t = null;
		Place pf = getFinalPlace(pn);
		for (PetrinetEdge<? extends PetrinetNode, ? extends PetrinetNode> e : pn.getInEdges(pf)) {
			t = (Transition) e.getSource();
		}
		return t;
	}

	public static boolean initialOrFinalTran(Petrinet pn, Transition t) {
		if (getFinalTranistion(pn).equals(t) || getInitialTranistion(pn).equals(t))
			return true;
		return false;
	}

	public static boolean initialPlace(Place p, Petrinet pn) {
		//		if (p.getLabel().equals("i"))
		//			return true;
		//		else {
		Collection<PetrinetEdge<? extends PetrinetNode, ? extends PetrinetNode>> e = pn.getInEdges(p);
		if (e.isEmpty())
			return true;
		else
			return false;
		//		}
	}

	public static boolean initialPlace(Set<Place> ps, Petrinet pn) {
		//		if (p.getLabel().equals("i"))
		//			return true;
		//		else {
		Place p = null;
		if (ps.size() > 1)
			return false;
		else {

			for (Place p1 : ps)
				p = p1;
		}
		Collection<PetrinetEdge<? extends PetrinetNode, ? extends PetrinetNode>> e = pn.getInEdges(p);
		if (e.isEmpty())
			return true;
		else
			return false;
		//		}
	}

	public static boolean initialOrFinalTransition(Transition t) {
		if (t.isInvisible() && (t.getLabel() == "taustart" || t.getLabel() == "tauend"))
			return true;
		else
			return false;
	}

	public static boolean initialTransition(Transition t) {
		if (t.isInvisible() && (t.getLabel() == "taustart"))
			return true;
		else
			return false;
	}

	public static boolean finalTransition(Transition t) {
		if (t.isInvisible() && (t.getLabel() == "tauend"))
			return true;
		else
			return false;
	}

	//Add new transition and place as seq after old place
	public static void addTransition(Place oldPlace, Place newPlace, Transition newTransition, Petrinet petriNet) {
		for (PetrinetEdge<? extends PetrinetNode, ? extends PetrinetNode> a : petriNet.getOutEdges(oldPlace)) {
			Transition tOutOld = (Transition) a.getTarget();
			petriNet.addArc(newPlace, tOutOld);
			petriNet.removeEdge(a);
		}
		petriNet.addArc(newTransition, newPlace);
		petriNet.addArc(oldPlace, newTransition);

	}

	public static void addTransitionBetweenPlaceAndTran(Place oldPlace, Transition oldTransition, Place newPlace,
			Transition newTransition, Petrinet petriNet) {
		addTransitionBetweenPlaceAndTran(oldPlace, oldTransition, newPlace, newTransition, petriNet, false);
	}

	//Add new transition between place and transition
	public static void addTransitionBetweenPlaceAndTran(Place oldPlace, Transition oldTransition, Place newPlace,
			Transition newTransition, Petrinet petriNet, boolean isSilent) {
		if (isSilent)
			newTransition.setInvisible(true);
		for (PetrinetEdge<? extends PetrinetNode, ? extends PetrinetNode> edge : petriNet.getOutEdges(oldPlace)) {
			if (edge.getTarget().equals(oldTransition)) {
				petriNet.removeEdge(edge);
				petriNet.addArc(newPlace, oldTransition);
			}
		}
		petriNet.addArc(newTransition, newPlace);
		petriNet.addArc(oldPlace, newTransition);
	}

	public static void addTransitionBetweenTranAndPlace(Transition oldTransition, Place oldPlace, Place newPlace,
			Transition newTransition, Petrinet petriNet) {
		for (PetrinetEdge<? extends PetrinetNode, ? extends PetrinetNode> edge : petriNet.getOutEdges(oldTransition)) {
			if (edge.getTarget().equals(oldPlace)) {
				petriNet.addArc(oldTransition, newPlace);
				petriNet.removeEdge(edge);
			}
		}
		petriNet.addArc(newPlace, newTransition);
		petriNet.addArc(newTransition, oldPlace);
	}

	public static Set<Transition> getEnabledTransitions(Petrinet ptnet, Set<Place> marking) {
		return getEnabledTransitions(ptnet, HashMultiset.create(marking));
	}

	public static Set<Transition> getEnabledTransitions(Petrinet ptnet, Multiset<Place> marking) {
		Set<Transition> enabledTransitions = new HashSet<Transition>();
		Collection<Transition> netTransitions = ptnet.getTransitions();
		for (Transition t : netTransitions) {
			Collection<PetrinetEdge<? extends PetrinetNode, ? extends PetrinetNode>> inflow = ptnet.getInEdges(t);
			boolean enabled = true;
			for (PetrinetEdge<? extends PetrinetNode, ? extends PetrinetNode> edge : inflow) {
				if (!marking.contains((edge.getSource()))) {
					enabled = false;
					break;
				}
			}
			if (enabled)
				enabledTransitions.add(t);
		}
		return enabledTransitions;
	}

	public static Place getMiddlePlace(Petrinet pn) {
		for (Place p : pn.getPlaces())
			if (getInitialPlace(pn) != p && getFinalPlace(pn) != p)
				return p;
		return null;

	}

	public static Set<String> getEnabledTransitionsStrings(Petrinet ptnet, Set<Place> marking) {
		Set<String> enabledTransitions = new HashSet<String>();
		Collection<Transition> netTransitions = ptnet.getTransitions();
		for (Transition t : netTransitions) {
			Collection<PetrinetEdge<? extends PetrinetNode, ? extends PetrinetNode>> inflow = ptnet.getInEdges(t);
			boolean enabled = true;
			for (PetrinetEdge<? extends PetrinetNode, ? extends PetrinetNode> edge : inflow) {
				if (!marking.contains((edge.getSource()))) {
					enabled = false;
					break;
				}
			}
			if (enabled)
				enabledTransitions.add(t.toString());
		}
		return enabledTransitions;
	}

	public static Marking getInitialMarking(Petrinet pn) {
		Marking m = new Marking();
		m.add(getInitialPlace(pn));
		return m;
	}

	public static Marking getFinalMarking(Petrinet pn) {
		Marking m = new Marking();
		m.add(getFinalPlace(pn));
		return m;
	}
}
