package org.processmining.reduction;

import java.util.ArrayList;
import java.util.Collection;
import java.util.Collections;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Set;

import org.ejml.simple.SimpleMatrix;
import org.processmining.framework.plugin.PluginContext;
import org.processmining.models.graphbased.directed.petrinet.Petrinet;
import org.processmining.models.graphbased.directed.petrinet.PetrinetEdge;
import org.processmining.models.graphbased.directed.petrinet.PetrinetNode;
import org.processmining.models.graphbased.directed.petrinet.elements.Place;
import org.processmining.models.graphbased.directed.petrinet.elements.Transition;
import org.processmining.prabhakardixit.fcnets.matrix.MatrixUtilsL;
import org.processmining.prabhakardixit.utils.PetrinetUtilsL;
import org.processmining.reduction.ReductionTrackerL.Rule;

public class ReductionL {
	private List<ReductionTrackerL> reductionTracker = new ArrayList<ReductionTrackerL>();
	private Map<Transition, Integer> tranMap = new HashMap<Transition, Integer>(); //id's of transitions
	private Map<Place, Integer> placeMap = new HashMap<Place, Integer>(); //ids of places
	private boolean getClusters;

	public ReductionL(Petrinet pn, PluginContext context) {
		getClusters = false;
		reduceNet(pn, context);
	}

	public ReductionL(Petrinet pn, Set<PetrinetNode> transToKeep) {
		getClusters = false;
		reduceNet(pn, transToKeep);
	}

	public ReductionL(Petrinet pn) {

		reduceNet(pn, new HashSet<PetrinetNode>());
	}

	public ReductionL(Petrinet pn, boolean keepChoices) {
		if (keepChoices)
			reduceNet_KeepAllChoices(pn);
	}

	public ReductionL() {

	}

	public List<ReductionTrackerL> getReductionTracker() {
		return reductionTracker;
	}

	public void setReductionTracker(List<ReductionTrackerL> reductionTracker) {
		this.reductionTracker = reductionTracker;
	}

	public void reduceNet_KeepAllChoices(Petrinet pn) {
		boolean reduced = true;
		init(pn);
		Set<PetrinetNode> empty = new HashSet<PetrinetNode>();

		//first reduce using abstraction rule
		while (reduced) {
			reduced = useAbstractionRule(pn, empty);
			if (!reduced) //reduce using l.d. place rule
				reduced = useLDPlaceRule(pn);
		}
		reduced = useAbstractionRule(pn, empty);
	}

	public void reduceNet(Petrinet pn, Set<PetrinetNode> keepNodes) {
		boolean reduced = true;
		init(pn);

		//first reduce using abstraction rule
		while (reduced) {
			reduced = useAbstractionRule(pn, keepNodes);
			if (!reduced) //reduce using l.d. place rule
				reduced = useLDPlaceRule(pn);
			if (!reduced)
				reduced = reduceTranLDUsingLocalInfo(pn, keepNodes, false);
			//			if (!reduced)
			//				reduced = reduceTranLDUsingLocalInfo(pn, keepNodes, true); // explore all combinations
		}
		reduced = useAbstractionRule(pn, keepNodes);
	}

	//all strings that are collections
	public Set<Set<String>> reduceNetGetClusters(Petrinet pn) {
		boolean reduced = true;
		init(pn);
		Set<String> duplActs = new HashSet<String>();
		Set<String> duplActsTemp = new HashSet<String>();
		for (Transition t : pn.getTransitions())
			if (!t.isInvisible() && !duplActsTemp.add(t.getLabel()))
				duplActs.add(t.getLabel());

		Set<Set<Transition>> abstractionMap = new HashSet<Set<Transition>>();
		//first reduce using abstraction rule
		while (reduced)
			reduced = getAllClusters(pn, abstractionMap);

		boolean flag = true;
		Set<Set<Transition>> abstractionMap2 = new HashSet<Set<Transition>>(abstractionMap);
		while (flag) {
			flag = mergeCommonActs(abstractionMap2, abstractionMap);
		}

		Set<Set<String>> absMapStr = new HashSet<Set<String>>();
		for (Set<Transition> tSet : abstractionMap2) {
			Set<String> sSet = new HashSet<String>();
			boolean add = true;

			for (Transition t : tSet)
				if (!t.isInvisible())
					if (duplActs.contains(t.getLabel()))
						add = false;
					else
						sSet.add(t.getLabel());
			if (add)
				absMapStr.add(sSet);
		}

		return absMapStr;
	}

	private boolean mergeCommonActs(Set<Set<Transition>> abstractionMap2, Set<Set<Transition>> abstractionMap) {

		for (Set<Transition> t : abstractionMap)
			for (Set<Transition> t2 : abstractionMap2)
				if (!t.equals(t2)) {
					if (!Collections.disjoint(t, t2)) {
						t.addAll(t2);
						abstractionMap2.remove(t2);
						return true;
					}

				}
		return false;
	}

	private List<ReductionTrackerL> reduceNet(Petrinet pn, PluginContext context) {
		boolean reduced = true;

		init(pn);
		//update keepnodes
		Set<PetrinetNode> keepTrans = new HashSet<PetrinetNode>();
		for (Transition t : tranMap.keySet())
			if (tranMap.get(t) == 0 || tranMap.get(t) == 1)
				keepTrans.add(t);

		//reduce using reduction rules
		while (reduced) {
			reduced = false;

			reduced = useLDPlaceRule(pn);//reduce using l.d. place rule
			if (!reduced)//reduce using linearly dependent transition rule
				reduced = useLDTranRule(pn);
			if (!reduced)
				reduced = useAbstractionRule(pn, keepTrans);

		}

		//finally the net is of the form i - tran - o: use this to build the first abstraction rule
		//		Set<Integer> initTran = new HashSet<Integer>();
		//		Set<Integer> initPlace = new HashSet<Integer>();
		//		initTran.add(0);
		//		Transition t = null;
		//		for (Transition t2 : pn.getTransitions())
		//			t = t2;
		//		initPlace.add(1);
		//		Place firstPlace = null;
		//		for (Place p : pn.getPlaces())
		//			if (pn.getInEdges(p).size() < 1)
		//				firstPlace = p;
		//		reductionTracker.add(new ReductionTracker(Rule.ABSTRACTION, initTran, initPlace, t.getLabel(), t.isInvisible(),
		//				placeMap.get(firstPlace), tranMap.get(t)));

		return null;
	}

	private void init(Petrinet pn) {
		int i = 2; //start from 2 to account for 2 initial trans from ipn
		for (Transition t : pn.getTransitions())
			tranMap.put(t, i++);
		int j = 3; //start from 3 to account for the 3 initial places from ipn
		for (Place p : pn.getPlaces()) {
			if (pn.getOutEdges(p).size() < 1)
				placeMap.put(p, (int) 1); //map end place with 1 (equivalent to mid)
			else
				placeMap.put(p, j++);
		}

		//add i-dummy-mid-dummy-o structure to the Petri net
		Transition t1 = pn.addTransition("DUMMY");
		t1.setInvisible(true);
		tranMap.put(t1, (int) 0);
		Transition t2 = pn.addTransition("DUMMY");
		t2.setInvisible(true);
		tranMap.put(t2, (int) 1);
		Place p1 = pn.addPlace("i");
		placeMap.put(p1, (int) 0);
		Place p2 = pn.addPlace("o");
		placeMap.put(p2, (int) 2);
		pn.addArc(p1, t1);
		pn.addArc(t2, p2);

		//get the initial place(s) of the net
		Set<Place> startPlaces = new HashSet<Place>();
		startPlaces.add(PetrinetUtilsL.getInitialPlace(pn));
		for (Place p : startPlaces)
			pn.addArc(t1, p);

		Set<Place> endPlaces = new HashSet<Place>();
		endPlaces.add(PetrinetUtilsL.getFinalPlace(pn));
		for (Place p : endPlaces)
			pn.addArc(p, t2); //mid to t2

		//		System.out.println("Place map is " + placeMap);
		//		System.out.println("Tran map is " + tranMap);
	}

	//reduces by keeping only one escape route (and one loop)
	private boolean reduceTranLDUsingLocalInfo(Petrinet pn, Set<PetrinetNode> keepNodes, boolean exploreAllCombi) {
		//		System.out.println("here");
		Map<Transition, Integer> tranInciMap = new HashMap<Transition, Integer>();
		Map<Place, Integer> placeInciMap = new HashMap<Place, Integer>();
		SimpleMatrix inciMat = MatrixUtilsL.createInciMap(tranInciMap, placeInciMap, pn);
		SimpleMatrix augMat3 = inciMat;//MatrixUtils.getSCIncidenceMatrix2(inciMat, placeInciMap, pn);
		//				for (Transition t : tranInciMap.keySet()) {
		//					if (keepNodes.contains(t))
		//						System.out.println("Transition " + t.getLabel() + " inci id : " + tranInciMap.get(t));
		//				}

		for (Transition t : pn.getTransitions()) {
			if (keepNodes.contains(t))
				continue;
			//			if (reduceSameIpOpChoices(t, pn, keepNodes))
			//				return true;

			boolean reduce = false;
			int inciIndexOftNew = tranInciMap.get(t);
			SimpleMatrix inciMatWithoutNewLD = new SimpleMatrix(augMat3.numRows(), augMat3.numCols() - 1);
			SimpleMatrix ldCol = new SimpleMatrix(augMat3.numRows(), 1);
			MatrixUtilsL.splitInciMat(augMat3, inciMatWithoutNewLD, ldCol, inciIndexOftNew);
			boolean isLoop = false;
			Set<Set<Integer>> ldCols = ReductionUtilsL.getPossibleDistinctPathsInBetween(inciMatWithoutNewLD, ldCol,
					inciIndexOftNew, isLoop);
			if (ldCols.size() < 1) {
				isLoop = true;
				ldCols.addAll(ReductionUtilsL.getPossibleDistinctPathsInBetween(inciMatWithoutNewLD, ldCol,
						inciIndexOftNew, isLoop)); //also loops
			}

			if (exploreAllCombi) {
				ldCols.addAll(ReductionUtilsL.getPossibleDistinctPathsUsingCombinations(inciMatWithoutNewLD, ldCol,
						inciIndexOftNew, isLoop, ldCols)); //also loops

			}

			for (Set<Integer> ldColSet : ldCols) {
				Set<Transition> allTransInBw = new HashSet<Transition>();
				if (!reduce) {
					for (int i : ldColSet) {
						for (Transition tranInci : tranInciMap.keySet())
							if (tranInciMap.get(tranInci).equals(i))
								allTransInBw.add(tranInci);
					}

					if (Collections.disjoint(allTransInBw, keepNodes))
						reduce = true;
				}
				//				System.out.println(ldColSet + " ");
			}
			if (reduce) {
				for (PetrinetEdge<? extends PetrinetNode, ? extends PetrinetNode> e : pn.getInEdges(t)) {
					pn.removeArc(e.getSource(), e.getTarget());
				}
				for (PetrinetEdge<? extends PetrinetNode, ? extends PetrinetNode> e : pn.getOutEdges(t)) {
					pn.removeArc(e.getSource(), e.getTarget());
				}
				pn.removeTransition(t);
				return true;
			}
		}
		return false;
	}

	private boolean reduceSameIpOpChoices(Transition t, Petrinet pn, Set<PetrinetNode> keepNodes) {
		Place pIn = null;
		Place pOut = null;
		boolean reduce = false;
		for (PetrinetEdge<? extends PetrinetNode, ? extends PetrinetNode> e : pn.getInEdges(t)) {
			pIn = ((Place) e.getSource());
		}
		for (PetrinetEdge<? extends PetrinetNode, ? extends PetrinetNode> e : pn.getOutEdges(t)) {
			pOut = ((Place) e.getTarget());
		}

		//reduce self loop
		if (pIn.equals(pOut)) {
			reduce = true;
		}

		//if not self loop - check if another escape path exists with same ip op
		for (PetrinetEdge<? extends PetrinetNode, ? extends PetrinetNode> e : pn.getOutEdges(pIn)) {
			for (PetrinetEdge<? extends PetrinetNode, ? extends PetrinetNode> e2 : pn.getOutEdges(e.getTarget())) {
				if (!reduce && !e.getTarget().equals(t) && !keepNodes.contains(e.getTarget())
						&& e2.getTarget().equals(pOut)) {
					reduce = true;
				}
			}
		}

		if (reduce) {
			for (PetrinetEdge<? extends PetrinetNode, ? extends PetrinetNode> e : pn.getInEdges(t)) {
				pn.removeArc(e.getSource(), e.getTarget());
			}
			for (PetrinetEdge<? extends PetrinetNode, ? extends PetrinetNode> e : pn.getOutEdges(t)) {
				pn.removeArc(e.getSource(), e.getTarget());
			}
			pn.removeTransition(t);
			return true;
		}

		return false;
	}

	private boolean useAbstractionRule(Petrinet pn, Set<PetrinetNode> keepNodes) {
		for (Transition t : pn.getTransitions()) {

			Collection<PetrinetEdge<? extends PetrinetNode, ? extends PetrinetNode>> inEdges = pn.getInEdges(t);
			Collection<PetrinetEdge<? extends PetrinetNode, ? extends PetrinetNode>> outEdges = pn.getOutEdges(t);
			boolean prevPlaceIsChoiceOrMoreThanOneIncomingArcs = false;
			boolean prevPlaceIsJoin = false;
			//if prev place is choice split then cannot apply abstraction
			for (PetrinetEdge<? extends PetrinetNode, ? extends PetrinetNode> e : inEdges) {
				if (pn.getOutEdges(e.getSource()).size() > 1 || pn.getInEdges(e.getSource()).size() == 0) //if choice or init place
					prevPlaceIsChoiceOrMoreThanOneIncomingArcs = true;
				if (pn.getInEdges(e.getSource()).size() > 1)
					prevPlaceIsJoin = true;
			}
			if (pn.getInEdges(t).size() > 1)
				prevPlaceIsChoiceOrMoreThanOneIncomingArcs = true; //if more than one incoming edges, then do not use abstraction
			//reduce using abstraction rule

			if (!prevPlaceIsChoiceOrMoreThanOneIncomingArcs) {
				if (!keepNodes.contains(t)) {
					//red track parameters
					Set<Integer> tranStart = new HashSet<Integer>();
					Set<Integer> placeEnd = new HashSet<Integer>();
					boolean isSilentTran = t.isInvisible() ? true : false;
					String tranLabel = t.getLabel();
					String placeLabel = "";
					int placeRemovedID = 0;
					pn.removeNode(t);
					Set<Transition> preTranSet = new HashSet<Transition>(); //Pre tran set of the pre place of t
					Set<Place> postPlaceSet = new HashSet<Place>(); //Post place sets of t
					for (PetrinetEdge<? extends PetrinetNode, ? extends PetrinetNode> e : inEdges) {
						Collection<PetrinetEdge<? extends PetrinetNode, ? extends PetrinetNode>> inEdges2 = pn
								.getInEdges(e.getSource());
						for (PetrinetEdge<? extends PetrinetNode, ? extends PetrinetNode> e2 : inEdges2) {
							preTranSet.add((Transition) e2.getSource());
							tranStart.add(tranMap.get(e2.getSource()));
						}
						placeLabel = e.getSource().getLabel();
						pn.removeNode(e.getSource());
						placeRemovedID = placeMap.get(e.getSource());
						pn.removeArc(e.getSource(), e.getTarget());

					}
					for (PetrinetEdge<? extends PetrinetNode, ? extends PetrinetNode> e : outEdges) {
						postPlaceSet.add((Place) e.getTarget());
						placeEnd.add(placeMap.get(e.getTarget()));
						pn.removeArc(e.getSource(), e.getTarget());

					}
					//					System.out.println("Prev Tran : " + preTranSet);
					//					System.out.println("Prev Tran : " + postPlaceSet);
					for (Transition preTran : preTranSet) {
						//keep te clusters

						for (Place postPlace : postPlaceSet) {
							pn.addArc(preTran, postPlace);
						}
					}

					//create red tracker and add to trackRed

					ReductionTrackerL redTrack = new ReductionTrackerL(Rule.ABSTRACTION, tranStart, placeEnd, tranLabel,
							placeLabel, isSilentTran, placeRemovedID, tranMap.get(t));
					reductionTracker.add(redTrack);

					return true;
					//				} else if (!prevPlaceIsJoin) { //reduce the tran by replacing the prev *single* tau tran with keepnode tran
					//					//get prev tran
					//					Transition prevTran = null;
					//					for (PetrinetEdge<? extends PetrinetNode, ? extends PetrinetNode> e : inEdges) {
					//						for (PetrinetEdge<? extends PetrinetNode, ? extends PetrinetNode> e2 : pn.getInEdges(e
					//								.getSource())) {
					//							prevTran = (Transition) e2.getSource();
					//						}
					//					}
					//
					//					if (!keepNodes.contains(prevTran)) {
					//						pn.removeNode(prevTran);
					//						Set<Place> prePlaceSet = new HashSet<Place>(); //pre place sets of prevTran
					//						for (PetrinetEdge<? extends PetrinetNode, ? extends PetrinetNode> e : pn.getInEdges(prevTran)) {
					//							prePlaceSet.add((Place) e.getSource());
					//							pn.removeArc(e.getSource(), e.getTarget());
					//						}
					//						for (PetrinetEdge<? extends PetrinetNode, ? extends PetrinetNode> e : pn.getOutEdges(prevTran)) {
					//							pn.removeArc(e.getSource(), e.getTarget());
					//							pn.removeArc(e.getTarget(), t);
					//							pn.removeNode(e.getTarget());
					//						}
					//
					//						for (Place prePlace : prePlaceSet) {
					//							pn.addArc(prePlace, t);
					//						}
					//					}
					//					return true;
				}
			}
		}
		return false;

	}

	private boolean useLDTranRule(Petrinet pn) {
		Map<Transition, Integer> tranInciMap = new HashMap<Transition, Integer>();
		Map<Place, Integer> placeInciMap = new HashMap<Place, Integer>();
		SimpleMatrix inciMat = MatrixUtilsL.createInciMap(tranInciMap, placeInciMap, pn);
		SimpleMatrix augMat3 = MatrixUtilsL.getSCIncidenceMatrix2(inciMat, placeInciMap, pn);

		for (Transition t : pn.getTransitions()) {
			int inciIndexOftNew = tranInciMap.get(t);
			SimpleMatrix inciMatWithoutNewLD = new SimpleMatrix(augMat3.numRows(), augMat3.numCols() - 1);
			SimpleMatrix ldCol = new SimpleMatrix(augMat3.numRows(), 1);
			MatrixUtilsL.splitInciMat(augMat3, inciMatWithoutNewLD, ldCol, inciIndexOftNew);
			if (MatrixUtilsL.isAugMatLD(inciMatWithoutNewLD, ldCol)) {
				//red track params
				Set<Integer> startPlaces = new HashSet<Integer>();
				Set<Integer> endPlaces = new HashSet<Integer>();
				//red track params

				Collection<PetrinetEdge<? extends PetrinetNode, ? extends PetrinetNode>> inEdges = pn.getInEdges(t);
				Collection<PetrinetEdge<? extends PetrinetNode, ? extends PetrinetNode>> outEdges = pn.getOutEdges(t);
				for (PetrinetEdge<? extends PetrinetNode, ? extends PetrinetNode> e : inEdges) {
					startPlaces.add(placeMap.get(e.getSource()));
					pn.removeArc(e.getSource(), e.getTarget());
				}
				for (PetrinetEdge<? extends PetrinetNode, ? extends PetrinetNode> e : outEdges) {
					endPlaces.add(placeMap.get(e.getTarget()));
					pn.removeArc(e.getSource(), e.getTarget());
				}
				pn.removeTransition(t);
				reductionTracker.add(new ReductionTrackerL(Rule.LDTRAN, startPlaces, endPlaces, t.getLabel(), "", t
						.isInvisible(), (int) 0, tranMap.get(t)));
				return true;
			}
		}
		return false;
	}

	private boolean useLDPlaceRule(Petrinet pn) {
		Map<Transition, Integer> tranInciMap = new HashMap<Transition, Integer>();
		Map<Place, Integer> placeInciMap = new HashMap<Place, Integer>();
		SimpleMatrix inciMat = MatrixUtilsL.createInciMap(tranInciMap, placeInciMap, pn);
		SimpleMatrix augMat3 = MatrixUtilsL.getSCIncidenceMatrix2(inciMat, placeInciMap, pn);

		for (Place p : pn.getPlaces()) {
			if (pn.getInEdges(p).size() < 1) //if init place then continue
				continue;

			int inciIndexOfpNew = placeInciMap.get(p);
			SimpleMatrix inciMatWithoutNewLD = new SimpleMatrix(augMat3.numCols(), augMat3.numRows() - 1);
			SimpleMatrix ldCol = new SimpleMatrix(augMat3.numCols(), 1);
			MatrixUtilsL.splitInciMat(augMat3.transpose(), inciMatWithoutNewLD, ldCol, inciIndexOfpNew);
			if (MatrixUtilsL.isAugMatLD(inciMatWithoutNewLD, ldCol)) {
				//red track params
				Set<Integer> startTrans = new HashSet<Integer>();
				Set<Integer> endTrans = new HashSet<Integer>();
				//red track params

				Collection<PetrinetEdge<? extends PetrinetNode, ? extends PetrinetNode>> inEdges = pn.getInEdges(p);
				Collection<PetrinetEdge<? extends PetrinetNode, ? extends PetrinetNode>> outEdges = pn.getOutEdges(p);
				for (PetrinetEdge<? extends PetrinetNode, ? extends PetrinetNode> e : inEdges) {
					startTrans.add(tranMap.get(e.getSource()));
					pn.removeArc(e.getSource(), e.getTarget());
				}
				for (PetrinetEdge<? extends PetrinetNode, ? extends PetrinetNode> e : outEdges) {
					endTrans.add(tranMap.get(e.getTarget()));
					pn.removeArc(e.getSource(), e.getTarget());
				}
				pn.removePlace(p);
				reductionTracker.add(new ReductionTrackerL(Rule.LDPLACE, startTrans, endTrans, "", p.getLabel(), false,
						placeMap.get(p), (int) 0));
				return true;
			}
		}
		//performance starts
		return false;
	}

	private boolean getAllClusters(Petrinet pn, Set<Set<Transition>> actsSeqMap) {
		for (Transition t : pn.getTransitions()) {

			Collection<PetrinetEdge<? extends PetrinetNode, ? extends PetrinetNode>> inEdges = pn.getInEdges(t);
			Collection<PetrinetEdge<? extends PetrinetNode, ? extends PetrinetNode>> outEdges = pn.getOutEdges(t);
			boolean prevPlaceIsChoiceOrMoreThanOneIncomingArcs = false;
			boolean prevPlaceIsJoin = false;
			//if prev place is choice split then cannot apply abstraction
			for (PetrinetEdge<? extends PetrinetNode, ? extends PetrinetNode> e : inEdges) {
				if (pn.getOutEdges(e.getSource()).size() > 1 || pn.getInEdges(e.getSource()).size() == 0) //if choice or init place
					prevPlaceIsChoiceOrMoreThanOneIncomingArcs = true;
				if (pn.getInEdges(e.getSource()).size() > 1)
					prevPlaceIsJoin = true;
			}
			if (pn.getInEdges(t).size() > 1)
				prevPlaceIsChoiceOrMoreThanOneIncomingArcs = true; //if more than one incoming edges, then do not use abstraction
			//reduce using abstraction rule

			if (!prevPlaceIsChoiceOrMoreThanOneIncomingArcs) {

				//red track parameters
				Set<Integer> tranStart = new HashSet<Integer>();
				Set<Integer> placeEnd = new HashSet<Integer>();
				boolean isSilentTran = t.isInvisible() ? true : false;
				String tranLabel = t.getLabel();
				String placeLabel = "";
				int placeRemovedID = 0;
				pn.removeNode(t);
				Set<Transition> preTranSet = new HashSet<Transition>(); //Pre tran set of the pre place of t
				Set<Place> postPlaceSet = new HashSet<Place>(); //Post place sets of t
				for (PetrinetEdge<? extends PetrinetNode, ? extends PetrinetNode> e : inEdges) {
					Collection<PetrinetEdge<? extends PetrinetNode, ? extends PetrinetNode>> inEdges2 = pn.getInEdges(e
							.getSource());
					for (PetrinetEdge<? extends PetrinetNode, ? extends PetrinetNode> e2 : inEdges2) {
						preTranSet.add((Transition) e2.getSource());
						tranStart.add(tranMap.get(e2.getSource()));
					}
					pn.removeNode(e.getSource());
					placeRemovedID = placeMap.get(e.getSource());
					pn.removeArc(e.getSource(), e.getTarget());

				}
				for (PetrinetEdge<? extends PetrinetNode, ? extends PetrinetNode> e : outEdges) {
					postPlaceSet.add((Place) e.getTarget());
					placeEnd.add(placeMap.get(e.getTarget()));
					pn.removeArc(e.getSource(), e.getTarget());

				}

				Set<Transition> preTranCl = new HashSet<Transition>();
				for (Transition preTran : preTranSet) {

					preTranCl.add(preTran);

					if (preTran.getLabel().contains("110") || preTran.getLabel().contains("120")
							|| preTran.getLabel().contains("180") || t.getLabel().contains("110")
							|| t.getLabel().contains("120") || t.getLabel().contains("180"))
						System.out.println();

					for (Place postPlace : postPlaceSet) {
						pn.addArc(preTran, postPlace);
					}
				}

				preTranCl.add(t);

				boolean added = false;
				for (Set<Transition> clu : actsSeqMap)
					if (!preTranCl.isEmpty() && !Collections.disjoint(preTranCl, clu)) {
						clu.addAll(preTranCl);
						added = true;
					}
				if (!added)
					actsSeqMap.add(preTranCl);

				return true;

			}
		}
		return false;

	}
}
