package org.processmining.reduction;

import java.util.Collection;
import java.util.HashMap;
import java.util.Map;
import java.util.Set;

import org.ejml.simple.SimpleMatrix;
import org.processmining.models.graphbased.directed.petrinet.Petrinet;
import org.processmining.models.graphbased.directed.petrinet.PetrinetEdge;
import org.processmining.models.graphbased.directed.petrinet.PetrinetNode;
import org.processmining.models.graphbased.directed.petrinet.elements.Place;
import org.processmining.models.graphbased.directed.petrinet.elements.Transition;
import org.processmining.prabhakardixit.fcnets.matrix.MatrixUtilsL;

public class ReductionNewL {

	public ReductionNewL(Petrinet pn, Set<Transition> trans) {
		reduceNet(pn, trans);
	}

	public void reduceNet(Petrinet pn, Set<Transition> trans) {
		boolean reduced = true;

		//first reduce using abstraction rule
		while (reduced) {
			reduced = useAbstractionRule(pn, trans);
			//			if (!reduced) //reduce using l.d. place rule
			//				reduced = useLDPlaceRule(pn);
			//			if (!reduced)
			//				reduced = useLDTranRule(pn, trans);
		}
		reduced = useAbstractionRule(pn, trans);
	}

	//reduces by keeping only one escape route (and one loop)
	private boolean useAbstractionRule(Petrinet pn, Set<Transition> keepNodes) {
		for (Transition t : pn.getTransitions()) {

			Collection<PetrinetEdge<? extends PetrinetNode, ? extends PetrinetNode>> inEdges = pn.getInEdges(t);
			Collection<PetrinetEdge<? extends PetrinetNode, ? extends PetrinetNode>> outEdges = pn.getOutEdges(t);
			boolean prevPlaceIsChoiceOrMoreThanOneIncomingArcs = false;
			Place p = null;
			//if prev place is choice split then cannot apply abstraction
			for (PetrinetEdge<? extends PetrinetNode, ? extends PetrinetNode> e : inEdges) {
				if (pn.getOutEdges(e.getSource()).size() > 1 || pn.getInEdges(e.getSource()).size() == 0) //if choice or init place
					prevPlaceIsChoiceOrMoreThanOneIncomingArcs = true;

			}
			if (pn.getInEdges(t).size() > 1)
				prevPlaceIsChoiceOrMoreThanOneIncomingArcs = true; //if more than one incoming edges, then do not use abstraction
			//reduce using abstraction rule

			if (!prevPlaceIsChoiceOrMoreThanOneIncomingArcs) {
				if (!keepNodes.contains(t)) {

					//					Set<Transition> preTranSet = new HashSet<Transition>(); //Pre tran set of the pre place of t
					//					Set<Place> postPlaceSet = new HashSet<Place>(); //Post place sets of t
					for (PetrinetEdge<? extends PetrinetNode, ? extends PetrinetNode> e : pn.getInEdges(t)) {
						p = (Place) e.getSource();
						for (PetrinetEdge<? extends PetrinetNode, ? extends PetrinetNode> e2 : pn.getInEdges(p)) { //inEdges of place
							for (PetrinetEdge<? extends PetrinetNode, ? extends PetrinetNode> e3 : pn.getOutEdges(t)) { //outedges of tran
								pn.addArc((Transition) e2.getSource(), (Place) e3.getTarget());
								//								pn.removeArc(t, e3.getTarget());
							}
							//							pn.removeArc(e2.getSource(), p);
						}

						pn.removeNode(p);
						pn.removeNode(t);
						pn.removeArc(p, t);

					}

					return true;

//				} else {
//					//if R is invisi
//					boolean rInvi = true;
//					for (PetrinetEdge<? extends PetrinetNode, ? extends PetrinetNode> e : pn.getInEdges(t)) {
//						p = (Place) e.getSource();
//						for (PetrinetEdge<? extends PetrinetNode, ? extends PetrinetNode> e2 : pn.getInEdges(p))
//							//inEdges of place
//							if (!((Transition) e2.getSource()).isInvisible())
//								rInvi = false;
//						if (rInvi) {
//							for (PetrinetEdge<? extends PetrinetNode, ? extends PetrinetNode> e2 : pn.getInEdges(p)) {
//								e2.getSource().getAttributeMap().put(AttributeMap.LABEL, t.getLabel());
//								keepNodes.remove(t);
//								keepNodes.add((Transition) e2.getSource());
//								for (PetrinetEdge<? extends PetrinetNode, ? extends PetrinetNode> e3 : pn
//										.getOutEdges(t))
//									pn.addArc((Transition) e2.getSource(), (Place) e3.getTarget());
//
//							}
//							pn.removeNode(p);
//							pn.removeNode(t);
//							pn.removeArc(p, t);
//							return true;
//						}
//					}
				}
			}
		}
		return false;

	}

	private boolean useLDTranRule(Petrinet pn, Set<Transition> trans) {
		Map<Transition, Integer> tranInciMap = new HashMap<Transition, Integer>();
		Map<Place, Integer> placeInciMap = new HashMap<Place, Integer>();

		SimpleMatrix inciMat = MatrixUtilsL.createInciMapWithoutSomeTrans(tranInciMap, placeInciMap, trans, pn);
		SimpleMatrix augMat3 = MatrixUtilsL.getSCIncidenceMatrix2(inciMat, placeInciMap, pn);

		for (Transition t : pn.getTransitions()) {
			if (!trans.contains(t)) {
				int inciIndexOftNew = tranInciMap.get(t);
				SimpleMatrix inciMatWithoutNewLD = new SimpleMatrix(augMat3.numRows(), augMat3.numCols() - 1);
				SimpleMatrix ldCol = new SimpleMatrix(augMat3.numRows(), 1);
				MatrixUtilsL.splitInciMat(augMat3, inciMatWithoutNewLD, ldCol, inciIndexOftNew);
				if (MatrixUtilsL.isAugMatLD(inciMatWithoutNewLD, ldCol)) {
					for (PetrinetEdge<? extends PetrinetNode, ? extends PetrinetNode> e : pn.getInEdges(t)) {
						pn.removeArc(e.getSource(), e.getTarget());
					}
					for (PetrinetEdge<? extends PetrinetNode, ? extends PetrinetNode> e : pn.getOutEdges(t)) {
						pn.removeArc(e.getSource(), e.getTarget());
					}
					pn.removeNode(t);

					return true;
				}
			}
		}
		return false;
	}

	private boolean useLDPlaceRule(Petrinet pn) {
		Map<Transition, Integer> tranInciMap = new HashMap<Transition, Integer>();
		Map<Place, Integer> placeInciMap = new HashMap<Place, Integer>();
		SimpleMatrix inciMat = MatrixUtilsL.createInciMap(tranInciMap, placeInciMap, pn);
		SimpleMatrix augMat3 = MatrixUtilsL.getSCIncidenceMatrix2(inciMat, placeInciMap, pn);

		for (Place p : pn.getPlaces()) {
			if (pn.getInEdges(p).size() < 1 || pn.getOutEdges(p).size() < 1) //if init place then continue
				continue;

			int inciIndexOfpNew = placeInciMap.get(p);
			SimpleMatrix inciMatWithoutNewLD = new SimpleMatrix(augMat3.numCols(), augMat3.numRows() - 1);
			SimpleMatrix ldCol = new SimpleMatrix(augMat3.numCols(), 1);
			MatrixUtilsL.splitInciMat(augMat3.transpose(), inciMatWithoutNewLD, ldCol, inciIndexOfpNew);
			if (MatrixUtilsL.isAugMatLD(inciMatWithoutNewLD, ldCol)) {

				for (PetrinetEdge<? extends PetrinetNode, ? extends PetrinetNode> e : pn.getInEdges(p)) {
					pn.removeArc(e.getSource(), e.getTarget());
				}
				for (PetrinetEdge<? extends PetrinetNode, ? extends PetrinetNode> e : pn.getOutEdges(p)) {
					pn.removeArc(e.getSource(), e.getTarget());
				}
				pn.removeNode(p);

				return true;
			}
		}
		//performance starts
		return false;
	}

}
