package org.processmining.reduction;

import java.util.ArrayList;
import java.util.Collection;
import java.util.HashSet;
import java.util.Set;

import org.apache.commons.math3.optim.MaxIter;
import org.apache.commons.math3.optim.PointValuePair;
import org.apache.commons.math3.optim.linear.LinearConstraint;
import org.apache.commons.math3.optim.linear.LinearConstraintSet;
import org.apache.commons.math3.optim.linear.LinearObjectiveFunction;
import org.apache.commons.math3.optim.linear.NonNegativeConstraint;
import org.apache.commons.math3.optim.linear.Relationship;
import org.apache.commons.math3.optim.linear.SimplexSolver;
import org.apache.commons.math3.optim.nonlinear.scalar.GoalType;
import org.ejml.simple.SimpleMatrix;
import org.processmining.prabhakardixit.alignments.CombinationUtilsL;

public class ReductionUtilsL {
	//gets all nodes in between
	public static Set<Set<Integer>> getPossibleDistinctPathsInBetween(SimpleMatrix inciMatrixWithoutLastCol,
			SimpleMatrix linDepMatrix, int indexOfNewNodeInInciMat, boolean isLoop) {
		Set<Set<Integer>> ldCols = new HashSet<Set<Integer>>();
		Set<Integer> allLdCols = new HashSet<Integer>();
		//		MatrixUtils.printMatrix(inciMatrixWithoutLastCol);
		//		MatrixUtils.printMatrix(linDepMatrix);
		boolean exec = true;
		while (exec) {
			double[] coeff = new double[inciMatrixWithoutLastCol.numCols()];
			LinearObjectiveFunction f = new LinearObjectiveFunction(coeff, 0);
			Collection<LinearConstraint> constraints = new ArrayList<LinearConstraint>();
			Collection<LinearConstraint> constraints_Loop = new ArrayList<LinearConstraint>();
			for (int r = 0; r < inciMatrixWithoutLastCol.numRows(); r++) {
				double[] inciMat = new double[inciMatrixWithoutLastCol.numCols()];
				for (int c = 0; c < inciMatrixWithoutLastCol.numCols(); c++) {
					if (!allLdCols.contains(c))
						inciMat[c] = inciMatrixWithoutLastCol.get(r, c);
				}
				constraints.add(new LinearConstraint(inciMat, Relationship.EQ, linDepMatrix.get(r, 0)));
				constraints_Loop.add(new LinearConstraint(inciMat, Relationship.EQ, -linDepMatrix.get(r, 0)));
			}
			try {
				SimplexSolver solver = new SimplexSolver();
				PointValuePair solution;
				if (!isLoop)
					solution = solver.optimize(new MaxIter(100), f, new LinearConstraintSet(constraints),
							GoalType.MINIMIZE, new NonNegativeConstraint(true));
				else
					solution = solver.optimize(new MaxIter(100), f, new LinearConstraintSet(constraints_Loop),
							GoalType.MINIMIZE, new NonNegativeConstraint(true));
				Set<Integer> ldColSet = new HashSet<Integer>();
				for (int c = 0; c < solution.getKey().length; c++) {
					double val = solution.getKey()[c];
					if (val != 0.)
						if (val != 1.)
							System.out.println("Error!!");
						else {
							if (c >= indexOfNewNodeInInciMat) {
								ldColSet.add(c + 1); //since ld col is removed from position 'indexOfNewNodeInInciMat', add 1 to compensate 
								allLdCols.add(c);
							} else {
								ldColSet.add(c);
								allLdCols.add(c);
							}
						}
				}
				if (ldColSet.size() == 0)//if self loop, then exit
					exec = false;
				ldCols.add(ldColSet);
			} catch (Exception e) {
				exec = false;
			}
		}

		return ldCols;
	}

	public static Set<Set<Integer>> getPossibleDistinctPathsUsingCombinations(SimpleMatrix inciMatrixWithoutLastCol,
			SimpleMatrix linDepMatrix, int indexOfNewNodeInInciMat, boolean isLoop, Set<Set<Integer>> transInBw) {
		Set<Set<Integer>> ldCols = new HashSet<Set<Integer>>();
		Set<Set<Integer>> combinations = new HashSet<Set<Integer>>();
		//get all the combinations 
		for (Set<Integer> set : transInBw) {
			combinations.addAll(CombinationUtilsL.getAllCombs(set));
		}
		for (Set<Integer> set : combinations) {
			double[] coeff = new double[inciMatrixWithoutLastCol.numCols()];
			LinearObjectiveFunction f = new LinearObjectiveFunction(coeff, 0);
			Collection<LinearConstraint> constraints = new ArrayList<LinearConstraint>();
			Collection<LinearConstraint> constraints_Loop = new ArrayList<LinearConstraint>();
			for (int r = 0; r < inciMatrixWithoutLastCol.numRows(); r++) {
				double[] inciMat = new double[inciMatrixWithoutLastCol.numCols()];
				for (int c = 0; c < inciMatrixWithoutLastCol.numCols(); c++) {
					if (!set.contains(c))
						inciMat[c] = inciMatrixWithoutLastCol.get(r, c);
				}
				constraints.add(new LinearConstraint(inciMat, Relationship.EQ, linDepMatrix.get(r, 0)));
				constraints_Loop.add(new LinearConstraint(inciMat, Relationship.EQ, -linDepMatrix.get(r, 0)));
			}
			try {
				SimplexSolver solver = new SimplexSolver();
				PointValuePair solution;
				if (!isLoop)
					solution = solver.optimize(new MaxIter(100), f, new LinearConstraintSet(constraints),
							GoalType.MINIMIZE, new NonNegativeConstraint(true));
				else
					solution = solver.optimize(new MaxIter(100), f, new LinearConstraintSet(constraints_Loop),
							GoalType.MINIMIZE, new NonNegativeConstraint(true));
				Set<Integer> ldColSet = new HashSet<Integer>();
				for (int c = 0; c < solution.getKey().length; c++) {
					double val = solution.getKey()[c];
					if (val != 0.)
						if (val != 1.)
							System.out.println("Error!!");
						else {
							if (c >= indexOfNewNodeInInciMat) {
								ldColSet.add(c + 1); //since ld col is removed from position 'indexOfNewNodeInInciMat', add 1 to compensate 
							} else {
								ldColSet.add(c);
							}
						}
				}
				ldCols.add(ldColSet);

			} catch (Exception e) {
				//catch ex and continue - non ld so do nothing
			}
		}
		return ldCols;
	}
}
